/* 2000, Tue 03 Jun 03 (PDT)

   FLOWHASH.C:  RTFM Meter, using hash tables for counts

   Copyright (C) 1992-2003 by Nevil Brownlee,
   CAIDA | University of Auckland */

/*
 * $Log: flowhash.c,v $
 * Revision 1.4  2003/12/08 22:01:55  uid512
 * First production release of NeTraMet++
 *
 * Revision 1.1.1.2.2.22  2002/10/31 23:02:30  nevil
 * 45b8: fd_filter 'tail' option implemented, \t supported as format separator
 *
 * Revision 1.1.1.2.2.21  2002/09/13 15:05:18  nevil
 * Implement VLANid, Priority and ECNCodeBits attributes
 *
 * Revision 1.1.1.2.2.20  2002/08/27 17:13:12  nevil
 * Improve thread locking.  Implement STREAM_TO_FLOW, ECNBits and TCPStreamData
 *
 * Revision 1.1.1.2.2.19  2002/07/12 01:41:36  nevil
 * Fix attribute bug in fd_filter
 *
 * Revision 1.1.1.2.2.18  2002/07/03 01:45:49  nevil
 * Implement OAM attributes
 *
 * Revision 1.1.1.2.2.17  2002/06/30 01:07:57  nevil
 * Make *Interface attribs 16-bit (instead of 8)
 *
 * Revision 1.1.1.2.2.16  2002/06/29 01:55:50  nevil
 * Miscellaneous fixes for warnings found by Compaq Unix cc
 *
 * Revision 1.1.1.2.2.12  2001/05/24 02:19:49  nevil
 * LfapMet implemented by Remco Poortinga.
 * MinPDUs implemented by Nevil.
 *
 * Revision 1.1.1.2.2.10  2000/12/28 00:48:13  nevil
 * Count unrequested DNS responses as LostPDUs
 *
 * Revision 1.1.1.2.2.8  2000/10/26 23:23:58  nevil
 * Write log_msg messages to meter.log file
 *
 * Revision 1.1.1.2.2.7  2000/08/08 19:44:49  nevil
 * 44b8 release
 *
 * Revision 1.1.1.2.2.5  2000/06/06 03:38:17  nevil
 * Combine NEW_ATR with TCP_ATR, various bug fixes
 *
 * Revision 1.1.1.2.2.2  2000/01/12 02:57:08  nevil
 * Implement 'packet pair matched' turnaroundtime distribution attributes.
 * Fix ASN-related bugs in NeTraMet, distribution-related bugs in fd_filter.
 *
 * Revision 1.1.1.2.2.1  1999/11/29 00:17:24  nevil
 * Make changes to support NetBSD on an Alpha (see version.history for details)
 *
 * Revision 1.1.1.2  1999/10/03 21:06:22  nevil
 * *** empty log message ***
 *
 * Revision 1.1.1.1.2.17  1999/09/29 22:29:14  nevil
 * Changes (mainly changing // to /* comments) for Irix
 *
 * Revision 1.1.1.1.2.16  1999/09/29 02:42:47  nevil
 * Fix problems discovered on PC meters
 * - Use NETRAMET compile-time variable to leave out C++ declarations
 * - Clean up mistakes in porting OCX_BGP code from meter_pc to meter_oc
 *
 * Revision 1.1.1.1.2.15  1999/09/24 02:58:37  nevil
 * Polish up code to get rid of warning messages from Borland (DOS) compiler.
 * Make manager PeerAddress buffers NSAP_ADDR_LEN bytes long.
 * Add asn_lookup variable - only call FindSubnet if ruleset uses ASNs.
 *
 * Revision 1.1.1.1.2.14  1999/09/22 05:38:40  nevil
 * Improve code to work properly on 64-bit machines
 * - Add OS=ALPHA handling to configure.in
 * - Clean up the Alpha compiler warnings
 * - Change all the snmp-related code to use Bit32 instead of unsigned long
 *
 * Revision 1.1.1.1.2.13  1999/09/14 00:45:26  nevil * 4.3 Release ..
 *  - Implement -D option to run NeTraMet as a daemon
 *  - Tidy up the on-line help displays
 *
 * Revision 1.1.1.1.2.12  1999/05/26 02:41:37  nevil
 * Integrate V6 and ASN code into PC versions of the meter.
 * This required a rework of the makefiles, using @cflags.opt files
 * to provide a much longer command line to the Borland C compiler.
 *
 * Revision 1.1.1.1.2.11  1999/05/18 03:36:26  nevil
 * Implement IPv6 in NeTraMet, and its manager/collectors.
 * - This is controlled by the V6 #define
 * - NeTraMet recognises v6 packets and fishes through their extension
 *     headers until it finds the actual payload.
 * - NeMaC et al display v6 addresses in the fom specified by RFC 2373
 * - fd_util and fd_extract allow colons in addresses (for defining tags)
 *
 * Revision 1.1.1.1.2.10  1999/05/17 00:11:53  nevil
 * Fixed 'reuse of ruleset row' problem:
 *   When a ruleset row is destroyed, set_RowStatus calls free_rulespace()
 *   and recover_flows().  free_rulespace() zeroed its varibales in the
 *   ri[] row, but recover_flows() didn't (it should have zeroed
 *   ri_flow_chain).  This left ri_flow_chain pointing to a free flow,
 *   which cause problems if a manager tried to reuse this row for a
 *   new ruleset.  Fixed by zeroing all of the ri[] row. (Also commented
 *   out the zeroing code from free_rulesace).
 *
 */

#if HAVE_CONFIG_H
#include <ntm_conf.h>
#endif

#define DEBUG 1   /* Required if any of the following are set */

#define PME_TRACE         0  /* Trace rules during pattern match */
#define PM_DEBUG          0

#define PP_DEBUG          0
#define PP_DEBUG_XX       0  /* 2 */
#define DNS_ROOT_DEBUG    0  /* 1 */
#define DNS_ROOT_DEBUG_2  0
#define PP_DNS_TRACE      0  /* Needs DNS_ROOT_DEBUG <<<<<<<<<<<<<<<< */

#define STREAM_CHAIN_DEBUG  0
#define SCH_CHECK         0  /* Check stream hash */
#define STH_CHECK         0  /* Check stream termination state */
#define STHC_CHECK        0  /* Check stream hash chains */

#define DIST_DBG          0
#define RS_CH_DEBUG       0
#define V6_TESTV4         0  /* Turns IP4 addresses into ::IP4 addresses */
#define noSTACK_CHECK     /* Keep an eye on pattern & return stack extents */
#define noGC_TEST         /* Monitor garbage collector */

#define noMATCH_TRACE     /* Testing, testing .. */
#define noCOMPARE_TRACE
#define noCOUNT_TRACE
#define noP_TRACE
#define noGC_DEBUG
#define noTESTING
#define GC_TEST1          0
#define H_TEST            0  /* Dump ruleset hash table info */
#define TCP_TESTING       0
#define TCP_TRACE         0
#define TCP_TAT_TRACE     0
#define OAM_TRACE         0

#define V6_TRACE         V6  /* Print V6 packets as we see them */
#define USER_DUMMY_IDS    0  /* Add new users when we see them */

#define LOG_STREAM_DATA   0  /* Write streams.log file */

#define STR_TO_TRACE      0  /* Trace check_*_idle_streams() */
#define EVNT_CHK_TRACE    0  /* Trace event handler */
#define SCC_CHECK         0  /* Check stream chains */
#define STF_DEBUG         0  /* Trace stream-to-flow */
#define P2PK_DEBUG        0  /* Trace p2pKind attribute handling */

#define P2P_TEST_TAGS     0  /* Tag pkts as 'string-matched p2p', use
                                   those tags to identify further streams */
#define LOG_P2P_TAGS      0  /* Write p2p.log file */
#define LOG_HOST_LIST     0  /* Display host-host records */

#define DECNET            0
 
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdarg.h>
#if NEW_ATR
#include <math.h>
#endif

# if HAVE_MALLOC_H
# include <malloc.h>
# endif
#include <sys/time.h>
#include <netinet/in.h>
#include "../bgp/types/types.h"  /* Needed here for subnet.h */

#include "ausnmp.h"

#include "asn1.h"
#include "snmp.h"
#include "snmpimpl.h"

#include <sys/types.h>
#include <string.h>

#if MULTI
# include <pthread.h>
# include "threads/semaphore.h"
#endif

#include "pktsnap.h"

#define EXTFLOW
#include "flowhash.h"
#include "decnet.h"
#include "met_vars.h"

#if MULTI
void init_mutex(pthread_mutex_t *m, char *name);
void destroy_mutex(pthread_mutex_t *m, char *name);

# if MUTEX_TRACE
#define MUTEX_HISTORY_SZ  100
char mutex_name[MUTEX_HISTORY_SZ][5];
pthread_mutex_t *mutex_addr[MUTEX_HISTORY_SZ];
int mutex_lk[MUTEX_HISTORY_SZ];
int mutex_nxt = 0;

void save_mutex_history(char *caller, pthread_mutex_t *m, char lk)
{
   strcpy(mutex_name[mutex_nxt], caller);
   mutex_addr[mutex_nxt] = m;
   mutex_lk[mutex_nxt] = lk;
   if (++mutex_nxt == MUTEX_HISTORY_SZ) mutex_nxt = 0;
   }
# endif
#endif

#ifdef DEBUG
void quack(int x) { }  /* Use for a debugger break point */

void paddr(Bit32 a);
void daddr(Bit32 a);
void dport(Bit16 a);
void fdaddr(FILE *f, Bit32 a);
void fdport(FILE *f, Bit16 a);
void pkey(char *msg, struct flow *n);
#endif

#if PM_DEBUG
int pm_dbg = 0;
char dbg_msg[80];
#endif


#if NF_OCX_BGP
#  include "../bgp/integrat/readbgp.h"
#  include "../bgp/integrat/subnet.h"
#endif

void dump_flow(struct flow *fp)
{
   Bit32 pa;
   log_msg(LOG_WARNING, -1,
      "   flow %u (%x), Ruleset=%d, PeerAddrType=%d, TransAddrType=%d,"
      " FlowKind=%d, FlowClass=%d",
      flow_nbr(fp),fp, fp->FlowRuleSet, fp->fk.PeerAddrType, 
      fp->fk.TransAddrType, fp->fk.FlowKind, fp->fk.FlowClass);
#if QBYTE_PEER
   pa = ntohl(fp->fk.Low.PeerAddress.qbyte);
#else
   pa = ntohl(fp->fk.Low.PeerAddress.peer32[0]);
#endif
   log_msg(LOG_WARNING, -1,
      "   Source Interface=%d, PeerAddress=%d.%d.%d.%d, TransAddress=%u",
      fp->fk.Low.Interface,
      pa >> 24, (pa >> 16) & 0xFF, (pa >> 8) & 0xFF, pa & 0xFF,
      ntohs(fp->fk.Low.Interface), ntohs(fp->fk.Low.TransAddress));
#if QBYTE_PEER
   pa = ntohl(fp->fk.High.PeerAddress.qbyte);
#else
   pa = ntohl(fp->fk.High.PeerAddress.peer32[0]);
#endif
   log_msg(LOG_WARNING, -1,
      "   Dest Interface=%d, PeerAddress=%d.%d.%d.%d, TransAddress=%u",
      pa >> 24, (pa >> 16) & 0xFF, (pa >> 8) & 0xFF, pa & 0xFF,
      ntohs(fp->fk.High.Interface), ntohs(fp->fk.High.TransAddress));
   }

void dump_bp(struct pkt *bp)
{
   Bit32 pa;
   log_msg(LOG_WARNING, -1,
      "pkt %x, PeerAddrType=%d, TransAddrType=%d",
      bp, bp->PeerAddrType, bp->TransAddrType);
#if QBYTE_PEER
   pa = ntohl(bp->Low.PeerAddress.qbyte);
#else
   pa = ntohl(bp->Low.PeerAddress.peer32[0]);
#endif
   log_msg(LOG_WARNING, -1,
      "   Source PeerAddress=%d.%d.%d.%d, TransAddress=%u",
      pa >> 24, (pa >> 16) & 0xFF, (pa >> 8) & 0xFF, pa & 0xFF,
      ntohs(bp->Low.TransAddress));
#if QBYTE_PEER
   pa = ntohl(bp->High.PeerAddress.qbyte);
#else
   pa = ntohl(bp->High.PeerAddress.peer32[0]);
#endif
   log_msg(LOG_WARNING, -1,
      "   Dest PeerAddress=%d.%d.%d.%d, TransAddress=%u",
      pa >> 24, (pa >> 16) & 0xFF, (pa >> 8) & 0xFF, pa & 0xFF,
      ntohs(bp->High.TransAddress));
   }

void dump_sp_key(struct stream *sp)
{
   Bit32 pa;
   struct stream_name *sfn = &sp->sfn;
   log_msg(LOG_WARNING, -1, "stream %x: TransAddrType=%d",
      sp, sfn->TransType);
#if QBYTE_PEER
   pa = ntohl(sfn->SrcPeerAddr);
#else
   pa = ntohl(sfn->SrcPeerAddr[0]);
#endif
   log_msg(LOG_WARNING, -1,
      "   Source PeerAddress=%d.%d.%d.%d, TransAddress=%u",
      pa >> 24, (pa >> 16) & 0xFF, (pa >> 8) & 0xFF, pa & 0xFF,
      ntohs(sfn->SrcTransAddr));
#if QBYTE_PEER
   pa = ntohl(sfn->DestPeerAddr);
#else
   pa = ntohl(sfn->DestPeerAddr[0]);
#endif
   log_msg(LOG_WARNING, -1,
      "   Dest PeerAddress=%d.%d.%d.%d, TransAddress=%u",
      pa >> 24, (pa >> 16) & 0xFF, (pa >> 8) & 0xFF, pa & 0xFF,
      ntohs(sfn->DestTransAddr));
   }

#if STREAM_CHAIN_DEBUG
int mx_chain_len = 0;

int check_flow_hash_chains(char *msg)
{
   int x, n, rv = 0;
   struct flow **sf;
   struct flow *f, *lf;

   for (x = 0; x != fthashmod; ++x) {
      sf = &flow_ht[x];
      if ((f = sf[0]) == NULL) {
         lf = NULL;  /* Empty hash chain */
         }
      else {
         n = 0; do {
            lf = f;
            if (++n > mx_chain_len) {
               printf("check_f_s_h(%s): hash %d has > %d flows <<<<\n",
                  msg, x, ++mx_chain_len);
               if (n > 15) rv = 1;
               }
            } while ((f = lf->ht_next) != (struct flow *)sf);
         f = NULL;  /* Loop stops with lf -> last flow in chain */
         }
      }
   return rv;
   }

void dump_flow_key(struct flow_key *fkp, int hash, char *msg)
{
   int j;
   printf("flow_key(%s) fkp=%x, hash=%d: PeerAddrType=%d, TransAddrType=%d\n"
      "      FlowClass=%d, FlowKind=%d, MeterId=%d, DSCP=%d\n",
      msg, fkp, hash, fkp->PeerAddrType, fkp->TransAddrType,
      fkp->FlowClass, fkp->FlowKind,fkp-> MeterId, fkp->DSCodePoint);
   printf("   Low: ");
   for (j = 0; j != sizeof(struct key); ++j)
      printf(" %02x", ((Bit8 *)(&fkp->Low))[j]);
   printf("\n  High: ");
   for (j = 0; j != sizeof(struct key); ++j)
      printf(" %02x", ((Bit8 *)(&fkp->High))[j]);
   printf("\n");
   }
#endif /* STREAM_CHAIN_DEBUG */

#if SCC_CHECK || STREAM_CHAIN_DEBUG
int check_stream_chains(char *msg)
{  /* Consistency check for a stream queue */
   struct stream *fsp,*rsp, *l_fsp, *l_rsp;
   char buf[100];
   int k, nas, rv;

   nas = t_active_streams;
   fsp = t_sq;  rsp = t_sq_tail;
   l_fsp = l_rsp = NULL;

   for (rv = 1, k = 0; ; ++k) {  /* Walk forward and reverse chains */
      if (fsp == NULL || rsp == NULL) {  /* Both NULL when done */
         if (fsp != NULL || fsp != rsp) {
	    log_msg(LOG_WARNING, 0,
               "s_c_c(%s): fsp=%d != rsp=%x\n", msg, stream_nbr(fsp), rsp);
 quack(419);
            rv = 0;
            }
         break;
         }
      if (k > nas+10) {
         log_msg(LOG_WARNING, 0,
            "s_c_c(%s): loop in chain, k=%d, nas=%d\n", msg, k, nas);
         fflush(stdout);
 quack(420);
         rv = 0;
         break;
         }
      l_fsp = fsp;  l_rsp = rsp;  /* For gdb debugging */
      fsp = fsp->next_sp;  rsp = rsp->prev_sp;
      }

   if (k != nas) {
      log_msg(LOG_WARNING, 0,
         "s_c_c(%s): k=%u != nas=%u\n", msg, k, nas);
 quack(421);
#if 0  //--?
      for (k=0, fsp = t_sq; fsp != NULL; fsp = fsp->next_sp) {
         printf(" %d:%x", k++, fsp);
         }
#endif  //--?
      printf("\n");
      rv = 0;
      }
   return rv;
   }

int check_stream_hash_chains(char *msg)
{
   int x;
   struct stream **sha;
   struct stream *sp, *lsp, *isp;

   for (x = 0; x != sfhashmod; ++x) {
      sha = &sfht[x];
         /* Address of stream hash table entry */
      if ((sp = *sha) == NULL)
         lsp = NULL;  /* Empty hash chain */
      else {
         isp = sp;
         for (;;) {
	    if (sp->next_hc == NULL) {
               printf("check_stream_h_c(%s): x=%d, isp=%d, lsp=%d, sp=%d\n",
                  msg, x, stream_nbr(isp), stream_nbr(lsp), stream_nbr(sp));
               return 0;
	       }
            lsp = sp;
            if ((sp = lsp->next_hc) == (struct stream *)sha) {
               sp = NULL;  break;
               /* Loop stops with lsp -> last stream in chain */
               }
            }
         }
      }
   return 1;
   }
#endif /* SCC_CHECK || STREAM_CHAIN_DEBUG */

void free_flow(struct flow *q)  /* Free memory allocated for a flow */
   /* Caution: must call mark_idle() before calling free_flow() <<< */
{
   struct distribution *d, *nd;
   struct pktdata *p, *np;
   struct stream_data *sdp;

   if (q->stdata != NULL) {
      sdp = q->stdata;

      q->stdata = NULL;  /* Don't let it point into stream_data free list! */
      unmatch_flow_streams(q);  /* May update distribs! */
/*  printf("@@@ free_flow(): flow=%d @@@\n", flow_nbr(q));  /**/
      free_st_data(sdp);
      }

   for (d = q->distrib_list; d != NULL; d = nd) {  
      nd = d->next; 
#if MULTI
      mutex_lock("ffl", &d->dyn_dst_mutex);  /* Make sure no-one's using it! */
      destroy_mutex(&d->dyn_dst_mutex, "dyn_dst_mutex");
#endif
      free_dist(d);
      }
#if OAM
   if (q->fk.OAMIdent != 0) {
      release_oam_data(q);
      }
#endif

   /* q->FlowRuleSet = 0;  /* Now it doesn't look like part of an rs chain! */
   /* Don't need this, it's done in mark_idle().  Nevil, 15 May 99 */
   }

#if NEW_ATR
# if STR_ANAL
static int complained_about_hlists = 0;
struct host_list *get_hlist(void)  /* Get an unused host_list */
{
   struct host_list *hlp;
   mutex_lock("Ghl", &hlist_fl_mutex);
   if (free_hlists != NULL) {
      hlp = free_hlists;  free_hlists = free_hlists->next;
      --n_free_hlists;
      mutex_unlock("Ghl", &hlist_fl_mutex);
      return hlp;
      }
   else {
      mutex_unlock("Ghl", &hlist_fl_mutex);
      if (!complained_about_hlists) {
         log_msg(LOG_WARNING, 0,
            "Not enough host_list records, increase -H <<<");
         complained_about_hlists = 1;
         }
      return NULL;
      }
   }

void free_hlist(struct host_list *hlp)  /* Return host_list to free list */
{
   mutex_lock("Fhl", &hlist_fl_mutex);
   hlp->next = free_hlists;  free_hlists = hlp;
   ++n_free_hlists;
   complained_about_hlists = 0;
   mutex_unlock("Fhl", &hlist_fl_mutex);
   }
# endif  /* STR_ANAL */

# if OAM
static int complained_about_oamrecs = 0;
struct OAMrec *get_oamrec(void)  /* Get an unused OAMrec */
{
   struct OAMrec *o;
   mutex_lock("Goa", &oamrec_fl_mutex);
   if (free_oamrecs != NULL) {
      o = free_oamrecs;  free_oamrecs = free_oamrecs->next;
      --n_free_oamrecs;
      mutex_unlock("Goa", &oamrec_fl_mutex);
      return o;
      }
   else {
      mutex_unlock("Goa", &oamrec_fl_mutex);
      if (!complained_about_oamrecs) {
         log_msg(LOG_WARNING, 0, "Not enough OAM records, increase -O <<<");
         complained_about_oamrecs = 1;
         }
      return NULL;
      }
   }

void free_oamrec(struct OAMrec *o)  /* Return OAMrec to free list */
{
   mutex_lock("Foa", &oamrec_fl_mutex);
   o->next = free_oamrecs;  free_oamrecs = o;
   ++n_free_oamrecs;
   complained_about_oamrecs = 0;
   mutex_unlock("Foa", &oamrec_fl_mutex);
   }

void release_oam_data(struct flow *q)
{
   struct flow *fp, *lfp;
   struct OAMrec *oap, *nap;

   if (q->stdata == NULL) {
      log_msg(LOG_ERR, 0, "release_oam_data(), flow %d", flow_nbr(q));
      return;
      }
   mutex_lock("rod", &oam_q_mutex);
   if ((lfp = oaf[q->fk.OAMIdent-1]) == q) {  /* Delink from head */
      oaf[q->fk.OAMIdent-1] = q->stdata->OAM_next_flow;
      }
   else {
      fp = lfp->stdata->OAM_next_flow;
      while (fp != q) {  /* Link before this flow */
         lfp = fp;  fp = fp->stdata->OAM_next_flow;
         }
      lfp->stdata->OAM_next_flow = q->stdata->OAM_next_flow;  /* Delink it */
      }
   mutex_unlock("rod", &oam_q_mutex);

   for (oap = q->stdata->OAM_head; oap != NULL; oap = nap) {
      nap = oap->next;
      free_oamrec(oap);  /* Delink flow's OAM records */
      }

#if MULTI
   destroy_mutex(&q->stdata->OAM_mutex, "OAM_mutex");
#endif
   }
# endif  /* OAM */

static int complained_about_distribs = 0;
struct distribution *get_dist(void)  /* Get an unused distribution */
{
   struct distribution *d;
   mutex_lock("Gds", &dist_fl_mutex);
   if (free_distribs != NULL) {
      d = free_distribs;  free_distribs = free_distribs->next;
      --n_free_distribs;
      mutex_unlock("Gds", &dist_fl_mutex);
      return d;
      }
   else {
      mutex_unlock("Gds", &dist_fl_mutex);
      if (!complained_about_distribs) {
         log_msg(LOG_WARNING, 0, "Not enough distributions, increase -v <<<");
         complained_about_distribs = 1;
         }
      return NULL;
      }
   }

void free_dist(struct distribution *d)  /* Return distribution to free list */
{
   mutex_lock("Fds", &dist_fl_mutex);
   switch (d->selector) {
   case FTTOBITRATE:
   case FTFROMBITRATE:
   case FTTOPDURATE:
   case FTFROMPDURATE:
      unlink_distrib_from_event(d);
      break;
      }
   d->next = free_distribs;  free_distribs = d;
   ++n_free_distribs;
   complained_about_distribs = 0;
   mutex_unlock("Fds", &dist_fl_mutex);
   }

#if CHECK_EV_QUEUE
void check_free_events()
{
  struct event *ep;  int j;
   if (free_events == NULL) {
     printf("cfe(): free_events == NULL\n");
      quack(701);
      }
   for (ep = free_events, j = 0; ep != NULL; ++j) {
      ep = ep->next;
      if (j > n_free_events) printf(" %x", ep);
      if (j > n_free_events+10) {
	printf("cfe(): looping!\n");
	break;
         }
      }
   if (j != n_free_events) {
      printf("cfe(): %d free events counted, n_free_events=%d\n",
         j, n_free_events);
      quack(702);
      }
   }
#endif

static int complained_about_events = 0;
struct event *get_event(int t)  /* Get an unused event */
{
   struct event *e;

   mutex_lock("Gev", &event_fl_mutex);
   if (free_events != NULL) {
      e = free_events;  free_events = free_events->next;
      e->type = t;  --n_free_events;
      mutex_unlock("Gev", &event_fl_mutex);
      return e;
      }
   else {
      mutex_unlock("Gev", &event_fl_mutex);
      if (!complained_about_events) {
         log_msg(LOG_WARNING, 0, "Not enough events, increase -e <<<");
         complained_about_events = 1;
         }
      return NULL;
      }
   }

void free_event(struct event *e)  /* Return event to free list */
{
   if (e == NULL) {
      log_msg(LOG_WARNING, 0, "Tried to free NULL event");
      return;
      }

   mutex_lock("Fev", &event_fl_mutex);
   if (e->type == FREE_EVENT) {
      mutex_unlock("Fev", &event_fl_mutex);
      log_msg(LOG_WARNING, 0, "Tried to free FREE_EVENT %x", e);
      return;
      }
   e->next = free_events;  free_events = e;
   e->type = FREE_EVENT;  ++n_free_events;
   complained_about_events = 0;
   mutex_unlock("Fev", &event_fl_mutex);
   }

void schedule_event(struct event *ev)
{
   struct event *eqp, *lqp;

   if (ev == NULL)
      log_msg(LOG_WARNING, 0, "Tried to schedule NULL event!");

   switch (ev->type) {
   case DISTRIB_EVENT:
      lqp = &d_eventq;  eqp = d_eventq.next;
      for ( ; eqp != NULL; lqp = eqp, eqp = eqp->next) {
         if (ev->t <= eqp->t) break;  /* Closest to head */
         }
      if (eqp == NULL) d_eventq_tail = ev;
      lqp->next = ev;  ev->next = eqp;
      break;
      }

#if 0
   printf("66 sch_ev(): queue = ");
   for (eqp = eventq.next; eqp != NULL; eqp = eqp->next)
     printf("%0x/%d/%d ", eqp, eqp->type, eqp->t);
   printf("\n");
#endif
   }

void unschedule_event(struct event *ev)
{
   struct event *eqp, *lqp;

   if (ev == NULL) {
      printf("unschedule_event(NULL) !!!\n");
      quack(333);
      }

   if (ev == NULL) return;
   mutex_lock("ue", &dev_q_mutex);
   switch (ev->type) {
   case DISTRIB_EVENT:
      lqp = &d_eventq;  eqp = d_eventq.next;
      for ( ; eqp != NULL; lqp = eqp, eqp = eqp->next) {
         if (eqp == ev) {
            if ((lqp->next = eqp->next) == NULL) d_eventq_tail = NULL;
            mutex_unlock("ue", &dev_q_mutex);
            return;
            }
         }
      break;
      }
   log_msg(LOG_WARNING, 0,
      "Couldn't unschedule type %d event %0x", ev->type, ev);
   mutex_unlock("ue", &dev_q_mutex);
   }

static int complained_about_pktdata = 0;
struct pktdata *get_pktdata(void)  /* Get an unused pktdata */
{
   struct pktdata *pd;
   mutex_lock("Gpd", &pktdata_fl_mutex);
   if (free_pkt_dat != NULL) {
      pd = free_pkt_dat;  free_pkt_dat = pd->next;
      --n_free_pktdata;  ++PktsCreated;
      mutex_unlock("Gpd", &pktdata_fl_mutex);
      return pd;
      }
   else {
      mutex_unlock("Gpd", &pktdata_fl_mutex);
      if (!complained_about_pktdata) {
         log_msg(LOG_WARNING, 0, "Not enough pktdata blocks, increase -a <<<");
         complained_about_pktdata = 1;
         }
      return NULL;
      }
   }

void free_pktdata(struct pktdata *pd)  /* Return pktdata to free list */
{
   mutex_lock("Fpd", &pktdata_fl_mutex);
   pd->next = free_pkt_dat;  free_pkt_dat = pd;
   ++n_free_pktdata;  ++PktsRecovered;
   complained_about_pktdata = 0;
   mutex_unlock("Fpd", &pktdata_fl_mutex);
   }

static int complained_about_pkts = 0;
struct pkt *get_pkt(void)
{
   struct pkt *pf;
   mutex_lock("Gpk", &pkt_fl_mutex);
   if (free_pkts != NULL) {
      pf = free_pkts;  free_pkts = pf->next_pp;
      --n_free_pkts;
      mutex_unlock("Gpk", &pkt_fl_mutex);
#if PKT_DEBUG
      pf->pkt_state = 1;
# if PKT_DEBUG > 1
      printf("g:");  fflush(stdout);
# endif
#endif
      return pf;
      }
   else {
      mutex_unlock("Gpk", &pkt_fl_mutex);
      if (!complained_about_pkts) {
/*         log_msg(LOG_WARNING, 0, "Not enough pkt space, increase -P <<<"); /**/
         log_msg(LOG_WARNING, 0, 
            "Not enough pkt space, increase -P, n_free_pkts=%d <<<",
            n_free_pkts);
         complained_about_pkts = 1;
         }
      return NULL;
      }
   }

void free_pkt(struct pkt *pp)
{
#if PKT_DEBUG
   pp->pkt_state = 0;
# if PKT_DEBUG > 1
   printf("f");  fflush(stdout);
# endif
#endif
   mutex_lock("Fpk", &pkt_fl_mutex);
   pp->next_pp = free_pkts;  free_pkts = pp;
   ++n_free_pkts;
   complained_about_pkts = 0;
   mutex_unlock("Fpk", &pkt_fl_mutex);
   }

void bump_dist(  /* Add point y into distrib d */
   struct distribution *d, double y);
void accum_dist(  /* Add v to point y of distrib d */
   struct distribution *d, double y, Bit32 v);
double inv_scale(struct distribution *d, Bit16 j);  /* limit -> value */
Bit32 scale(struct distribution *d, double y);

Bit32 times_pwr_of_10(Bit16 x, int p);

void compute_rate(struct distribution *d)
{
   struct flow *f;
   counter64 r;
   double delta;
   f = d->fp;
   switch (d->selector) {
   case FTTOBITRATE:
      subtr_64(r, f->UpOctets,f->LastUpOctets);
      assign_64(f->LastUpOctets, f->UpOctets);
      delta = (doublefrom64(r))*8.0;  /* Bits */
      break;
   case FTFROMBITRATE:
      subtr_64(r, f->DownOctets,f->LastDownOctets);
      assign_64(f->LastDownOctets, f->DownOctets);
      delta = (doublefrom64(r))*8.0;  /* Bits */
      break;
   case FTTOPDURATE:
      subtr_64(r, f->UpPDUs,f->LastUpPDUs);
      assign_64(f->LastUpPDUs, f->UpPDUs);
      delta = doublefrom64(r);  /* Packets */
      break;
   case FTFROMPDURATE:
      subtr_64(r, f->DownPDUs,f->LastDownPDUs);
      assign_64(f->LastDownPDUs, f->DownPDUs);
      delta = doublefrom64(r);  /* Packets */
      break;
      }
   bump_dist(d, delta/((double)d->RateInterval));
   }

void check_events(Bit32 now)
{
   struct event *ep;
   struct distribution *d;
   struct flow *fp;
   struct stream_data *sdp;
   int re_queue, n_term;

#if EVNT_CHK_TRACE
   printf("check_ev(%d) n_sc=%d, sc_ev=%d: ", now, n_sc_events, sc_ev);
   for (ep = d_eventq.next; ep != NULL; ep = ep->next)
      printf(" %d.%x=%d", ep->type, flow_nbr(ep->det.dp->fp), ep->t);
   printf("\n");
#endif
   for (;;) {  /* Distribution event queue */
      mutex_lock("cde", &event_q_mutex);
      ep = d_eventq.next;
      if (ep == NULL || ep->t > now) {
         mutex_unlock("cde", &event_q_mutex);
         break;
         }
      d_eventq.next = ep->next;  /* Remove from distrib event queue */
      mutex_unlock("cde", &event_q_mutex);

      re_queue = 0;
      if ((d = ep->det.dp) != NULL) {
         fp = ep->det.fp;
/*  printf("66: DISTRIB_EVENT flow=%d\n",flow_nbr(fp));  /**/
         mutex_lock("e0", &dev_q_mutex);
         do {
            compute_rate(d);
            d = d->nrate;
            } while (d != NULL);
         mutex_unlock("e0", &dev_q_mutex);
         re_queue = 1;
         ep->t += ep->interval;
         }

      if (re_queue) {
         mutex_lock("cdr", &event_q_mutex);
         schedule_event(ep);      
         mutex_unlock("cdr", &event_q_mutex);
         }
      else free_event(ep);
      }

   check_idle_streams();
      /* Only one stream queue to check */
      /* Don't care about TCP vs other (!) */
   }

struct event *setup_event(int type, void *p, int interval)
{
   struct event *eqp;

   if ((eqp = get_event(type)) != NULL) {
      eqp->det.dp = p;
      eqp->interval = interval;
      eqp->t = uptime_s() + eqp->interval;
      mutex_lock("sue", &event_q_mutex);
      schedule_event(eqp);
      mutex_unlock("sue", &event_q_mutex);
/* printf("66 setup_e(3): type %d created, flow=%d, event=%0x, now=%d, t=%d\n",
   type, flow_nbr(eqp->det.fp), eqp, eqp->t - eqp->interval, eqp->t); /**/
      }
   return eqp;
   }

void link_distrib_to_event(struct distribution *d)
{
   struct event *eqp;
   struct distribution *dq;
   mutex_lock("elf", &dev_q_mutex);
   for (eqp = d_eventq.next; eqp != NULL; eqp = eqp->next) { 
      if (eqp->type == DISTRIB_EVENT && 
         eqp->interval == d->RateInterval) break;
      }
   mutex_unlock("elf", &dev_q_mutex);
   if (eqp == NULL)  /* Make new event for this interval */
      eqp = setup_event(DISTRIB_EVENT, NULL, d->RateInterval);

   mutex_lock("ell", &dev_q_mutex);
   if (eqp != NULL) {  /* Add distrib to head of event chain */
      /* Was eqp == NULL, fixed Sat 3 Jun 00 */
      d->nrate = eqp->det.dp;
      eqp->det.dp = d;
      }
   mutex_unlock("ell", &dev_q_mutex);
   }

#if DIST_DBG
void print_dist_chain(struct event *ep)
{
   struct distribution *dq;
   for (dq = ep->dp; dq != NULL; dq = dq->nrate)
      printf("%d@%lu ", dq->fp->fk.PeerAddrType, dq);
   printf("\n");
   }
#endif

void unlink_distrib_from_event(struct distribution *d)
{
   struct event *eqp;
   struct distribution *dq, *lq;
   mutex_lock("eu", &dev_q_mutex);
   for (eqp = d_eventq.next; eqp != NULL; eqp = eqp->next) { 
      if (eqp->type == DISTRIB_EVENT &&
         eqp->interval == d->RateInterval) break;
      }
#if DIST_DBG
   printf("unlink_distrib(%lu): ", d);  /* $$$ */
   print_dist_chain(eqp);
#endif
   if (eqp == NULL) {
      mutex_unlock("eu", &dev_q_mutex);
      scpos(0,scr_lrow);
      log_msg(LOG_ERR, 0, 
         "No event for %ds rate distribution <<<", d->RateInterval);
      return;
      }
   for (lq = NULL, dq = eqp->det.dp; dq != d; lq = dq, dq = dq->nrate) {
      if (dq == NULL) {
         mutex_unlock("eu", &dev_q_mutex);
         scpos(0,scr_lrow);
         log_msg(LOG_ERR, 0, "Couldn't find rate distribution <<<");
         return;
         }
      }
   if (lq == NULL) eqp->det.dp = d->nrate;
   else lq->nrate = d->nrate;
   mutex_unlock("eu", &dev_q_mutex);

   if (eqp->det.dp == NULL) {  /* No distribs left in chain */
      mutex_lock("ude", &event_q_mutex);
      unschedule_event(eqp);
      mutex_unlock("ude", &event_q_mutex);
      free_event(eqp);
      }

#if DIST_DBG
   printf("   ");  /* $$$ */
   print_dist_chain(eqp);
#endif
   }

static int complained_about_str_blocks = 0;
struct stream_data *get_st_data(void)  /* Get unused st_data */
{
   struct stream_data *sh;
   mutex_lock("G", &stdata_fl_mutex);
   if (free_str_blocks != NULL) {
      sh = free_str_blocks;  free_str_blocks = free_str_blocks->next;
      --n_free_str_blocks;
      mutex_unlock("G", &stdata_fl_mutex);
      memset(sh, 0, sizeof(struct stream_data));
      return sh;
      }
   else {
      mutex_unlock("G", &stdata_fl_mutex);
      if (!complained_about_str_blocks) {
         log_msg(LOG_WARNING, 0,
            "Not enough stream data blocks, increase -b <<<");
         complained_about_str_blocks = 1;
         }
      return NULL;
      }
   }

void free_st_data(struct stream_data *sh)  /* Return st_data to free list */
{
   mutex_lock("F", &stdata_fl_mutex);
   sh->next = free_str_blocks;  free_str_blocks = sh;
   ++n_free_str_blocks;
   complained_about_str_blocks = 0;
   mutex_unlock("F", &stdata_fl_mutex);
   }

static int complained_about_streams = 0;
struct stream *get_stream(void)
{
   struct stream *sf;
   mutex_lock("G", &stream_fl_mutex);
   if (free_streams != NULL) {
      sf = free_streams;  free_streams = free_streams->next_sp;
      if (free_streams == NULL) free_stream_tail = NULL;
      sf->state = SS_GOT;
      mutex_unlock("G", &stream_fl_mutex);
      --n_free_streams;  ++StreamsCreated;
      return sf;
      }
   else {
      mutex_unlock("G", &stream_fl_mutex);
      ++n_lost_streams;
      if (!complained_about_streams) {
         log_msg(LOG_WARNING, 0, "Not enough stream space, increase -t <<<");
         complained_about_streams = 1;
         }
      return NULL;
      }
   }

void free_stream(struct stream *sp)
{
   mutex_lock("F", &stream_fl_mutex);
   if (free_stream_tail == NULL) free_streams = sp;
   else free_stream_tail->next_sp = sp;
   free_stream_tail = sp;
   sp->next_sp = NULL;  sp->state = SS_FREE;
   mutex_unlock("F", &stream_fl_mutex);
   ++n_free_streams;  ++StreamsRecovered;
   complained_about_streams = 0;
   }
#endif  /* NEW_ATR */

Bit8 v6addr[IP6_ADDR_LEN*3];
#if V6 || USER_INFO
char *sfmt_v6addr(char *d, unsigned char *a)
{
   char buf[50];
   int j, k, st,len, stx,lenx;
   unsigned int v, a2[IP6_ADDR_LEN/2], av;
   st = len = lenx = 0;
   for (k = j = 0; j != IP6_ADDR_LEN; j += 2) {
      v = (a[j] << 8) | a[j+1];
      a2[k++] = v;          /* Build array of two-byte pairs */
      if (v == 0) ++len;
      else {
         if (len > lenx) {  /* Find longest run of zero pairs */
            stx = st;  lenx = len;
	    }
         st = k;  len = 0;
         }
      }
   if (len > lenx) {
      stx = st;  lenx = len;
      }
   if (lenx != 0 && stx == 0) {  /* Longest run at left */
      d = strmov(d, ":");  j = lenx;
      }
   else {
      sprintf(buf, "%04X", a2[0]);
      d = strmov(d,buf);  j = 1;
      }
   for (; j < IP6_ADDR_LEN/2; ) {
      if (lenx != 0 && j == stx) {
         d = strmov(d,":");  j += lenx;
         }
      else {
         sprintf(buf, ":%04X", a2[j]);
         d = strmov(d, buf);  ++j;
         }
      }
   if (j == stx+lenx) d = strmov(d, ":");  /* Longest run at right */
   *d = '\0';
   return d;
   }

void print_v6addr(Bit8 *a)
{
   char buf[IP6_ADDR_LEN*4];
   sfmt_v6addr(buf, a);  
   printf("%s\n", buf);
   }

#endif

#ifdef DEBUG
FILE *logfl;
int logf_nbr = 1;

void open_log()
{
   char fn[30];
   sprintf(fn,"flows.%03d",logf_nbr++);
   logfl = fopen(fn, "w");
   }

void paddr(Bit32 a)
{
   a = ntohl(a);
   fprintf(logfl,"%u.%u.%u.%u",
      a >> 24, (a >> 16) & 0xFF, (a >> 8) & 0xFF, a & 0xFF);
   }

void daddr(Bit32 a)
{
   a = ntohl(a);
   printf("%u.%u.%u.%u",
      a >> 24, (a >> 16) & 0xFF, (a >> 8) & 0xFF, a & 0xFF);
   }
  
void fdaddr(FILE *f, Bit32 a)
{
   a = ntohl(a);
   fprintf(f, "%u.%u.%u.%u",
      a >> 24, (a >> 16) & 0xFF, (a >> 8) & 0xFF, a & 0xFF);
   }

void dport(Bit16 a)
{
   a = ntohs(a);
   printf("%u", a);
   }

void fdport(FILE *f, Bit16 a)
{
   a = ntohs(a);
   fprintf(f, "%u",a);
   }

void pkey(char *msg, struct flow *f)
{
   if (!logfl) open_log();
   fprintf(logfl,"%s: %lu  %u\n   {", msg,f, f->fk.PeerAddrType);
#if QBYTE_PEER
   paddr(f->fk.Low.PeerAddress.qbyte); fprintf(logfl," & ");
   paddr(Masks[f->fk.Low.PeerMaskVal].qbyte); 
#else
   paddr(f->fk.Low.PeerAddress.peer32[0]); fprintf(logfl," & ");
   paddr(Masks[f->fk.Low.PeerMaskVal].peer32[0]); 
#endif
   fprintf(logfl,"} -> {");
#if QBYTE_PEER
   paddr(f->fk.High.PeerAddress.qbyte); fprintf(logfl," & ");
   paddr(Masks[f->fk.High.PeerMaskVal].qbyte);
#else
   paddr(f->fk.High.PeerAddress.peer32[0]); fprintf(logfl," & ");
   paddr(Masks[f->fk.High.PeerMaskVal].peer32[0]);
#endif
   fprintf(logfl,"}\n");
   }
#endif

#define DRT_SIZE  3  /* Default rule set: 3 rules */
#ifdef OLD_CT
#define DCT_SIZE  1  /*   1 count */
#endif
struct rule default_rule_table[DRT_SIZE] = {
   /* Value,    Selector, Mask, JumpIndex, Action */
   {
# if CLNS
      {AT_DUMMY,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
# elif V6 || FULL_IPX
      {AT_DUMMY,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
# else  /* v4 */
      {AT_DUMMY,0,0,0,0,0},
# endif
      FTLOWPEERTYPE, 1, 0, RA_IGNORE},
      /* SourcePeerType & 255 = dummy: Ignore, 0; # 1 */
   {
# if CLNS
      {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
# elif V6 || FULL_IPX
      {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
# else  /* v4 */
      {0,0,0,0,0,0},
# endif
      RS_NULL, 0, 3, RA_GOTOACT},
      /* Null & 0 = 0: GotoAct, Next; # 2 */
   {
# if CLNS
      {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
# elif V6 || FULL_IPX
      {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
# else  /* v4 */
      {0,0,0,0,0,0},
# endif
      FTLOWPEERTYPE, 1, 1, RA_COUNTPKT}
      /* SourcePeerType & 255 = 0: CountPkt, 1; # 3 */
   };

Bit32 p2primes[] = {  /* Primes just below powers of two */
   7, 13, 31, 61, 127, 251, 509, 1021,      /*  0 .. 7  1024 */
   2039, 4093, 8191, 16381, 32749, 65521,   /*  8 .. 13  65536 */
   131071, 262139, 524287, 1048573};        /* 14 .. 17 1048576 */

#if 0
 int primes[22] = {  /* Two for Trans bytes + enough for NSAP bytes */
   211, 251,
   113, 139, 173, 227, 179, 193, 181, 131,  67,  73, 
    83,  97, 103, 127, 149,  23,  53,  89,  59,  47};
#endif

Bit32 primes[22] = {  /* Two for Trans bytes + enough for NSAP bytes */
   97241, 98467, 97171, 95881, 96953,
   95027, 93559, 95089, 96289, 97039,
   97301, 99761, 95803, 94873, 95131,
   96457, 91867, 94007, 97501, 99961,
   96337, 95383};

#define  prime_0  95027
#define  prime_1  93559
#define  prime_2  95089
#define  prime_3  96289
#define  prime_4  97039
#define  prime_5  97301
#define  prime_6  99761
#define  prime_7  95803
#define  prime_8  94873
#define  prime_9  95131
#define prime_10  96457
#define prime_11  91867
#define prime_12  94007
#define prime_13  97501
#define prime_14  99961
#define prime_15  96337
#define prime_16  95383
#define prime_17  97241
#define prime_18  98467
#define prime_19  97171
#define prime_20  95881
#define prime_21  96953

Bit32 compute_rule_hash(int attrib, union address *rvp)
{  /* Must compute these **exactly** as for hashed compare in match */
#if !QBYTE_PEER
   Bit32 hash;
   int j;
#endif

   switch (attrib) {
   case FTLOWPEERADDRESS:
   case FTHIPEERADDRESS:
#if QBYTE_PEER
      return rvp->qbyte*prime_0;
#else
      for (j = hash = 0;  j !=(PEER_ADDR_LEN+3)/4;  ++j) hash += 
         rvp->peer32[j]*primes[j];
      return hash;
#endif

   case FTLOWPEERTYPE:
   case FTHIPEERTYPE:
      return rvp->byte*prime_1;
   case FTLOWTRANSADDRESS:
   case FTHITRANSADDRESS:
      return rvp->trans*prime_2;
   case FTLOWTRANSTYPE:
   case FTHITRANSTYPE:
      return rvp->byte*prime_3;
   case FTLOWADJACENTADDRESS:
   case FTHIADJACENTADDRESS:
      return rvp->ma.ms4*prime_4 + rvp->ma.ls2*prime_5;
   case FTSOURCECLASS:
   case FTDESTCLASS:
   case FTFLOWCLASS:
      return rvp->byte*prime_9;
   case FTSOURCEKIND:
   case FTDESTKIND:
   case FTFLOWKIND:
      return rvp->byte*prime_10;
   case FTPAIRINTERFACE:
   case FTLOWINTERFACE:
   case FTHIINTERFACE:
   case FTOBSINTERFACE:
      return rvp->trans*prime_7;
   case FTLOWADJACENTTYPE:
   case FTHIADJACENTTYPE:
      return rvp->byte*prime_6;
   case FTPRIORITY:
      return rvp->byte*prime_16;
   case FTDSCODEPOINT:
      return rvp->byte*prime_8;
   case FTECNCODEBITS:
      return rvp->byte*prime_14;
   case FTVLANID:
      return rvp->byte*prime_15;
# if USER_INFO
   case FTHISUBSCRIBERID:  /* UserCoS */
      return rvp->byte*prime_17;
   case FTSESSIONID:  /* UserInfo */
      return rvp->byte*prime_18;
# endif

#if NF_ASN_ATT
   case FTLOWROUTEASN:
   case FTHIROUTEASN:
      return rvp->byte*prime_11;
   case FTLOWROUTEPREFIX:
   case FTHIROUTEPREFIX:
      return rvp->byte*prime_12;
#endif
#if NF_OTHER_ATT
   case FTMETERID:
      return rvp->byte*prime_13;
#endif
      }
   }

int optimise_rule_table(int rs)
{
   struct rtinf_rec *rip = &ri[rs-1];
   Bit16 crs, j, k, g, h, x, l,
      nht;  /* Nbr of rule_hash_tbls */
   Bit32 ths,  /* Sum of rule_hash_tbl sizes */
      rhss;  /* Total rule_hash sequence size */
   struct rule *crt, *rp, *hp;
   Bit16 *rs_rule_hash;
   struct rule_hash_tbl *c_rh;
   Bit32 hash;
   char msg[60];

   crt = rip->ri_rule_table;
   crs = rip->ri_Size;

   crt[0].push_rule = 1;  /* Mark 'could test attrib' rules */
   for (j = 0; j != crs; ++j) {
      rp = &crt[j];
      if (rp->RuleAction == RA_PUSHPKTTO || rp->RuleAction == RA_GOTO ||
            rp->RuleAction == RA_PUSHTO || rp->RuleAction == RA_POPTO ||
            rp->RuleAction == RA_GOSUB || rp->RuleAction == RA_ASSIGN) {
         if (rp->RuleJumpIndex != 0)
            crt[rp->RuleJumpIndex-1].push_rule = 1;
         }
      }

   ri[rs-1].oa_count = 0;
   for (j = 0; j != crs; ++j) {
      rp = &crt[j];
      if (rp->RuleAction == RA_PUSHPKTTO ||  /* SAVE rules */
            rp->RuleAction == RA_PUSHPKTTOACT) {
         if (atr_oa[rp->RuleSelector]) ++ri[rs-1].oa_count;
         }
      if (rp->push_rule &&  /* Test can be performed */
            rp->RuleSelector != RS_NULL) {  /* Attribute is tested */
         if (atr_oa[rp->RuleSelector]) ++ri[rs-1].oa_count;
         crt[j+1].push_rule = 1;  /* If it fails, next rule will test */
         }
      }

#if H_TEST
   scpos(0,scr_lrow);  printf("\nRT: crs=%u, RULE_ADDR_LEN=%d\n",
      crs,RULE_ADDR_LEN);
#endif
   hp = &crt[0];  hp->hash_tbl_index = 1;
   for (j = 1;  j != crs;  ++j) {  /* Mark match break points */
      rp = &crt[j];

      rp->push_rule = (  /* For build_search_key */
         (rp->RuleAction != RA_PUSHPKTTO) &&
         (rp->RuleAction != RA_PUSHPKTTOACT) &&
         (rp->RuleAction != RA_COUNTPKT) );

      rp->hash_tbl_index = rp->RuleSelector != hp->RuleSelector ||
         rp->RuleMaskVal != hp->RuleMaskVal;
      hp = rp;
      }
   for (j = 0;  j != crs;  ++j) {  /* Mark goto break points */
      rp = &crt[j];  rp->hash_link = 0;
      k = rp->RuleAction;
      if (k == RA_PUSHTO  /* OK to gotoact in middle of compare group */
            || k == RA_PUSHPKTTO || k == RA_GOTO
            || k == RA_GOSUB     || k == RA_ASSIGN
            || k == RA_POPTO
            ) {
         hp = &crt[rp->RuleJumpIndex-1];
         hp->hash_tbl_index = 2; 
         }
      }
#if H_TEST
   scpos(0,scr_lrow);  printf("rule  index  link  MARKED\n");
   for (j = 0;  j != crs;  ++j) {
      rp = &crt[j];
      printf("%3d %5d %5d  %d  ",
         j,rp->hash_tbl_index,rp->hash_link, rp->RuleSelector);
#if QBYTE_PEER
      daddr(Masks[rp->RuleMaskVal].qbyte);
#else
      daddr(Masks[rp->RuleMaskVal].peer32[0]);
#endif
      printf("\n");
      }
#endif
   nht = ths = 0;
   hp = &crt[0];  k = 1;
   for (j = 1;  j != crs;  ++j) {  /* Find compare groups */
      rp = &crt[j];
      if ((g = rp->hash_tbl_index) == 0) ++k;  /* Same group */
      if (g != 0 || j == crs-1) {  /* End of a group */
         if (k >= MNCGRPSZ && hp->RuleSelector != RS_NULL) {
            for (h = 0; h != sizeof(p2primes)/sizeof(Bit32)-1 &&
               p2primes[h] <= (Bit32)k; ++h) ;  /* h = next prime above k */
            if (p2primes[h] > MXRHTSZ) h = MXRHTSZ;
            else h = p2primes[h];
            hp->hash_tbl_index = k;  /* Remember group and */
            hp->hash_link = h;       /*   hash table sizes */
            ++nht;  ths += h;
            }
         else hp->hash_tbl_index = 0;
         hp = rp;  k = 1;
         }
      rp->hash_tbl_index = 0;  /* Last row can't start a group */
      }
   rhss = 1 + nht*2 + ths;  /* rule_hash set size (Bit16s) */
   if (rhss > 0xFFFF) {
      log_msg(LOG_ERR, 0, msg,"Ruleset '%s': rhss = %lu (> 65535)", 
         rip->ri_Name, rhss);
      return 0;  /* Fail */
      }
#if H_TEST
   printf("rule  index  link  GROUPED\n");
   for (j = 0;  j != crs;  ++j) {
      rp = &crt[j];
      printf("%3d %5d %5d\n", j, rp->hash_tbl_index,rp->hash_link);
      }
#endif

   if (!alloc_rulespace(rs, 0, rhss))  /* Alloc rule hash space */
      return 0;  /* No room for rule hash table! */

   rs_rule_hash = rip->ri_rule_hash;
   rs_rule_hash[0] = 0;  k = 1;  rp = &crt[j = 0];
   for ( ; nht != 0; --nht, k += 2+h) {
      c_rh = (struct rule_hash_tbl *)&rs_rule_hash[k];
      while (rp->hash_tbl_index == 0) {
         ++rp; ++j;
         }
      g = rp->hash_tbl_index;  /* Group size */
      h = rp->hash_link;  /* Hash table size */
      rp->hash_tbl_index = k;  /* Index of hash_tbl in rule_hash */
      rp->hash_link = 0;
      c_rh->group_last = j+g;
      c_rh->hashmod = h;
      for (x = 0; x != h; ++x) c_rh->hash_ent[x] = 0;
      for ( ; g != 0; --g) {  /* Build rule hash tables */
         hash = compute_rule_hash(rp->RuleSelector, 
            &rp->RuleMatchedValue) % h;
         if ((l = c_rh->hash_ent[hash]) == 0)
            c_rh->hash_ent[hash] = j+1;  /* First rule with this hash */
         else {
            do {
               hp = &crt[l-1];  l = hp->hash_link;
               } while (l != 0);
            hp->hash_link = j+1;  /* Add to end of hash chain */
            }
         ++rp; ++j;
         }
      }
#if H_TEST
   printf("Rule Hash Tables ..\n\nrule index  n_rules n_hash n_zero\n\n");
   for (j = 0;  j != crs;  ++j) {
      rp = &crt[j];
      if ((k = rp->hash_tbl_index) == 0) continue;
      c_rh = (struct rule_hash_tbl *)&rs_rule_hash[k];
      h = c_rh->hashmod+1;
      for (g = 0, hash = 0; hash != h; ++hash)
         if (c_rh->hash_ent[hash] == 0) ++g;
      printf("%5d%5d  %5d%5d%5d\n     ",
         j+1, k, c_rh->group_last-j+1, h, g);
      printf("\n");
      }
#endif
   return 1;
   }

void empty_hash_chain(int rs)  /* Unlink flows for ruleset rs */
{
   Bit32 k;
   struct flow **sf;
   struct flow *fp, *lfp;
   struct flow *chp, *ctp;
   struct stream_data *sdp;

   /* Could walk the rs chain, rather than linear searching the
      whole table!  But we don't shut down rule sets very often! */

   for (k = 0; k != fthashmod; ++k) {
      sf = &flow_ht[k];  /* Address of hash table entry */
      if ((fp = *sf) != NULL) {
         chp = ctp = NULL;  /* New chain for remaining active flows */
         do {
            lfp = fp;  fp = lfp->ht_next;
            if (lfp->FlowRuleSet == rs) {
               mark_orphan(lfp);
               if (lfp->stdata != NULL) {
                  sdp = lfp->stdata;
                  lfp->stdata = NULL;  /* Don't point into stdata free list! */
                  }
               }
            else {
               if (chp == NULL) chp = ctp = lfp;
               else {
                  ctp->ht_next = lfp;
                  ctp = lfp;
                  }
               }
            } while (fp != (struct flow *)sf);
         if (chp != NULL) {
            ctp->ht_next = fp;
            *sf = chp;
            }
         else {  /* No remaining active flows */
            *sf = NULL;
            --n_hash_ents;
            }
         }
      }
   }

#if PME_TRACE
#define MXTRACE      250
#define MXTRACE_MSGS   5  /* Per rule set */

static Bit16 ruletrace[MXTRACE], rtx;
static int tr_msgs_left[MXRTBLS];
#endif

void open_rule_set(int n, int open_rs)
{
   int lrx,rrx, k, p;
   Bit16 rsx, c_rsz;
   struct rule *rp, *c_rt;

   if (n == 0) return;  /* 0 = no rule set to run */
   if (ri[n-1].ri_Status != RS_ACTIVE) return;

   if (open_rs) {
      if (++ri[n-1].ri_running != 1)  /* Already running */
         return;
      else {
#if MULTI
         for (k = 0; k != N_PME_THREADS; ++k) 
            ntm_semaphore_down("ors", &r_rts_sem);
#endif
         if (running_rts == 0) {  /* Starting first ruleset */
            running_rts = n;  n_running_rts = 1;
            }
         else {
            for (rrx = running_rts; ; ) {
               if (ri[rrx-1].next_running_rt == 0) break;
               rrx = ri[rrx-1].next_running_rt;
               }
            ri[rrx-1].next_running_rt = n;  ++n_running_rts;
            }
         ri[n-1].next_running_rt = 0;
#if PME_TRACE
         tr_msgs_left[n-1] = MXTRACE_MSGS;
#endif
         }

      unmatch_all_streams();  /* Any stream could match new ruleset */
      }
   else {  /* Close ruleset */
      if (--ri[n-1].ri_running != 0)  /* Still running */
         return;

#if MULTI
      for (k = 0; k != N_PME_THREADS; ++k) 
         ntm_semaphore_down("ors", &r_rts_sem);
#endif
      --n_running_rts;
      for (lrx = 0, rrx = running_rts; rrx != n; ) {
         lrx = rrx;  rrx = ri[rrx-1].next_running_rt;
         }
      if (lrx == 0) running_rts = ri[rrx-1].next_running_rt;
      else ri[lrx-1].next_running_rt = ri[rrx-1].next_running_rt;
      ri[n-1].next_running_rt = 0;

      empty_hash_chain(n);
      unmatch_ruleset_streams(n);  /* May update distribs! */
      }

   c_rt = ri[n-1].ri_rule_table;  c_rsz = ri[n-1].ri_Size;
   for (rsx = 0; rsx != c_rsz; ++rsx) {
      rp = &c_rt[rsx];
      if (rp->RuleSelector == FTLOWPEERTYPE ||
            rp->RuleSelector == FTHIPEERTYPE) {
         p = rp->RuleMatchedValue.rule[0];
         if (rp->RuleAction == RA_PUSHPKTTO ||  /* All peer types */
               rp->RuleAction == RA_PUSHPKTTOACT ||
               rp->RuleAction == RA_COUNTPKT)
            for (k = 0; k != sizeof(proto_reqd)/sizeof(int); ++k) {
               if (open_rs) ++proto_reqd[k];  
               else --proto_reqd[k];
               }
         else if (p <= MX_PROTOCOLS && *proto_names[p] != '\0') {
 	    /* Was & above!  Sebastian Zander, 6 Jan 2000 */
            if (open_rs) ++proto_reqd[p];  /* Specific peer type */
            else --proto_reqd[p];
            }
         }
      if ((rp->RuleSelector == FTLOWADJACENTADDRESS ||
            rp->RuleSelector == FTHIADJACENTADDRESS) ||
            ((rp->RuleAction == RA_ASSIGN ||
            rp->RuleAction == RA_ASSIGNACT) &&
            (rp->RuleMatchedValue.rule[0] == FTLOWADJACENTADDRESS ||
            rp->RuleMatchedValue.rule[0] == FTHIADJACENTADDRESS)) )
         if (open_rs) ++adj_reqd;
         else --adj_reqd;
#if NF_OCX_BGP
      if (rp->RuleSelector == FTLOWROUTEASN ||
            rp->RuleSelector == FTHIROUTEASN)
         if (open_rs) ++asn_lookup;
         else --asn_lookup;
#endif
      }
   if (adj_reqd) {
      for (k = 0; k != sizeof(proto_reqd)/sizeof(int); ++k)
         if (proto_reqd[k]) {
#if MULTI
            for (k = 0; k != N_PME_THREADS; ++k) 
               ntm_semaphore_up("ors", &r_rts_sem);
#endif
            return;  /* At least one protocol specified */
            }
      for (k = 0; k != sizeof(proto_reqd)/sizeof(int); ++k) {
         if (open_rs) ++proto_reqd[k];  /* All peer types */
         else --proto_reqd[k];
         }
      }
#if MULTI
   for (k = 0; k != N_PME_THREADS; ++k) 
      ntm_semaphore_up("ors", &r_rts_sem);
#endif
   }

#ifdef GC_DEBUG
int dup_flow(struct flow *f, char *msg)
{
   Bit32 x;
   for (x = 253; x <= mxflows; ++x)
      if (flow_ix[x-1] != NULL) {
         scpos(0,scr_lrow);
         printf("   flow_ix[%d]=%lu !!\n", x,flow_ix[x-1]);
         }
   for (x = 1; x <= mxflows; ++x)
      if (flow_ix[x-1] == f) {
         scpos(0,scr_lrow);
         printf("Flow %lu is flow_ix[%d], msg=%s\n", f,x, msg);
         exit(0);
         }
   return 0;
   }
#endif

Bit32 rsc_n_searches,
   rsc_mx_range, rsc_av_range,  /* Flow index range searched */
   rsc_mx_tests, rsc_av_tests;  /* Nbr of TimeFilter tests */

# if RS_CH_DEBUG
void check_rs_chain(int rs, char *msg)
{
   Bit32 fx;
   int count;
   for (count = 0, fx = ri[rs-1].ri_flow_chain;
         fx != 0;  fx = fa[fx-1].rs_next) {
      if (++count > ri[rs-1].ri_FlowRecords) break;
      }
   if (count != ri[rs-1].ri_FlowRecords)
      printf("check_rs_chain(%d, %s): count=%d, FlowRecords=%d\n",
         rs,msg, count, ri[rs-1].ri_FlowRecords);
   }
# endif

void mark_idle(Bit32 fx)
{
   struct flow *fp = &fa[fx-1];
   fp->FlowRuleSet = 0;   /* Mark idle */
   if (free_flow_tail == 0) free_flow_head = fx;
   else fa[free_flow_tail-1].rs_next = fx;  /* Add to tail of free chain */
   free_flow_tail = fx;
   fp->rs_next = 0;  /* End marker for rs chain */
   }

Bit32 alloc_flow_nbr(void)  /* Allocate nbr for a new flow */
{
   Bit32 fx;

   if (already_complained) {  /* We know there are no free flows */
      ++n_lost_flows;
      return 0;  /* already_complained reset by set_LastTime() */
      }

   if (free_flow_head == 0) {  /* No idle flows: force garbage collection */
      if (garbage_collect(0))
         already_complained = 0;
      }
   if (free_flow_head != 0) {
      fx = free_flow_head;
      free_flow_head = fa[fx-1].rs_next;  /* Take from head of free chain */
      if (free_flow_head == 0) free_flow_tail = 0;
      if (nflows == 0) gcf_ix = fx;  /* First flow being allocated */
      return fx;
      }
 
   ++n_lost_flows;
   if (already_complained) return 0;
   more_garbage();  /* Increase garbage collecting rate */
   log_msg(LOG_WARNING, 0, "Recovering flows too slow !!!");
   already_complained = 1;
   return 0;
   }

struct flow *find_flow(Bit32 x)  /* Find flow with flow_nbr x */
{
   if (x > mxflows || x < 1) {
      log_msg(LOG_ERR, 0, "Invalid flow number (%d)", x);
      return NULL;
      }
#if FLOW_INDEX
   return flow_ix[x-1];
#else
   return &fa[x-1];
#endif
   }

void more_garbage()
{
   int max_gc_f;
   gc_interval >>= 1;  /* Decrease garbage collection interval */
   if (gc_interval < 2) {
      gc_interval = 2;
      max_gc_f = (mxflows)/3;
      //**      max_gc_f = (mxflows)/7;
      gc_f <<= 1;  /* Increase nbr of flows checked each collection */
      if (gc_f > max_gc_f) gc_f = max_gc_f;
      }
   }

#define GC_TRACE 0
int garbage_collect(int incremental)
{
   int f,e, gcf_rs, starting;
   Bit32 start_gcf_ix, next_gcf_ix;
   Bit32 GCtime;
   struct rule *rp;
   struct flow *fp, *np, *tp, *ltp;
#ifdef GC_DEBUG
   int k;
#endif
   Bit32 rs_first_flow;
   struct flow *pf;
   int rs;
   int result = 0;

   if (nflows < gc_f) return 1;  /* No point in garbage collecting yet! */
   e = gc_f;
   f = incremental ? gc_f/2 : mxflows;
   /* mxflows catches case where non-incremental recovers the start_gcf_ix flow */
   //**   e = gc_f;  f = gc_f/4;

   /* gcf_ix = index of the next flow garbage_collect() will test */
#if FLOW_INDEX
   fp = flow_ix[gcf_ix-1];
#else
   fp = &fa[gcf_ix-1];
#endif

   for (starting = 1; ; ) {
      if (gcf_ix == 0) {  /* End of a ruleset chain */
	 do {  
            gcf_rs = ++gcf_rs % MXRTBLS;
            } while ((gcf_ix = ri[gcf_rs].ri_flow_chain) == 0);
         GCtime = ri[gcf_rs].ri_gc_time;
#if GC_TRACE
         printf("\n   gc(2): gcf_rs=%d, GCtime=%u, gcf_ix=%u ",
            gcf_rs, GCtime, gcf_ix); 
	 if (--count == 0) quack(6666);  //***
#endif
         }

#if FLOW_INDEX
      fp = flow_ix[gcf_ix-1];
#else
      fp = &fa[gcf_ix-1];
#endif
      if (starting) {  /* First time round loop! */
         gcf_rs = fp->FlowRuleSet-1;  /* 0-org */
         GCtime = ri[gcf_rs].ri_gc_time;
         start_gcf_ix = gcf_ix;
#if GC_TRACE
         printf("\n\ngc(1): gcf_rs=%d, GCtime=%u, gcf_ix=%u, start_gcf_ix=%u, f=%u ",
            gcf_rs, GCtime, gcf_ix, start_gcf_ix, f); 
#endif
         starting = 0;
         }
      else if (gcf_ix == start_gcf_ix)
         break;  /* No free flows */

      next_gcf_ix = fp->rs_next;  /* Save for next time around loop */

      if (flow_idle(gcf_ix)) {  /* Unused flow index */
         gcf_ix = next_gcf_ix;
         result = 1;
         if (incremental && --e == 0) return result;
         continue;
         }

#if GC_TRACE
      //         printf(" %u:%u", gcf_ix, fp->ntm_LastTime); 
#endif
      if (fp->ntm_LastTime <= GCtime) {  /* Inactive, recover it */
#if GC_TRACE
         printf("=%u", fp->sf_state); 
#endif

#if STREAM_TO_FLOW
         if (fp->sf_state == SF_ACTIVE) { /* Stream still active */
            gcf_ix = next_gcf_ix;
            if (incremental && --e == 0) return result;
            continue;
	    }
#endif
         result = 1;
         mutex_lock("A", &flow_table_mutex);
         if (!flow_orphan(fp)) {  /* Remove from hash chain */

#ifdef GC_DEBUG
            scpos(0,scr_lrow);
            printf("GC: set=%u, ix=%u, fp=%lu, np=%lu\n",
               fp->FlowRuleSet,gcf_ix, fp,fp->ht_next);
            k = 0;
#endif
            np = fp->ht_next;
            tp = fp;  do {
               ltp = tp;
#ifdef GC_DEBUG
printf("   k=%d, ltp=%lu, tp=%lu\n", ++k, ltp,ltp->ht_next);
if (tp->ht_next == NULL || k > 10) exit(0);  /* Infinite loop - stop */
#endif
               } while ((tp = tp->ht_next) != fp);
            if (ltp == np) {  /* Single flow in hash chain */    
               mark_orphan(ltp);
               --n_hash_ents;
	       }
            else ltp->ht_next = np;
            mark_orphan(fp);  /* Added 15 May 99, Nevil */
            }

         --ri[fp->FlowRuleSet-1].ri_FlowRecords;
         --ri[fp->FlowRuleSet-1].ri_recoverable_flows;
         rs = fp->FlowRuleSet;
         rs_first_flow = ri[rs-1].ri_flow_chain;
         if (gcf_ix == rs_first_flow) {  /* First flow of chain */
            ri[rs-1].ri_flow_chain = fp->rs_next;
            }
         else {  /* Find previous link in ruleset chain */
            pf = fp;  do {
               --pf;
               } while (pf->FlowRuleSet != rs);
            pf->rs_next = fp->rs_next;
            }
# if RS_CH_DEBUG
         check_rs_chain(rs, "GC");
# endif
         mark_idle(gcf_ix);

         free_flow(fp);
         mutex_unlock("A", &flow_table_mutex);

         --nflows;  ++FlowsRecovered;
         gcf_ix = next_gcf_ix;
         if (!incremental) return result;  /* Forced collection succeeded,
                                         flow index gcf_ix is now free */
         }
      else {  /* Not inactive */
         gcf_ix = next_gcf_ix;
         }
      if (--f == 0) return result;
      } 
   return 0;  /* Failed to recover any flows */
   }

#ifdef GC_DEBUG
void print_chain(struct hash_tbl *t, Bit16 hash, char *msg)
{
   struct flow **sf;
   struct flow *f, *lf;
   sf = &t->hash_ent[hash];  /* Address of hash table entry */
   scpos(0,scr_lrow);  printf("Chain: t=%lu, hash=%u, sf=%lu, msg=%s\n",
      t,hash,sf, msg);
   lf = f = sf[0];
   do {
      printf("   f=%lu\n", f);
      if ((lf = f) == NULL) return;
      } while ((f = lf->ht_next) != (struct flow *)sf);
   }
#endif

void update_flows_for_stream(struct stream *sp,
   void  (*update_flow)(struct flow *fp, struct stream *sp))
{
   int j,k;
   struct flow *fp;
   Bit32 fx;

   for (j = 0; j != sp->f_pmi.n_flows; ++j) {  /* Forward flows */
      fp = flow_ptr(sp->f_pmi.fx[j]);
      update_flow(fp, sp);
       }

   for (j = 0; j != sp->r_pmi.n_flows; ++j) {  /* Reverse flows */
      fx = sp->r_pmi.fx[j];
      for (k = 0; k != sp->f_pmi.n_flows; ++k) {
         if (sp->f_pmi.fx[k] == fx) break;
         }
      if (k == sp->f_pmi.n_flows)  /* Not also a forward flow */
         update_flow(flow_ptr(fx), sp);
      }
   }

#if APP_DATA

#define OLD_P2P_TEST  0  /* Trying Thomas' new test code */
# if OLD_P2P_TEST
struct p2p {
   int t_mp,u_mp;  /* Test TCP, UDP pkts */
   int mn_rm;      /* Stream not p2p if we don't get this many matches */
   int mx_tests;   /* Stop testing after this many packets */
   int delta_j;    /* Nbr of strings to skip on first-char != */
   Bit8 *str;  /* String to match */
   int p;          /* p2p protocol */
   int strlen;     /* strlen(str); computed in init_monitor() */
   };
#define MX  14
// #define MX  20

/* Dag parameters: -V -L64   (legacy format, -L48)
      => 16 chars payload, allowing for SDLC header and MPLS shim */

struct p2p strings[] = {
   { 0, 0,  0,   0,  0, "",                          0, 0 }, /* Not p2p */

   { 1, 1,  2,  MX,  1, "\xe3",                      1, 0 }, /* EDONKEY*/
   { 1, 1,  2,  MX,  1, "\xc5",                      1, 0 },

   { 1, 0,  1,  MX,  5, "GET /uri-res/",             4, 0 }, /* GNUTELLA */
   { 1, 0,  1,  MX,  4, "GET /get",                  4, 0 },
   { 1, 0,  1,  MX,  3, "GIV ",                      4, 0 },
   { 0, 1,  1,  MX,  2, "GND",                       4, 0 },
   { 1, 0,  1,  MX,  1, "GNUTELLA",                  4, 0 },
   { 1, 0,  1,  MX,  1, "Server",                    4, 0 },
   { 1, 0,  1,  MX,  9, "X-Ul",                      4, 0 },
   { 1, 0,  1,  MX,  8, "X-Ve",                      4, 0 },
   { 1, 0,  1,  MX,  7, "X-Dy",                      4, 0 },
   { 1, 0,  1,  MX,  6, "X-Ma",                      4, 0 },
   { 1, 0,  1,  MX,  5, "X-Qu",                      4, 0 },
   { 1, 0,  1,  MX,  4, "X-Gu",                      4, 0 },
   { 1, 0,  1,  MX,  3, "X-Ex",                      4, 0 },
   { 1, 0,  1,  MX,  2, "X-Tr",                      4, 0 },
   { 1, 0,  1,  MX,  1, "X-De",                      4, 0 },

   { 1, 0,  1,  MX,  2, "GET /.hash",                5, 0 }, /* FASTTRACK */
   { 1, 0,  1,  MX,  3, "GIVE",                      5, 0 },
   { 0, 1,  1,  MX,  1, "\x27\x00\x00\x00\x29\x80",  5, 0 },
   { 0, 1,  1,  MX,  1, "\x28\x00\x00\x00\x29\x00",  5, 0 },
   { 0, 1,  1,  MX,  1, "\x29\x00\x00\x00\x29",      5, 0 },

   { 1, 0,  1,  MX, 12, "$Direction",                6, 0 }, /* DIRECT    */
   { 1, 0,  1,  MX, 11, "$Connect",                  6, 0 }, /*   CONNECT */
   { 1, 0,  1,  MX, 10, "$Search",                   6, 0 },
   { 1, 0,  1,  MX,  9, "$Hello",                    6, 0 },
   { 1, 0,  1,  MX,  8, "$Send",                     6, 0 },
   { 1, 0,  1,  MX,  7, "$Lock",                     6, 0 },
   { 1, 0,  1,  MX,  6, "$Quit",                     6, 0 },
   { 0, 1,  1,  MX,  5, "$Ping",                     6, 0 },
   { 1, 0,  1,  MX,  4, "$Get",                      6, 0 },
   { 1, 0,  1,  MX,  3, "$Sup",                      6, 0 },
   { 1, 0,  1,  MX,  2, "$My",                       6, 0 },
   { 0, 1,  1,  MX,  1, "$SR",                       6, 0 }

/* , { 1, 0,  1,  MX,  3, "GET",                    99, 0 } /* Testing */
   };

int n_strings = sizeof(strings) / sizeof(struct p2p);

int p2p_test(struct pkt *p_p)
{
   int mt,rm, pdus, result, j;
   struct stream *sp = p_p->sp;
   struct p2p *p2;

   if (sp->pm_state == PM_COUNTING)
     return sp->p2pProto;

   if (p_p->p2ptested) return p_p->p2pProto;
   p_p->p2ptested = 1;

   if (sp->pm_state != PM_NOTEST) {  /* Recover match state */
      mt = strings[sp->p2pKind].mx_tests;
      rm = strings[sp->p2pKind].mn_rm;
      if (sp->p2pmatches >= rm) {  /* Matched enough times */
         sp->pm_state == PM_COUNTING;
         return p_p->p2pProto = sp->p2pProto;
         }
      }
   else {
      mt = MX;  rm = 0;
      }

   pdus = p_p->opp_dir ? sp->FromPDUs : sp->ToPDUs;
   if (pdus > mt+1) {  /* Only test first mx_tests pkts */
      sp->pm_state == PM_COUNTING;
      return sp->p2pProto;
#if P2PK_DEBUG
     printf("... p2p_t test: p_p=%0x, strm %d, pre=%d, kind=%d, matches=%d\n",
        p_p, stream_nbr(sp), sp->pre_p2p_fx, sp->p2pKind, sp->p2pmatches);
#endif
      }
   if (pdus == mt+1 && sp->p2pmatches < rm) {
         /* Didn't reach mn_seen within first mx_tests pkts */
      sp->pm_state == PM_FAILED;
      return p_p->p2pProto = 0;
      }

   result = 0;
   for (j = 1; j != n_strings; ) {
      p2 = &strings[j];
      if (p_p->pktinfo_sz != 0 &&  /* At least one char to test */
            p2->str[0] != p_p->pi.tcp.data[0])  /* First char doesn't match */
         j += p2->delta_j;
      else if (p2->strlen <= p_p->pktinfo_sz &&  /* Enough chars to test */
            memcmp(p_p->pi.tcp.data, p2->str, p2->strlen) == 0) {
         if (p_p->TransAddrType == PT_TCP) {
            if (p2->t_mp) result = 1;
            break;
            }
         else if (p_p->TransAddrType == PT_UDP) {
            if (p2->u_mp) result = 1;
            break;
            }
         else ++j;
         }
      else ++j;
      }

   switch (sp->pm_state) {
   case PM_NOTEST:
      if (result == 0) sp->pm_state = PM_EARLY;
      else {  /* First match */
         sp->p2pKind = j;  /* Remember which string matched */
         result = p2->p;
         sp->pm_state = PM_MATCHED;
         }
      break;
   case PM_WAITING:
      if (result != 0) {  /* First match */
         sp->p2pKind = j;  /* Remember which string matched */
         result = p2->p;
         sp->pm_state = PM_MATCH_1;
         }
      break;
   case PM_MATCHED:
       if (result != 0) {
          if (p2->p == sp->p2pProto) {  /* Same protocol */
             ++sp->p2pmatches;
            sp->p2pKind = j;  /* Remember which string matched */
            result = p2->p;
            }
         else sp->pm_state = PM_FAILED;
	 }
      break;
   default:
      log_msg(LOG_WARNING, 0, "Stream %d, p2p_test in state %d!",
         stream_nbr(sp), sp->pm_state);
      }

   return p_p->p2pProto = sp->p2pProto = result;
   }

#else  /* OLD_P2P_TEST */

int n_p_tests = 0;  //05

static int checkudp(Bit8 *payload, u_short sport, u_short dport, u_short plen){   //<--- thomas

   //printf("Payload :%d\n", udppayload.caplen);
   //    printf("PPPPPPP %" PRIu64 "  \n",plen);
   if ( payload[0] == 0xe3 || payload[0] == 0xc5 )     
     //07      if (sport != 4665 && dport != 4665)  //don't care about standard port
	return 1;   // eDonkey   
    
   if ( memcmp(payload,"\x47\x4e\x44",3)==0)
     //07      if ( sport != 6347 && dport != 6347 && sport != 6346 && dport != 6346 )  //don't care about standard port
	return 2;  //Gnutella

   if ( memcmp(payload,"\x27\x00\x00\x00\x29\x80",6)==0 ||
	 memcmp(payload,"\x28\x00\x00\x00\x29\x00",6)==0  ||
	 memcmp(payload,"\x29\x00\x00\x00\x29",5)==0 ||
	 memcmp(payload,"\xc0\x28",2)==0  || 
	(memcmp(payload,"\xc1",1)== 0 && plen == 7) ||
	(memcmp(payload,"\x2a",1)== 0 && plen == 5)  )
     //07      if (sport != 1214 && dport != 1214)	
	return 3;   //Fasttrack

   if (strncmp(payload,"$SR ",4)==0 || strncmp(payload,"$Pin",4)==0 )
     //07      if (sport != 411 && dport != 411 && sport != 412 && dport != 412 )  //don't care about standard port
	return 4;  //DC

#if 0  // UA
   if (memcmp(&payload[2],"\x00\x00\x00",3)==0)
      if  (sport != 53 && dport != 53) // dns > false positives
	return 81;  //possible MP2P
#endif  // UA
	
   if (strncmp(&payload[2],"\xcf\x64",2)==0)
      return 10;  //ES5 ??

   return 0;	   
   }  // ---> thomas

static int checktcp(Bit8 *payload, u_short sport, u_short dport, u_short plen){   //<--- thomas
    
   if ( payload[0] == 0xe3 || payload[0] == 0xc5 )     
     //07      if (sport != 4661 && dport != 4661 && sport != 4662 && dport != 4662) //don't care about standard port
	return 1;  //eDonkey 

   if ( memcmp(payload,"GNUT",4)==0 || memcmp(payload,"GIV ",4)==0 
	 || (memcmp(payload,"X-Qu",4)==0) || (memcmp(payload,"X-Gu",4)==0) || (memcmp(payload,"X-Ul",4)==0)
	 || (memcmp(payload,"X-Ex",4)==0) || (memcmp(payload,"X-Tr",4)==0) || (memcmp(payload,"X-De",4)==0) 
	 || memcmp (payload,"GET /uri-res/",13)==0 || memcmp (payload,"GET /get/",9)==0 || 
	 memcmp (payload,"HTTP 503",8)==0 )
     //07      if ( sport != 6347 && dport != 6347 && sport != 6346 && dport != 6346 )  //don't care about standard port
	return 2; //Gnutella

   if ( memcmp(payload,"GIVE",4)==0 || memcmp(payload,"GET /.hash",10)==0 )
     //07      if (sport != 1214 && dport != 1214)  //don't care about standard port
	return 3; //Fasttrack

   if ( memcmp(payload,"$Sen",4)==0 || memcmp(payload,"$Get",4)==0 
	 || (memcmp(payload,"$Dir",4)==0) || (memcmp(payload,"$Con",4)==0) || (memcmp(payload,"$Sup",4)==0)
	 || (memcmp(payload,"$Sea",4)==0) || (memcmp(payload,"$MyI",4)==0) || (memcmp(payload,"$Hel",4)==0) 
	 || (memcmp(payload,"$Loc",4)==0) || (memcmp(payload,"$MyN",4)==0) || (memcmp(payload,"$Qui",4)==0) 
	 || (memcmp(payload,"$Key",4)==0) )	 
     //07      if (sport != 411 && dport != 411 && sport != 412 && dport != 412 )  //don't care about standard port
	return 4; //DC

   if (  memcmp(payload,"\x13\x42\x69\x74",4)==0 || memcmp(payload,"\x00\x00\x00\x05",4)==0 
	  || memcmp(payload,"\x00\x00\x40\x09",4)==0 || memcmp(payload,"\x00\x00\x00\x0d",4)==0 ||
	  memcmp(payload,"GET /announce?",14)==0 || memcmp(payload,"GET /torrents/",14)==0 || 
	  memcmp(payload,"GET /TrackPak",13)==0 )
     //07      if ((sport > 6889 || sport < 6881) && (dport > 6889 || dport < 6881))  //don't care about standard port
	return 5; //BitTorrent

   if ( (memcmp(&payload[2],"\x00\x00\x01\x00",4)==0 && plen-4 == *(u_short *) payload) // login
	 || (memcmp(&payload[2],"\x00\x00\x03\x00",4)==0 && plen-4 == *(u_short *) payload) // GetPeerAddress
	 || (memcmp(&payload[2],"\x00\x00\x07\x00",4)==0 && plen-4 == *(u_short *) payload) // GetUserStatus
	 || (memcmp(&payload[2],"\x00\x00\x12\x00",4)==0 && plen-4 == *(u_short *) payload) // ConnectToPeer
	 || (memcmp(&payload[2],"\x00\x00\x1a\x00",4)==0 && plen-4 == *(u_short *) payload) // FileSearch
	 || (memcmp(&payload[2],"\x00\x00\x09\x00",4)==0 && plen-4 == *(u_short *) payload) // Search resutls
	 || (memcmp(&payload[2],"\x00\x00\x28\x00",4)==0 && plen-4 == *(u_short *) payload) // Transfer request
	 || (memcmp(&payload[2],"\x00\x00\x29\x00",4)==0 && plen-4 == *(u_short *) payload) // Transfer response
	 || (memcmp(&payload[2],"\x00\x00\x2a\x00",4)==0 && plen-4 == *(u_short *) payload) // Place hold upload
	 || (memcmp(&payload[2],"\x00\x00\x03\x31\x00\x00",6)==0 )   // multiple ssk packets within one??
	 || (memcmp(&payload[2],"\x00\x00",2)==0 && plen == 4 )   // first packet when connecting to peer for dl (possibly firewalled dl)
	 )  // appears to be in little-endian do not use ntohs!!
     //07     if ((sport > 6889 || sport < 6881) && (dport > 6889 || dport < 6881))  // remove bitTorrent ports. false positives 
	return 9; //soulseek  !!! we have to compare soulseek before openap because openap has similar, more general patterns
       
   if (memcmp(payload,"SEND",4)==0 || (memcmp(payload,"GET",3)==0 && plen == 3) )
     //07      if (sport != 6699 && dport != 6699)  //don't care about standard port
	return 6;  //Opennap

   if ( (memcmp(payload,"\x31",1)==0 && plen == 1)   // start download  Openap is little-endian.. do not use ntohs!!!!!!!!!!!!!
	 || (memcmp(payload,"\x00\x00\xca\x00",4)==0 && plen == 4) // end of search respones
	 || (memcmp(&payload[2],"\xc8\x00",2)==0 && plen-4 == *(u_short *) payload) // search request 
	 || (memcmp(&payload[2],"\xc9\x00",2)==0 && plen-4 == *(u_short *) payload) // search response 
	 || (memcmp(&payload[2],"\xcb\x00",2)==0 && plen-4 == *(u_short *) payload)  // download request
	 || (memcmp(&payload[2],"\xcc\x00",2)==0 && plen-4 == *(u_short *) payload)  // download ack
	 || (memcmp(&payload[2],"\x00\x00",2)==0 && plen-4 == *(u_short *) payload)   // server error
	 || (memcmp(&payload[2],"\xd6\x00",2)==0 && plen-4 == *(u_short *) payload) // server stats
	 || (memcmp(&payload[2],"\x02\x00",2)==0 && plen-4 == *(u_short *) payload) // login
	 || (memcmp(&payload[2],"\x03\x00",2)==0 && plen-4 == *(u_short *) payload) // login ack
	 || (memcmp(&payload[2],"\x06\x00",2)==0 && plen-4 == *(u_short *) payload) // new user login
	 ) 
     //07     if (sport != 6699 && dport != 6699)  //don't care about standard port
	return 61;  //possible Opennap
            
   if (memcmp(payload,"GET hash:",9)==0 || (memcmp(payload,"PUSH ",5)==0) )
      return 7;  //Ares

   if (memcmp(payload,"GO!!",4)==0 || memcmp(payload,"MD5 ",4)==0 || 
	memcmp(payload,"SIZ ",4)==0  || memcmp(payload,"STR ",4)==0)
      return 8;  //MP2P

   if ( (memcmp(payload,"PUT http",8)==0 || memcmp(payload,"HEAD http",9)==0 || memcmp(payload,"GET http",8)==0 ) 
	 && sport !=80 && dport !=80  && sport !=8080 && dport !=8080 && sport !=8000 && dport !=8000
	 && sport !=81 && dport !=81  )
      return 101; //es5 ??

   if  (memcmp(payload,"GET /",5)==0 && memcmp(&payload[14],"/",1)==0){
      int i=5;
      while (isdigit(payload[i])!=0 && i<14)
	i++;
      if (i==14)
	return 10; //es5
    }
    
#if 0  // UA
   if ( (memcmp (payload,"HTTP",4)==0 || memcmp (payload,"GET ",4)==0) 
	 && sport !=80 && dport !=80  && sport !=8080 && dport !=8080 && sport !=8000 && dport !=8000
	 && sport !=81 && dport !=81 && sport !=1214 && dport !=1214 && sport !=6346 && dport !=6346 && sport !=6347 && dport !=6347
	 && sport !=8200 && dport !=8200 )  // http,http proxy ports, kazaa, gnutella, and expertcity port 8200 (remote web services)
      return 99; // unclassified HTTP possible p2p
#endif  // UA

   return 0;	   
   }  // ---> thomas

#if LOG_P2P_TAGS
FILE *p2p_log = NULL;
# if DAG_DIRECT  /* Really need fmt_start_ts() in meter_ux.c <<<<< */
extern long long meter_start_time;  /* Dag (64-bit) timestamp */
# else
extern struct timeval meter_start_time;
# endif

struct stream *last_sp = NULL, *last_s_sp = NULL, *last_d_sp = NULL;

void log_p2p(struct stream *sp, struct stream *s_sp, struct stream *d_sp)
{
   if (p2p_log == NULL) {
      if ((p2p_log = fopen("p2p.log", "w")) == NULL)
         log_msg(LOG_ERR, 1, "Couldn't open p2p.log <<<<<");
      fprintf(p2p_log, "%lu.%06lu meter_start_time\n",
# if DAG_DIRECT
         DAGT_S(meter_start_time), DAGT_US(meter_start_time));
# else
         meter_start_time.tv_sec, meter_start_time.tv_usec);
# endif
      }

if (sp == last_sp && s_sp == last_s_sp && d_sp == last_d_sp)
     quack(1234);
last_sp = sp;  last_s_sp = s_sp;  last_d_sp = d_sp;

   fprintf(p2p_log, "%u  %d  %d %d  ",
      sp->LastTime, sp->p2pProto, sp->sfn.PeerType, sp->sfn.TransType);

   if (sp->f_pmi.n_flows != 0 && sp->f_pmi.flow_order[0] ||
         /* Wire order is FORWARD */
      sp->r_pmi.n_flows != 0 && !sp->r_pmi.flow_order[0]) {
         /* Opposite order is REVERSE */
# if QBYTE_PEER
      fdaddr(p2p_log, sp->sfn.SrcPeerAddr);  fprintf(p2p_log, " ");
      fdaddr(p2p_log, sp->sfn.DestPeerAddr);
# else
      fdaddr(p2p_log, sp->sfn.SrcPeerAddr[0]);  fprintf(p2p_log, " ");
      fdaddr(p2p_log, sp->sfn.DestPeerAddr[0]);
# endif
      fprintf(p2p_log, "  %u %u",
         ntohs(sp->sfn.SrcTransAddr), ntohs(sp->sfn.DestTransAddr));
      if (s_sp != NULL || d_sp != NULL) {
         fprintf(p2p_log, "  T %u  %u %u  %u %u\n", sp->p2p_bytag,
            s_sp == NULL ? 0 : s_sp->ToPDUs,  /* Nbr of p2p tag calls */
            d_sp == NULL ? 0 : d_sp->ToPDUs,
            s_sp == NULL ? 0 : s_sp->FromPDUs,  /* Nbr of p2p lookup calls */
            d_sp == NULL ? 0 : d_sp->FromPDUs);
         }
      else {
         fprintf(p2p_log, "  F %u  %u %u  %u %u\n", sp->p2p_bytag,
            sp->ToPDUs, sp->FromPDUs, sp->ToOctets, sp->FromOctets);
         }
      }
   else {
# if QBYTE_PEER
      fdaddr(p2p_log, sp->sfn.DestPeerAddr);  fprintf(p2p_log, " ");
      fdaddr(p2p_log, sp->sfn.SrcPeerAddr);
# else
      fdaddr(p2p_log, sp->sfn.DestPeerAddr[0]);  fprintf(p2p_log, " ");
      fdaddr(p2p_log, sp->sfn.SrcPeerAddr[0]);
# endif
      fprintf(p2p_log, "  %u %u",
         ntohs(sp->sfn.DestTransAddr), ntohs(sp->sfn.SrcTransAddr));
      if (s_sp != NULL || d_sp != NULL) {
         fprintf(p2p_log, "  T %u  %u %u  %u %u\n", sp->p2p_bytag,
            s_sp == NULL ? 0 : s_sp->ToPDUs,  /* Nbr of p2p tag calls */
            d_sp == NULL ? 0 : d_sp->ToPDUs,
            s_sp == NULL ? 0 : s_sp->FromPDUs,  /* Nbr of p2p lookup calls */
            d_sp == NULL ? 0 : d_sp->FromPDUs);
         }
      else {
         fprintf(p2p_log, "  F %u  %u %u  %u %u\n", sp->p2p_bytag,
            sp->FromPDUs, sp->ToPDUs, sp->FromOctets, sp->ToOctets);
         }
      }
   }
#endif LOG_P2P_TAGS

void tag_p2p_host(struct stream *strp)
{
   struct pkt pp;
   struct stream *sp, *dp;

   memset(&pp, 0, sizeof(struct pkt));
   pp.arrival_time = strp->StrLastTime;
   pp.TransAddrType = strp->sfn.TransType;
   pp.PeerAddrType = AT_P2P;
#if QBYTE_PEER
   pp.Low.PeerAddress.qbyte = strp->sfn.SrcPeerAddr;
#else
   memcpy(pp.Low.PeerAddress.peer32, 
      strp->sfn.SrcPeerAddr, PEER_ADDR_LEN);
#endif
   if (lookup_stream(&pp, 1, 1)) {  /* Low host */
      sp = pp.sp;
      sp->n_pkts_queued = 0;
      ++sp->ToPDUs;  /* ToPDUs counts host's p2p streams */
      sp->ToOctets = strp->p2pProto;
      }

#if QBYTE_PEER
   pp.Low.PeerAddress.qbyte = strp->sfn.DestPeerAddr;
#else
   memcpy(pp.Low.PeerAddress.peer32, 
      strp->sfn.DestPeerAddr, PEER_ADDR_LEN);
#endif
   if (lookup_stream(&pp, 1, 1)) {  /* High host */
      dp = pp.sp;
      dp->n_pkts_queued = 0;
      ++dp->ToPDUs;  /* ToPDUs counts host's p2p streams */
      dp->ToOctets = strp->p2pProto;
      }

#if LOG_P2P_TAGS
   log_p2p(strp, sp, dp);  /* Start of stream */
#endif
   }

int p2p_host_tagged(struct stream *strp)
{
   struct pkt pp;
   struct stream *sp, *dp;
   int r,s;

   r = s = 0;
   memset(&pp, 0, sizeof(struct pkt));
   pp.arrival_time = strp->StrLastTime;
   pp.TransAddrType = strp->sfn.TransType;
   pp.PeerAddrType = AT_P2P;
#if QBYTE_PEER
   pp.Low.PeerAddress.qbyte = strp->sfn.SrcPeerAddr;
#else
   memcpy(pp.Low.PeerAddress.peer32, 
      strp->sfn.SrcPeerAddr, PEER_ADDR_LEN);
#endif
   if (lookup_stream(&pp, 0, 1)) {  /* Low host */
      sp = pp.sp;
      sp->n_pkts_queued = 0;
      ++sp->FromPDUs;  /* FromPDUs counts host's p2p lookups */
      r = sp->ToOctets;  /* p2pProto when last tagged */
//      printf(" %d.%d", sp->ToPDUs, sp->FromPDUs);  //06 
      }

#if QBYTE_PEER
   pp.Low.PeerAddress.qbyte = strp->sfn.DestPeerAddr;
#else
   memcpy(pp.Low.PeerAddress.peer32, 
      strp->sfn.DestPeerAddr, PEER_ADDR_LEN);
#endif
   if (lookup_stream(&pp, 0, 1)) {  /* High host */
      dp = pp.sp;
      dp->n_pkts_queued = 0;
      ++dp->FromPDUs;  /* FromPDUs counts host's p2p lookups */
//      printf(" %d:%d", sp->ToPDUs, sp->FromPDUs);  //06 
      s = dp->ToOctets;  /* p2pProto when last tagged */
      }

   if (r != 0 && s != 0)  /* Both hosts tagged */
      return r == s ? r : 0;
   else return 0;
   }

int p2p_test(struct pkt *p_p)
{
   int pdus, result, j;
   struct stream *sp = p_p->sp;
   struct p2p *p2;

   if (sp->pm_state == PM_COUNTING) return sp->p2pProto;

   if (p_p->p2ptested) return p_p->p2pProto;
   p_p->p2ptested = 1;

#define MX_P2P_PKTS  100  //06 Sep
   pdus = p_p->opp_dir ? sp->FromPDUs : sp->ToPDUs;
   if (pdus > MX_P2P_PKTS) {  /* Only test first MX_P2P_PKTS pkts */
      if (sp->p2pProto != 0) tag_p2p_host(sp);  //02d
      sp->pm_state == PM_COUNTING;
      return sp->p2pProto;
#if P2PK_DEBUG
     printf("... p2p_t test: p_p=%0x, strm %d, pre=%d, kind=%d, matches=%d\n",
        p_p, stream_nbr(sp), sp->pre_p2p_fx, sp->p2pKind, sp->p2pmatches);
#endif
      }

   result = 0;
   if (p_p->pktinfo_sz != 0) {
      if (p_p->TransAddrType == PT_TCP)
         result = checktcp(p_p->pi.tcp.data,
             ntohs(p_p->Low.TransAddress), ntohs(p_p->High.TransAddress),
             p_p->pktinfo_sz);
      else if (p_p->TransAddrType == PT_UDP)
         result = checkudp(p_p->pi.udp.d.data,
            ntohs(p_p->Low.TransAddress), ntohs(p_p->High.TransAddress),
            p_p->pktinfo_sz);
      }
#if 0
   printf("p2p_test(): r=%d,   sp=%u. dp=%u, sz=%u, data= %02x",  //05
	r, ntohs(p_p->Low.TransAddress), ntohs(p_p->High.TransAddress), p_p->pktinfo_sz, p_p->pi.udp.d.data[0]);
   for (j = 1; j < p_p->pktinfo_sz; ++j) printf(" %02x", p_p->pi.udp.d.data[j]);
   printf("\n");  //05
#endif

   switch (sp->pm_state) {
   case PM_NOTEST:  /* First packet of stream */
      if (result == 0)  /* p2p match failed, remember zero flow */
         sp->pm_state = PM_EARLY;
      else {
         if (result != 1 && result != 61)
            /* Other p2p protocols only need to match one pkt */
            sp->pm_state = PM_COMPLETE;
	 else {
            if (result == 1) ++sp->p2pcount_p1;
            else if (result == 61) ++sp->p2pcount_p61;
            sp->pm_state = PM_MATCH_1;
            }
         }
      break;
   case PM_WAITING:  /* Waiting for first p2p match */
      if (result == 0) {

#if P2P_TEST_TAGS
#define TAG_P2P_PKTS   4  //26 Sep
         if (sp->FromPDUs + sp->ToPDUs == TAG_P2P_PKTS) {
            result = p2p_host_tagged(sp);
            if (result != 0) {  /* Tagged as p2p */
               sp->p2p_bytag = 1;  /* Identifed by previous p2p tag */
               sp->pm_state = PM_COMPLETE;
               }
            }
#endif
         }
      else {
         if (result != 1 && result != 61)
            /* Other p2p protocols only need to match one pkt */
            sp->pm_state = PM_COMPLETE;
	 else {
            if (result == 1) ++sp->p2pcount_p1;
            else if (result == 61) ++sp->p2pcount_p61;
            sp->pm_state = PM_MATCH_1;
            }
         }
      break;
   case PM_MATCHED:  /* Waiting for later matches */
      if (result == 0) {
#if P2P_TEST_TAGS
         if (sp->FromPDUs + sp->ToPDUs == TAG_P2P_PKTS) {
            result = p2p_host_tagged(sp);
            if (result != 0) {  /* Tagged as p2p */
               sp->p2p_bytag = 1;  /* Identifed by previous p2p tag */
               sp->pm_state = PM_COMPLETE;
               }
            }
#endif
         }
      else {
         if (result == 1) ++sp->p2pcount_p1;
         else ++sp->p2pcount_p61;
         j = sp->p2pcount_p1 > sp->p2pcount_p61 ? 1 : 61;
         if (j != sp->p2pProto) { /* Switch to most frequent */
            result = j;
            sp->pm_state = PM_MATCH_1;
            }
	 }
      break;
   default:
      log_msg(LOG_WARNING, 0, "Stream %d, p2p_test in state %d!",
         stream_nbr(sp), sp->pm_state);
      }

   return p_p->p2pProto = sp->p2pProto = result;
   }
#endif  /* NEW_P2P_TEST */

#if STR_ANAL
int h_e(struct host_list *lp)  /* Host_list entropy */
{
   int j,k,  d, b,e;
   for (e = 0, j = 1;  j != lp->n_hosts; ++j) {
#if QBYTE_PEER
      d = lp->hl[j-1].ha ^ lp->hl[j].ha;
      for (b = 0; d != 0; ++b) 
         d &= ~(d ^ (d-1));
#else
      for (d = k = 0; k != (PEER_ADDR_LEN+3)/4, ++k) {
         d = lp->hl[j-1].ha[k] ^ lp->hl[j].ha[k];
         for (b = 0; d != 0; ++b) 
            d &= ~(d ^ (d-1));
         }
#endif
      e += b;
      }
   return e;
   }

void log_hostlist(struct stream *sp, int reason)
{
#if LOG_HOST_LIST
   struct host_list *lp = sp->hlp;
   int j, e;

   if (lp == NULL) return;
   e = h_e(lp);
   if (e < HOSTLIST_SZ*8) return;  /* Don't log scans */

   printf("@ ");  daddr(sp->sfn.SrcPeerAddr);
   printf(" %d %d  %d %d %d  %d", reason, lp->tot_streams, 
      lp->n_gt2pkt,lp->n_gt1kB,lp->n_bidir, lp->n_hosts);

   for (j = 0; j != lp->n_hosts; ++j) {
      printf(" ");  daddr(lp->hl[j].ha);
      printf("+%u-%lu", lp->hl[j].n_streams, lp->hl[j].lasttime);
      }
   printf("\n");
#endif  /* LOG_HOST_LIST */
   }

//29 void update_tandu_count(struct flow *fp, struct stream *sp)
//29 {
//29    ++fp->stdata->n_tandu_streams;
//29   }

void update_host_list(int wo, int terminating, struct stream *strp)
{
   struct pkt pp;
   struct stream *sp;
   struct host_list *lp;
   int j,k, m,f;
   Bit32 lru_time;

   struct pkt_match_info *f_pmi, *r_pmi;
   struct flow *fp;

   memset(&pp, 0, sizeof(struct pkt));
   pp.arrival_time = strp->StrLastTime;
   pp.TransAddrType = strp->sfn.TransType;
   pp.PeerAddrType = AT_HLIST;
#if QBYTE_PEER
   pp.Low.PeerAddress.qbyte = 
      wo ? strp->sfn.SrcPeerAddr : strp->sfn.DestPeerAddr;
#else
   memcpy(pp.Low.PeerAddress.peer32, 
      wo ? strp->sfn.SrcPeerAddr : strp->sfn.DestPeerAddr,
      PEER_ADDR_LEN);
#endif

   if (lookup_stream(&pp, 1, !terminating)) {  /* Low host */
      /* If stream is terminating, don't let lookup_stream move
         it to tail of stream queue.  check_idle_streams may
         be pointng to the AT_HLIST stream which gets moved! */
      sp = pp.sp;
      sp->n_pkts_queued = 0;
      ++sp->ToPDUs;  /* Don't want it to time out too soon! */

      if (sp->hlp == NULL) {
         if ((sp->hlp = get_hlist()) == NULL)
            return;  /* Couldn't get host_list */
         sp->hlp->tot_streams = sp->hlp->n_hosts = 0;
         }
      lp = sp->hlp;

      for (j = lp->n_hosts-1; j >= 0; --j) {
#if QBYTE_PEER
         if (lp->hl[j].ha ==
            (wo ? strp->sfn.DestPeerAddr : strp->sfn.SrcPeerAddr)) {
#else
         if (memcmp(lp->hl[j].ha,
            wo ? strp->sfn.DestPeerAddr : strp->sfn.SrcPeerAddr,
            PEER_ADDR_LEN) {
#endif
            /* Already in list, update entry */

            if (lp->hl[j].protocols_seen != (TU_TCP | TU_UDP)) {
               if (strp->sfn.TransType = PT_TCP)
                  lp->hl[j].protocols_seen |= TU_TCP;
               else if (strp->sfn.TransType = PT_UDP)
                  lp->hl[j].protocols_seen |= TU_UDP;
	       }

            if (terminating) {
//29               if (lp->hl[j].protocols_seen == (TU_TCP | TU_UDP))
//29                  update_flows_for_stream(strp, update_tandu_count);

               if (strp->ToPDUs > 2 || strp->FromPDUs > 2)
                  ++lp->n_gt2pkt;
               if (strp->ToOctets > 1000 || strp->FromOctets > 1000)
                  ++lp->n_gt1kB;
               if (strp->ToPDUs > 0 && strp->FromPDUs > 0)
                  ++lp->n_bidir;
//?	       printf("*%d%d%d ", lp->n_gt2pkt,lp->n_gt1kB,lp->n_bidir);
               }
            else {
               lp->hl[j].lasttime = strp->LastTime;
               ++lp->hl[j].n_streams;
               ++lp->tot_streams;
               }
            return;
            }
         }
      
      if (lp->n_hosts < HOSTLIST_SZ)  /* Not in host list, add it */
         j = lp->n_hosts++;
      else {  /* Not enough room, overwrite lru host */
         lru_time = lp->hl[j = 0].lasttime;
         for (k = 1; k != lp->n_hosts; ++k) {
            if (lp->hl[k].lasttime < lru_time)
               lru_time = lp->hl[j = k].lasttime;
            }
         }
#if QBYTE_PEER
      lp->hl[j].ha =
         (wo ? strp->sfn.DestPeerAddr : strp->sfn.SrcPeerAddr);
#else
      memcpy(lp->hl[j].ha,
         wo ? strp->sfn.DestPeerAddr : strp->sfn.SrcPeerAddr,
         PEER_ADDR_LEN);
#endif
      if (strp->sfn.TransType = PT_TCP)
         lp->hl[j].protocols_seen |= TU_TCP;
      else if (strp->sfn.TransType = PT_UDP)
         lp->hl[j].protocols_seen |= TU_UDP;

      if (terminating) {
        if (strp->ToPDUs > 2 || strp->FromPDUs > 2)
           ++lp->n_gt2pkt;
        if (strp->ToOctets > 1000 || strp->FromOctets > 1000)
           ++lp->n_gt1kB;
        if (strp->ToPDUs > 0 && strp->FromPDUs > 0)
           ++lp->n_bidir;
//?        printf("*%d%d%d ", lp->n_gt2pkt,lp->n_gt1kB,lp->n_bidir);
         }
      else {
         lp->hl[j].lasttime = strp->LastTime;
         lp->hl[j].n_streams = 1;
         ++lp->tot_streams;
         }

      if (lp->n_hosts > 30 && lp->tot_streams <= 2000 &&
            lp->tot_streams % 100 == 0)
         log_hostlist(sp, 1);
      }  /* lookup_stream() succeeded */
   }

#endif  /* STR_ANAL */

void move_stream_counts(struct flow *newfp, struct flow *oldfp,
   struct pkt *p_p)
{
   struct stream *streamp = p_p->sp;
   counter64 topdus,frompdus, tooctets,fromoctets;

   /* Stream counts have been bumped by process_pkt() */
   if (p_p->opp_dir) {
      frompdus = streamp->FromPDUs - 1;
      fromoctets = streamp->FromOctets - p_p->p_len;
      topdus = streamp->ToPDUs;
      tooctets = streamp->ToOctets;
      }
   else {
      frompdus = streamp->FromPDUs;
      fromoctets = streamp->FromOctets;
      topdus = streamp->ToPDUs - 1;
      tooctets = streamp->ToOctets - p_p->p_len;
      }

   /* Subtract this stream's counts from oldfp */
   if (oldfp->owisFWD) {
      if (oldfp->DownPDUs < frompdus) frompdus = oldfp->DownPDUs;
         subtr_64(oldfp->DownPDUs, oldfp->DownPDUs, frompdus);
      if (oldfp->DownOctets < fromoctets) fromoctets = oldfp->DownOctets;
         subtr_64(oldfp->DownOctets, oldfp->DownOctets, fromoctets);
      if (oldfp->UpPDUs < topdus) topdus = oldfp->UpPDUs;
         subtr_64(oldfp->UpPDUs, oldfp->UpPDUs, topdus);
      if (oldfp->UpOctets < tooctets) tooctets = oldfp->UpOctets;
         subtr_64(oldfp->UpOctets, oldfp->UpOctets, tooctets);
      }
   else {
      if (oldfp->DownPDUs < topdus) topdus = oldfp->DownPDUs;
         subtr_64(oldfp->DownPDUs, oldfp->DownPDUs, topdus);
      if (oldfp->DownOctets < tooctets) tooctets = oldfp->DownOctets;
         subtr_64(oldfp->DownOctets, oldfp->DownOctets, tooctets);
      if (oldfp->UpPDUs < frompdus) frompdus = oldfp->UpPDUs;
         subtr_64(oldfp->UpPDUs, oldfp->UpPDUs, frompdus);
      if (oldfp->UpOctets < fromoctets) fromoctets = oldfp->UpOctets;
         subtr_64(oldfp->UpOctets, oldfp->UpOctets, fromoctets);
      }

   /* Now add this stream's counts to newfp */
   if (newfp->owisFWD) {
      add_64(newfp->DownPDUs, newfp->DownPDUs, frompdus);
      add_64(newfp->DownOctets, newfp->DownOctets, fromoctets);
      add_64(newfp->UpPDUs, newfp->UpPDUs, topdus);
      add_64(newfp->UpOctets, newfp->UpOctets, tooctets);
      }
   else {
      add_64(newfp->DownPDUs, newfp->DownPDUs, topdus);
      add_64(newfp->DownOctets, newfp->DownOctets, tooctets);
      add_64(newfp->UpPDUs, newfp->UpPDUs, frompdus);
      add_64(newfp->UpOctets, newfp->UpOctets, fromoctets);
      }
   }
#endif  /* APP_DATA */

void build_search_key(Bit32 *search_hash, struct flow_key *key,
   int c_rtx, struct pme_env *pme,
   struct pkt *bp, struct pkt_key *Low, struct pkt_key *High)
   /* Builds a key which current() will use to search the flow table */
{
   Bit32 hash;
   struct rule *rp;
   int pal, j, k;
   union address *Maskp;
   struct rule *c_rt = ri[c_rtx-1].ri_rule_table;

   hash = c_rtx*prime_14;
      /* Include ruleset ix in search key.  Nevil, 11 Sep 98 */
   memset(key, 0, sizeof(struct flow_key));  /* Start with a null key */
   pal = (PEER_ADDR_LEN+3)/4;  /* Could be any type, start by using max */
#ifdef TESTING
   if (!logfl) open_log();
   fprintf(logfl,"p_s_depth=%d\n", pme->p_s_depth);
#endif
   for (j = 0; j != pme->p_s_depth; ++j) {
      rp = &c_rt[pme->p_stack[j]-1];
      Maskp = &Masks[rp->RuleMaskVal];
#ifdef TESTING
      fprintf(logfl,"   j=%d, attrib=%d\n", j,pme->a_stack[j]);
#endif
      switch (pme->a_stack[j]) {
      case FTLOWPEERADDRESS:
         key->Low.PeerMaskVal = rp->RuleMaskVal;
#if QBYTE_PEER
         if (rp->push_rule)
            hash += (key->Low.PeerAddress.qbyte =
               rp->RuleMatchedValue.qbyte)*prime_0;
         else
            hash += (key->Low.PeerAddress.qbyte =
               Low->PeerAddress.qbyte & Maskp->qbyte)*prime_0;
#else
         if (rp->push_rule) for (k = 0; k != pal; ++k)
            hash += (key->Low.PeerAddress.peer32[k] =
               rp->RuleMatchedValue.peer32[k])*primes[k];
         else for (k = 0; k != pal; ++k)
            hash += (key->Low.PeerAddress.peer32[k] =
               Low->PeerAddress.peer32[k] & Maskp->peer32[k])*primes[k];
#endif
         break;
      case FTHIPEERADDRESS:
         key->High.PeerMaskVal = rp->RuleMaskVal;
#if QBYTE_PEER
         if (rp->push_rule)
            hash += (key->High.PeerAddress.qbyte =
               rp->RuleMatchedValue.qbyte)*prime_0;
         else
            hash += (key->High.PeerAddress.qbyte =
               High->PeerAddress.qbyte & Maskp->qbyte)*prime_0;
#else
         if (rp->push_rule) for (k = 0; k != pal; ++k)
            hash += (key->High.PeerAddress.peer32[k] =
               rp->RuleMatchedValue.peer32[k])*primes[k];
         else for (k = 0; k != pal; ++k)
            hash += (key->High.PeerAddress.peer32[k] =
               High->PeerAddress.peer32[k] & Maskp->peer32[k])*primes[k];
#endif
         break;
      case FTLOWPEERTYPE:
      case FTHIPEERTYPE:
         if (rp->push_rule) hash +=
            (key->PeerAddrType = rp->RuleMatchedValue.byte)*prime_1;
         else hash +=
            (key->PeerAddrType = bp->PeerAddrType)*prime_1;
         pal = addr_len32[key->PeerAddrType];  /* Now we know peer type */
         break;
      case FTLOWTRANSADDRESS:
         key->Low.TransMaskVal = rp->RuleMaskVal;
         if (rp->push_rule)
            hash += (key->Low.TransAddress =
               rp->RuleMatchedValue.trans)*prime_2;
         else
            hash += (key->Low.TransAddress =
               Low->TransAddress & Maskp->trans)*prime_2;
         break;
      case FTHITRANSADDRESS:
         key->High.TransMaskVal = rp->RuleMaskVal;
         if (rp->push_rule)
            hash += (key->High.TransAddress =
               rp->RuleMatchedValue.trans)*prime_2;
         else
            hash += (key->High.TransAddress =
               High->TransAddress & Maskp->trans)*prime_2;
         break;
      case FTLOWTRANSTYPE:
      case FTHITRANSTYPE:
         if (rp->push_rule) hash +=
            (key->TransAddrType = rp->RuleMatchedValue.byte)*prime_3;
         else hash +=
            (key->TransAddrType = bp->TransAddrType)*prime_3;
         break;
      case FTLOWADJACENTADDRESS:
         key->Low.AdjMaskVal = rp->RuleMaskVal;
         if (rp->push_rule) {
            hash += (key->Low.AdjAddr_ms4 = 
               rp->RuleMatchedValue.ma.ms4)*prime_4;
            hash += (key->Low.AdjAddr_ls2 =
               rp->RuleMatchedValue.ma.ls2)*prime_5;
            }
         else {
            hash += (key->Low.AdjAddr_ms4 =
               Low->AdjAddr_ms4 & Maskp->ma.ms4)*prime_4;
            hash += (key->Low.AdjAddr_ls2 =
               Low->AdjAddr_ls2 & Maskp->ma.ls2)*prime_5;
            }
         break;
      case FTHIADJACENTADDRESS:
         key->High.AdjMaskVal = rp->RuleMaskVal;
         if (rp->push_rule) {
            hash += (key->High.AdjAddr_ms4 = 
               rp->RuleMatchedValue.ma.ms4)*prime_4;
            hash += (key->High.AdjAddr_ls2 =
               rp->RuleMatchedValue.ma.ls2)*prime_5;
            }
         else {
            hash += (key->High.AdjAddr_ms4 =
               High->AdjAddr_ms4 & Maskp->ma.ms4)*prime_4;
            hash += (key->High.AdjAddr_ls2 =
               High->AdjAddr_ls2 & Maskp->ma.ls2)*prime_5;
            }
         break;
      case FTSOURCECLASS:
         hash += (key->Low.Class = rp->RuleMatchedValue.byte)*prime_9;
         break;
      case FTDESTCLASS:
         hash += (key->High.Class = rp->RuleMatchedValue.byte)*prime_9;
         break;
      case FTFLOWCLASS:
         hash += (key->FlowClass = rp->RuleMatchedValue.byte)*prime_9;
         break;
      case FTSOURCEKIND:
         hash += (key->Low.Kind = rp->RuleMatchedValue.byte)*prime_10;
         break;
      case FTDESTKIND:
         hash += (key->High.Kind = rp->RuleMatchedValue.byte)*prime_10;
         break;
      case FTFLOWKIND:
         hash += (key->FlowKind = rp->RuleMatchedValue.byte)*prime_10;
         break;
      case FTPAIRINTERFACE:
      case FTLOWINTERFACE:
         if (rp->push_rule) hash +=
            (key->Low.Interface = rp->RuleMatchedValue.trans)*prime_7;
         else hash +=
            (key->Low.Interface = Low->Interface)*prime_7;
         break;
      case FTHIINTERFACE:
         if (rp->push_rule) hash +=
            (key->High.Interface = rp->RuleMatchedValue.trans)*prime_7;
         else hash +=
            (key->High.Interface = High->Interface)*prime_7;
         break;
      case FTOBSINTERFACE:
         if (rp->push_rule) hash +=
            (key->ObsInterface = rp->RuleMatchedValue.trans)*prime_7;
         else hash +=
            (key->ObsInterface = Low->ObsInterface)*prime_7;
         break;
      case FTLOWADJACENTTYPE:
         if (rp->push_rule) hash +=
            (key->Low.AdjAddrType = rp->RuleMatchedValue.byte)*prime_6;
         else hash +=
            (key->Low.AdjAddrType = Low->AdjType)*prime_6;
         break;
      case FTHIADJACENTTYPE:
         if (rp->push_rule) hash +=
            (key->High.AdjAddrType = rp->RuleMatchedValue.byte)*prime_6;
         else hash +=
            (key->High.AdjAddrType = High->AdjType)*prime_6;
         break;
      case FTDSCODEPOINT:
         if (rp->push_rule) hash +=
            (key->DSCodePoint = rp->RuleMatchedValue.byte)*prime_8;
         else hash +=
            (key->DSCodePoint = bp->DSCodePoint)*prime_8;
         break;
      case FTPRIORITY:
         if (rp->push_rule) hash +=
            (key->Priority = rp->RuleMatchedValue.byte)*prime_16;
         else hash +=
            (key->Priority = bp->Priority)*prime_16;
         break;
      case FTECNCODEBITS:
         if (rp->push_rule) hash +=
            (key->ECNCodeBits = rp->RuleMatchedValue.byte)*prime_14;
         else hash +=
            (key->ECNCodeBits = bp->ECNCodeBits)*prime_14;
         break;
      case FTVLANID:
         if (rp->push_rule) hash +=
            (key->VLANid = rp->RuleMatchedValue.trans)*prime_15;
         else hash +=
            (key->VLANid = bp->VLANid & Maskp->trans)*prime_15;
         break;
# if USER_INFO
      case FTLOWSUBSCRIBERID:  /* UserID */
         if (!rp->push_rule) for (k = 0; k != SESSION_ID_Q_LEN; ++k)
            key->UserID.qba[k] = uinfo[pme->user_ix].d.UserID.qba[k];
         break;
      case FTHISUBSCRIBERID:  /* UserCoS */
         if (rp->push_rule) hash +=
            (key->UserCoS = rp->RuleMatchedValue.trans)*prime_17;
         else hash +=
            (key->UserCoS = uinfo[pme->user_ix].d.UserCoS
               & Maskp->trans)*prime_17;
         break;
      case FTSESSIONID:  /* UserInfo */
         if (rp->push_rule) hash +=
            (key->UserInfo = rp->RuleMatchedValue.trans)*prime_18;
         else hash +=
            (key->UserInfo = pme->user_attrib & Maskp->trans)*prime_18;
         break;
# endif
#if NF_ASN_ATT
      case FTLOWROUTEASN:
         key->Low.ASNMaskVal = rp->RuleMaskVal;
         if (rp->push_rule)
            hash += (key->Low.RouteASN = rp->RuleMatchedValue.trans)*prime_11;
         else
            hash += (key->Low.RouteASN = Low->nf_ASN & Maskp->trans)*prime_11;
         break;
      case FTHIROUTEASN:
         key->High.ASNMaskVal = rp->RuleMaskVal;
         if (rp->push_rule)
            hash += (key->High.RouteASN = rp->RuleMatchedValue.trans)*prime_11;
         else
            hash += (key->High.RouteASN = High->nf_ASN & Maskp->trans)*prime_11;
         break;
      case FTLOWROUTEPREFIX:
         if (rp->push_rule) hash +=
            (key->Low.RoutePrefix = rp->RuleMatchedValue.byte)*prime_12;
         else hash +=
            (key->Low.RoutePrefix = Low->nf_mask)*prime_12;
         break;
      case FTHIROUTEPREFIX:
         if (rp->push_rule) hash +=
            (key->High.RoutePrefix = rp->RuleMatchedValue.byte)*prime_12;
         else hash +=
            (key->High.RoutePrefix = High->nf_mask)*prime_12;
         break;
#endif
#if NF_OTHER_ATT
      case FTMETERID:
         if (rp->push_rule) hash +=
            (key->MeterId = rp->RuleMatchedValue.byte)*prime_13;
         else hash +=
            (key->MeterId = bp->ntm_interface)*prime_13;
         break;
#endif
#if NEW_ATR && OAM
      case FTOAMIDENT:
         hash += (key->OAMIdent = rp->RuleMatchedValue.byte)*prime_14;
         break;
#endif
#if APP_DATA
      case FTAPPDNSTLD:
         break;
      case FTAPPP2PKIND:
         hash += (key->p2pProto = htons(p2p_test(bp)))*prime_19;
         break;
      case FTAPPATR4:
         break;
      case FTAPPATR3:
         break;
      case FTAPPATR2:
         break;
      case FTAPPATR1:
         break;
#endif
#if NEW_ATR
      case FTSPARE1:
         break;
#endif
         }
      }
#ifdef MATCH_TRACE
   printf("\nType,Class,Kind: %d  %d %d %d  %d %d %d  hash=%d\n",
      key->PeerAddrType,
      key->Low.Class,key->High.Class,key->FlowClass,
      key->Low.Kind,key->High.Kind,key->FlowKind,
      hash);
#endif

   *search_hash = hash % fthashmod;
   }

#ifdef STACK_CHECK
void dump_pat_stack(char *reason)
{
   int j;
   scpos(0,24);
   printf("match %s: stk=", reason);
   for (j = 0; j != RSTKSIZ; ++j)
      printf("%u+%u,", pme->a_stack[j],pme->p_stack[j]);
   printf("\n   LowPeer=");  daddr(search_key.Low.PeerAddress);
   printf(", HighPeer=");  daddr(search_key.High.PeerAddress);
   printf("\n");
   }
#endif

#if PME_TRACE
void log_PME_TRACE()
{
   int ptx;
   for (ptx = 0; ptx <= rtx-1; ptx+=10)
      log_msg(LOG_INFO, 0, "   %d, %d, %d, %d, %d, %d, %d, %d, %d, %d,",
         ruletrace[ptx], ruletrace[ptx+1], ruletrace[ptx+2],
         ruletrace[ptx+3], ruletrace[ptx+4], ruletrace[ptx+5],
         ruletrace[ptx+6], ruletrace[ptx+7], ruletrace[ptx+8],
         ruletrace[ptx+9]);
   }
#endif

int test_hash(int *rx, int *attrib,  /* PME: hash-comparison test */
   struct pkt *bp, struct pkt_key *from, struct pkt_key *to,
   struct rule *rp, struct pme_env *pme,
   struct rtinf_rec *rip)
{
   union address *Maskp = &Masks[rp->RuleMaskVal];
   int deref, j;
   Bit32 hash;
   int ral;  /* Rule address length (bytes) */

   union address masked_value;
   struct rule *rt;
   Bit16 *rht;  /* Ruleset's rule hash tbl */
   struct rule_hash_tbl *rh;

   do {
      deref = 0;
      switch (*attrib) {
      case FTLOWPEERADDRESS:
#if QBYTE_PEER
         hash = (masked_value.qbyte = from->PeerAddress.qbyte
            & Maskp->qbyte)*prime_0;
         ral = 4;
#else
         for (j = hash = 0;  j != pme->pal;  ++j) hash += 
            (masked_value.peer32[j] = from->PeerAddress.peer32[j]
               & Maskp->peer32[j])*primes[j];
         ral = pme->pal*4;
#endif
         break;
      case FTHIPEERADDRESS:
#if QBYTE_PEER
         hash = (masked_value.qbyte = to->PeerAddress.qbyte
            & Maskp->qbyte)*prime_0;
         ral = 4;
#else
         for (j = hash = 0;  j != pme->pal;  ++j) hash += 
            (masked_value.peer32[j] = to->PeerAddress.peer32[j]
               & Maskp->peer32[j])*primes[j];
         ral = pme->pal*4;
#endif
         break;
      case FTLOWPEERTYPE:
      case FTHIPEERTYPE:
         hash = (masked_value.byte = 
            bp->PeerAddrType & Maskp->byte)*prime_1;
         ral = 1;
         break;
      case FTLOWTRANSADDRESS:
         hash = (masked_value.trans = 
            from->TransAddress & Maskp->trans)*prime_2;
         ral=TRANS_ADDR_LEN;
         break;
      case FTHITRANSADDRESS:
         hash = (masked_value.trans = 
            to->TransAddress & Maskp->trans)*prime_2;
         ral=TRANS_ADDR_LEN;
         break;
      case FTLOWTRANSTYPE:
      case FTHITRANSTYPE:
         hash = (masked_value.byte = 
            bp->TransAddrType & Maskp->byte)*prime_3;
         ral = 1;
         break;
      case FTLOWADJACENTADDRESS:
         hash = (masked_value.ma.ms4 =
               from->AdjAddr_ms4 & Maskp->ma.ms4)*prime_4 +
            (masked_value.ma.ls2 =
               from->AdjAddr_ls2 & Maskp->ma.ls2)*prime_5;
         ral = MAC_ADDR_LEN;
         break;
      case FTHIADJACENTADDRESS:
         hash = (masked_value.ma.ms4 =
               to->AdjAddr_ms4 & Maskp->ma.ms4)*prime_4 +
            (masked_value.ma.ls2 =
               to->AdjAddr_ls2 & Maskp->ma.ls2)*prime_5;
         ral = MAC_ADDR_LEN;
         break;
      case FTSOURCECLASS:
         hash = (masked_value.byte = pme->psc & Maskp->byte)*prime_9;
         ral = 1;
         break;
      case FTDESTCLASS:
         hash = (masked_value.byte = pme->pdc & Maskp->byte)*prime_9;
         ral = 1;
         break;
      case FTFLOWCLASS:
         hash = (masked_value.byte = pme->pfc & Maskp->byte)*prime_9;
         ral = 1;
         break;
      case FTSOURCEKIND:
         hash = (masked_value.byte = pme->psk & Maskp->byte)*prime_10;
         ral = 1;
         break;
      case FTDESTKIND:
         hash = (masked_value.byte = pme->pdk & Maskp->byte)*prime_10;
         ral = 1;
         break;
      case FTFLOWKIND:
         hash = (masked_value.byte = pme->pfk & Maskp->byte)*prime_10;
         ral = 1;
         break;
      case FTPAIRINTERFACE:
      case FTLOWINTERFACE:
         hash = (masked_value.byte = 
            from->Interface & Maskp->trans)*prime_7;
         ral = 1;
         break;
      case FTHIINTERFACE:
         hash = (masked_value.byte = 
            to->Interface & Maskp->trans)*prime_7;
         ral = 1;
         break;
      case FTOBSINTERFACE:
         hash = (masked_value.byte = 
            from->ObsInterface & Maskp->trans)*prime_7;
         ral = 1;
         break;
      case FTLOWADJACENTTYPE:
         hash = (masked_value.byte = 
            from->AdjType & Maskp->byte)*prime_6;
         ral = 1;
         break;
      case FTHIADJACENTTYPE:
         hash = (masked_value.byte = 
            to->AdjType & Maskp->byte)*prime_6;
         ral = 1;
         break;
      case FTPRIORITY:
         hash = (masked_value.byte = 
            bp->Priority & Maskp->byte)*prime_16;
         ral = 1;
         break;
      case FTDSCODEPOINT:
         hash = (masked_value.byte = 
            bp->DSCodePoint & Maskp->byte)*prime_8;
         ral = 1;
         break;
      case FTECNCODEBITS:
         hash = (masked_value.byte = 
            bp->ECNCodeBits & Maskp->byte)*prime_14;
         ral = 1;
         break;
      case FTVLANID:
         hash = (masked_value.trans = 
            bp->VLANid & Maskp->trans)*prime_15;
         ral = TRANS_ADDR_LEN;
         break;
#if USER_INFO
      case FTHISUBSCRIBERID:  /* UserCoS */
         hash = (masked_value.byte = 
            uinfo[pme->user_ix].d.UserCoS & Maskp->byte)*prime_17;
         ral = 1;
         break;
      case FTSESSIONID:  /* UserInfo */
         hash = (masked_value.byte = 
            pme->user_attrib & Maskp->byte)*prime_18;
         ral = 1;
         break;
#endif

#if NF_ASN_ATT
      case FTLOWROUTEASN:
         hash = (masked_value.trans = 
            from->nf_ASN & Maskp->trans)*prime_11;
         ral = TRANS_ADDR_LEN;
         break;
      case FTHIROUTEASN:
         hash = (masked_value.trans = 
            to->nf_ASN & Maskp->trans)*prime_11;
         ral = TRANS_ADDR_LEN;
         break;
      case FTLOWROUTEPREFIX:
         hash = (masked_value.byte = 
            from->nf_mask & Maskp->byte)*prime_12;
         ral = 1;
         break;
      case FTHIROUTEPREFIX:
         hash = (masked_value.byte = 
            to->nf_mask & Maskp->byte)*prime_12;
         ral = 1;
         break;
#endif
#if NF_OTHER_ATT
      case FTMETERID:
         hash = (masked_value.byte = 
            bp->ntm_interface & Maskp->byte)*prime_13;
         ral = 1;
         break;
#endif

      case RS_NULL:
      default:
         return 1;  /* Don't test anything */

      case FTV1:
      case FTV2:
      case FTV3:
      case FTV4:
      case FTV5:
         *attrib = pme->user_vbls[rp->RuleSelector-FTV1];
	 deref = 1;
         break;
         }
      } while (deref);

   rt = rip->ri_rule_table - 1;
   rht = rip->ri_rule_hash;
   rh = (struct rule_hash_tbl *)&rht[rp->hash_tbl_index];
   hash %= rh->hashmod;

#ifdef P_TRACE
   ptx = *rx;
#endif
   if ((*rx = rh->hash_ent[hash]) != 0) {
      do {
         rp = &rt[*rx];  /* 1-org */
         if (memcmp((Bit8 *)masked_value.rule,
               rp->RuleMatchedValue.rule, ral) == 0) {
            return 1;  /* Matched */
            }
         } while ((*rx = rp->hash_link) != 0);
      }
   *rx = rh->group_last;  /* Not in group */
#ifdef P_TRACE
   printf("Hash cmp: rx=%d, att=%d, mtch=%d, m_rx=%d, act=%d, ix=%d\n",
      ptx,attrib, 0,*rx, rp->RuleAction,rp->RuleJumpIndex);
#endif
   return 0;
   }

int match(int forward, struct pkt *bp, int rtx, struct pme_env *pme,
   struct pkt_key *from, struct pkt_key *to)
{
   int rx, attrib, j, test_reqd;
   struct rule *rp;

   struct rtinf_rec *rip = &ri[rtx-1];
   struct rule *rt = rip->ri_rule_table - 1;  /* Ruleset, 1-org */
   int rsz = rip->ri_Size;  /* Rules in ruleset */
   Bit16 *rht = rip->ri_rule_hash;  /* Ruleset's rule hash tbl */
   union address *Maskp;
   int match;

#if PME_TRACE
   rtx = 0;  /* Initialise match trace */
#endif
   memset(pme, 0, sizeof(struct pme_env));
   pme->pal = (PEER_ADDR_LEN+3)/4;  pme->forward = forward;
   test_reqd = 1;

   for (rx = 1; rx <= rsz; ) {  /* Search the rule table */
#if PME_TRACE
      if (rtx != MXTRACE) ruletrace[rtx++] = rx;
      else {
         if (tr_msgs_left[rtx-1] != 0) {
            log_msg(LOG_ERR, 0, "PME match loop, ri[%d]: rtx=%d", rtx, rtx); 
            log_PME_TRACE();
            --tr_msgs_left[rtx-1];
	    }
         return MFAIL;
         }
#endif
      rp = &rt[rx];  /* 1-org */
      if ((attrib = rp->RuleSelector) != RS_NULL) {
         if (!test_reqd) {
            if (USER_ATTRIB(attrib))  /* User variable, deref it */
               attrib = pme->user_vbls[rp->RuleSelector-FTV1];
            }
         else {  /* Test required */
            if (rp->hash_tbl_index == 0) {  /* Simple compare */
            Maskp = &Masks[rp->RuleMaskVal];
            for (;;) {  /* May loop to test indirect attribs */
               switch (attrib) {
               case FTLOWTRANSTYPE:
               case FTHITRANSTYPE:
                  match = bp->TransAddrType == rp->RuleMatchedValue.byte;
                  goto test_done;
               case FTLOWPEERTYPE:
               case FTHIPEERTYPE:
                  if (match = bp->PeerAddrType == rp->RuleMatchedValue.byte)
                     pme->pal = addr_len32[bp->PeerAddrType];
                  goto test_done;
               case FTLOWTRANSADDRESS:
                  match = (from->TransAddress & Maskp->trans) ==
                     rp->RuleMatchedValue.trans;
                  goto test_done;
               case FTHITRANSADDRESS:
                  match = (to->TransAddress & Maskp->trans) ==
                     rp->RuleMatchedValue.trans;
                  goto test_done;
               case FTLOWPEERADDRESS:
#if QBYTE_PEER
                  match = (from->PeerAddress.qbyte & Maskp->qbyte) ==
                        rp->RuleMatchedValue.qbyte;
#else
                  for (j = 0;  j != pme->pal;  ++j) {
                     if ((from->PeerAddress.peer32[j] & Maskp->peer32[j]) !=
                           rp->RuleMatchedValue.peer32[j]) {
                        match = 0;  goto test_done;
                        }
                     }
                  match = 1;
#endif
                  goto test_done;
               case FTHIPEERADDRESS:
#if QBYTE_PEER
                  match = (to->PeerAddress.qbyte & Maskp->qbyte) ==
                        rp->RuleMatchedValue.qbyte;
#else
                  for (j = 0;  j != pme->pal;  ++j) {
                     if ((to->PeerAddress.peer32[j] & Maskp->peer32[j]) !=
                           rp->RuleMatchedValue.peer32[j]) {
                        match = 0;  goto test_done;
                        }
                     }
                  match = 1;
#endif
                  goto test_done;
               case FTLOWADJACENTADDRESS:
                  match = ((from->AdjAddr_ms4 & Maskp->ma.ms4) ==
                        rp->RuleMatchedValue.ma.ms4) &&
                     ((from->AdjAddr_ls2 & Maskp->ma.ls2) ==
                        rp->RuleMatchedValue.ma.ls2);
                  goto test_done;
               case FTHIADJACENTADDRESS:
                  match = ((to->AdjAddr_ms4 & Maskp->ma.ms4) ==
                        rp->RuleMatchedValue.ma.ms4) &&
                     ((to->AdjAddr_ls2 & Maskp->ma.ls2) ==
                        rp->RuleMatchedValue.ma.ls2);
                  goto test_done;
               case FTFORWARD:  /* Addresses in packet order */
                  match = pme->forward == rp->RuleMatchedValue.byte;
                  goto test_done;
               case FTSOURCECLASS:
                  match = (pme->psc & Maskp->byte) ==
                     rp->RuleMatchedValue.byte;
                  goto test_done;
               case FTDESTCLASS:
                  match = (pme->pdc & Maskp->byte) ==
                     rp->RuleMatchedValue.byte;
                  goto test_done;
               case FTFLOWCLASS:
                  match = (pme->pfc & Maskp->byte)
                     == rp->RuleMatchedValue.byte;
                  goto test_done;
               case FTSOURCEKIND:
                  match = (pme->psk & Maskp->byte) ==
                     rp->RuleMatchedValue.byte;
                  goto test_done;
               case FTDESTKIND:
                  match = (pme->pdk & Maskp->byte) ==
                     rp->RuleMatchedValue.byte;
                  goto test_done;
               case FTFLOWKIND:
                  match = (pme->pfk & Maskp->byte) ==
                     rp->RuleMatchedValue.byte;
                  goto test_done;
               case FTPAIRINTERFACE:
               case FTLOWINTERFACE:
                  match = from->Interface == rp->RuleMatchedValue.trans;
                     /* Kevin, 23 Nov 98 */
                  goto test_done;
               case FTHIINTERFACE:
                  match = to->Interface == rp->RuleMatchedValue.trans;
                     /* Kevin, 23 Nov 98 */
                  goto test_done;
               case FTOBSINTERFACE:
                  match = from->ObsInterface == rp->RuleMatchedValue.byte;
                  goto test_done;
               case FTLOWADJACENTTYPE:
                  match = from->AdjType == rp->RuleMatchedValue.byte;
                  goto test_done;
               case FTHIADJACENTTYPE:
                  match = to->AdjType == rp->RuleMatchedValue.byte;
                  goto test_done;
               case FTPRIORITY:
                  match = bp->Priority == rp->RuleMatchedValue.byte;
                  goto test_done;
               case FTDSCODEPOINT:
                  match = bp->DSCodePoint == rp->RuleMatchedValue.byte;
                  goto test_done;
               case FTECNCODEBITS:
                  match = bp->ECNCodeBits == rp->RuleMatchedValue.byte;
                  goto test_done;
               case FTVLANID:
                  match = bp->VLANid == rp->RuleMatchedValue.trans;
                  goto test_done;
#if USER_INFO
               case FTHISUBSCRIBERID:  /* UserCoS */
                  match =  uinfo[pme->user_ix].d.UserCoS ==
                     rp->RuleMatchedValue.byte;
                  goto test_done;
               case FTSESSIONID:  /* UserInfo */
                  match =  pme->user_attrib == rp->RuleMatchedValue.byte;
                  goto test_done;
#endif
#if NF_ASN_ATT
               case FTLOWROUTEASN:
                  match = (from->nf_ASN & Maskp->trans) !=
                     rp->RuleMatchedValue.trans;
                  goto test_done;
               case FTHIROUTEASN:
                  match = (to->nf_ASN & Maskp->trans) !=
                     rp->RuleMatchedValue.trans;
                  goto test_done;
               case FTLOWROUTEPREFIX:
                  match = from->nf_mask == rp->RuleMatchedValue.byte;
                  goto test_done;
               case FTHIROUTEPREFIX:
                  match = to->nf_mask == rp->RuleMatchedValue.byte;
                  goto test_done;
#endif
#if NF_OTHER_ATT
               case FTMETERID:
                  match = bp->ntm_interface == rp->RuleMatchedValue.byte;
                  goto test_done;
#endif

               case RS_NULL:
               default:
                  match = 1;  /* Don't test anything */
                  goto test_done;

               case FTV1:
               case FTV2:
               case FTV3:
               case FTV4:
               case FTV5:
                  attrib = pme->user_vbls[rp->RuleSelector-FTV1];
         	 continue;
                  }
               }
            test_done:
               if (!match) {
                  ++rx;  /* Test failed */
                  continue;
                  }
               }
            else {  /* Hashed compare */
               if (!test_hash(&rx, &attrib, bp, from, to, rp, pme, rip)) {
                  ++rx;  /* Test failed */
                  continue;
                  }
               else rp = &rt[rx];  /* Test succeeded */
               }
            }
         }

      test_reqd = 0;
      switch (rp->RuleAction) {
      case RA_IGNORE:
#if PM_DEBUG
         if (rtx == pm_dbg) {
            sprintf(dbg_msg, " f=%d, rx=%d, IGNORE", forward,rx);
            display_msg(0,dbg_msg);
            }
#endif
         return MIGNORE;
      case RA_RETRY:
#if PM_DEBUG
         if (rtx == pm_dbg) {
            sprintf(dbg_msg, " f=%d, rx=%d, RETRY", forward,rx);
            display_msg(0,dbg_msg);
            }
#endif
         return MFAIL;
      case RA_COUNT:
      case RA_COUNTPKT:
         if (attrib != 0 || rp->RuleMaskVal != 0) {  /* Non-zero mask */
               /* Count rule specified an attribute */
            pme->a_stack[pme->p_s_depth] = attrib;  /* Last matched rule */
            pme->p_stack[pme->p_s_depth++] = rx;
            }
#ifdef COUNT_TRACE
/*         if ((from->PeerAddress[2] == 4 && from->PeerAddress[3] == 33) ||
                  (to->PeerAddress[2] == 4 && to->PeerAddress[3] == 33)) */
               {
            printf("Count: %u+%u",pme->a_stack[0],pme->p_stack[0]);
            for (j = 1; j != pme->p_s_depth; ++j) printf(",%u+%u, ",
               pme->a_stack[j],pme->p_stack[j]);
            daddr(from->PeerAddress);  printf(":");
            dport(from->TransAddress);
            printf(" %s ", forward ? "->" : "<-");
            daddr(to->PeerAddress);  printf(":");
            dport(to->TransAddress);  printf("\n");
            }
#endif
#if PM_DEBUG
         if (rtx == pm_dbg) {
            sprintf(dbg_msg, " f=%d, rx=%d, COUNT", forward,rx);
            display_msg(0,dbg_msg);
            }
#endif
         return MSUC;
      case RA_RETURN:
         --pme->e_s_tos;
#ifdef STACK_CHECK
         if (pme->e_s_tos < 0) {
            scpos(0,scr_lrow);
            printf("gosub oflo: depth=%d, stk=", pme->e_s_tos);
            for (j = 0; j != ESTKSIZ; ++j) printf("%u,", pme->env_stack[j]);
            printf("\n");
            return MFAIL;
            }
#endif
#ifdef P_TRACE
         scpos(0,scr_lrow);
         printf("Return: rx=%d, attrib=%d, gosub@%d, index=%d\n",
            rx,rp->RuleSelector, pme->env_stack[pme->e_s_tos],rp->RuleJumpIndex);
#endif
         rx = pme->env_stack[pme->e_s_tos] + rp->RuleJumpIndex;
         continue;
      case RA_ASSIGN:
         test_reqd = 1;
      case RA_ASSIGNACT:
         pme->user_vbls[rp->RuleSelector-FTV1]
            = rp->RuleMatchedValue.byte;  /* Used to be RuleJumpIndex */
         /* ++rx;  Bug!  Nevil, 6 Mar 97 */
         rx = rp->RuleJumpIndex;
         continue;
      case RA_GOSUB:
         test_reqd = 1;
      case RA_GOSUBACT:
         pme->env_stack[pme->e_s_tos++] = rx;  /* Index of gosub rule */
#ifdef STACK_CHECK
         if (pme->e_s_tos >= ESTKSIZ) {
            scpos(0,scr_lrow);
            printf("gosub oflo: depth=%d, stk=", pme->e_s_tos);
            for (j = 0; j != pme->e_s_tos; ++j) printf("%u,", pme->env_stack[j]);
            printf("\n");
            return MFAIL;
            }
#endif
         rx = rp->RuleJumpIndex;
         continue;
      case RA_GOTO:
         test_reqd = 1;
      case RA_GOTOACT:
         rx = rp->RuleJumpIndex;
#ifdef TESTING
         scpos(0,scr_lrow);  printf("   GOTO %d\n", rx);
#endif
         continue;
      case RA_PUSHTO:
         test_reqd = 1;
      case RA_PUSHTOACT:
         switch (attrib) {  /* Save 'packet' computed attrib value */
         case FTSOURCECLASS:
            pme->psc = rp->RuleMatchedValue.byte;  break;
         case FTDESTCLASS:
            pme->pdc = rp->RuleMatchedValue.byte;  break;
         case FTFLOWCLASS:
            pme->pfc = rp->RuleMatchedValue.byte;  break;
         case FTSOURCEKIND:
            pme->psk = rp->RuleMatchedValue.byte;  break;
         case FTDESTKIND:
            pme->pdk = rp->RuleMatchedValue.byte;  break;
         case FTFLOWKIND:
            pme->pfk = rp->RuleMatchedValue.byte;  break;
#if USER_INFO
         case FTSESSIONID:  /* UserInfo */
            switch (pme->user_attrib = rp->RuleMatchedValue.byte) {
            case FTLOWPEERADDRESS:
               pme->user_ix = UserInfoLookup(&from->PeerAddress);
               break;
            case FTHIPEERADDRESS:
               pme->user_ix = UserInfoLookup(&to->PeerAddress);
               break;
               }
# if USER_DUMMY_IDS
            if (!pme->user_ix) {
               union ui_row ui;
               memset(ui.ui_bytes, 0, sizeof(ui));
               ui.d.Operation = UI_START;
               ui.d.UserCoS = 1;
               switch (pme->user_attrib) {
               case FTLOWPEERADDRESS:
                  memcpy(&ui.d.key_val,
                     &from->PeerAddress.byte, PEER_ADDR_LEN);
                  break;
               case FTHIPEERADDRESS:
                  memcpy(&ui.d.key_val,
                     &to->PeerAddress.byte, PEER_ADDR_LEN);
                  break;
                  }
/*               sprintf((char *)&ui.d.UserID, "U%03d", n_users);
                  /* 4 chars for IPv4 testing */
               memset(&ui.d.UserID, 0, sizeof(ui.d.UserID));
               ui.d.UserID.qba[0] = htonl(n_users);
                  /* 32-bit integer, network byte order */
               ui.d.UserID.ba[SESSION_ID_LEN-1] = '*';
	       UserInfoUpdate(pme->user_ix, UI_START, &ui);
               pme->user_ix = n_users-1;

	       printf("User %d registered:", n_users-1);
	       for (j = 0; j != sizeof(ui); ++j)
		 printf(" %02x", ui.ui_bytes[j]);
               printf("\n");  fflush(stdout);
               }
/* test to see that these get through to fdf OK ..
   bp->VLANid = htons(0x0FFF - n_users);
   bp->Priority = n_users & 0x0F;
   printf("VLAN/Priority set\n");  /**/
# endif
            break;
#endif
            }
         pme->a_stack[pme->p_s_depth] = attrib;
            /* Remember which rules matched */
         pme->p_stack[pme->p_s_depth++] = rx;
         rx = rp->RuleJumpIndex;
#ifdef STACK_CHECK
         if (pme->p_s_depth >= RSTKSIZ) {
            dump_pat_stack("oflo");
            return MFAIL;
            }
#endif
         continue;
      case RA_PUSHPKTTO:
         test_reqd = 1;
      case RA_PUSHPKTTOACT:
         pme->a_stack[pme->p_s_depth] = attrib;
            /* Remember which rules matched */
         pme->p_stack[pme->p_s_depth++] = rx;
         rx = rp->RuleJumpIndex;
#ifdef STACK_CHECK
         if (pme->p_s_depth >= RSTKSIZ) {
            dump_pat_stack("oflo");
            return MFAIL;
            }
#endif
         continue;
      case RA_POPTO:
         test_reqd = 1;
      case RA_POPTOACT:
         --pme->p_s_depth;  /* Forget last matched rule */
         rx = rp->RuleJumpIndex;
#ifdef STACK_CHECK
         if (pme->p_s_depth < 0) {
            dump_pat_stack("uflo");
            return MFAIL;
            }
#endif
         continue;

       default:  /* Not a valid action! */
         log_msg(LOG_ERR, 0, 
            "Bad action, ruleset %d, rule %d", rtx,rx);
#if PME_TRACE
         log_PME_TRACE();
#endif
         return MFAIL;
         }
      }
   log_msg(LOG_ERR, 0, "Bad index, ruleset %d, rule %d", rtx,rx);
#if PME_TRACE
   log_PME_TRACE();
#endif
   return MFAIL;  /* Not matched */
   }

void show_flow_hash(void)
{
   int j, inuse_slots,  entries, cl, hc_len[51], mx_cl;
   struct flow **sf, *f;

   inuse_slots = entries = 0;
   memset(hc_len, 0, sizeof(hc_len));
   for (j = 0; j != fthashmod; ++j) {
      if (flow_ht[j] != NULL) {
         ++inuse_slots;  ++entries;
         sf = &flow_ht[j];  f = *sf;
         for (cl = 1; ; ++cl, ++entries) {
            if ((f = f->ht_next) == (struct flow *)sf) break;
	    }
         ++hc_len[cl > sizeof(hc_len)/sizeof(int) ? 
            sizeof(hc_len)/sizeof(int)-1 : cl-1];
         }
      }

   printf("current() hash performance:\n");
   printf("   %d hash slots, %d in use (%.2f %%)\n",
      fthashmod, inuse_slots, inuse_slots*100.0/fthashmod);

   for (mx_cl = sizeof(hc_len)/sizeof(int)-1; mx_cl != 0; --mx_cl)
      if (hc_len[mx_cl] != 0) break;
   if (mx_cl >= 0) {
      ++mx_cl;
      printf("   %d entries in table, average %.2f entries/inuse_slot\n",
         entries, (float)entries/(inuse_slots == 0 ? 1 : inuse_slots));
      printf("   chain lengths:");
      for (j = 0; j != mx_cl; ++j) printf(" %d", hc_len[j]);
      printf(" (max %d)\n", mx_cl);
      }
   }

struct flow *current(
   struct search_result *s_r,
   Bit32 hash, struct flow_key *search_key)
{
   struct flow **sf;
   struct flow *f, *lf;
   Bit32 *qp, *hqp;
   int j;

   sf = &flow_ht[hash];  /* Address of hash table entry */
#ifdef MATCH_TRACE
   printf("  current(): cht=%Fp, sf=%Fp\n", cht,sf);
#endif
   ++n_hash_searches;
   if ((f = sf[0]) == NULL) {
      lf = NULL;  /* Empty hash chain */
      ++n_hash_compares;  /* Make sure compares >= searches! */
      }
   else {
      do {
         ++n_hash_compares;
         for (;;) {  /* Compare whole fields - data has been masked */
            for (hqp = (Bit32 *)&f->fk, qp = (Bit32 *)search_key, j = 0;
                  j != sizeof(struct flow_key)/sizeof(Bit32);  ++j) {
               if (*hqp++ != *qp++) goto s_no_match;
               }
            return f;  /* Matched */
            }

      s_no_match:
         lf = f;
         } while ((f = lf->ht_next) != (struct flow *)sf);
      f = NULL;  /* Loop stops with lf -> last flow in chain */
      }
   s_r->sf = sf;  s_r->lf = lf;
      /* Save sf and lf; we'll need them to insert a new flow */
#ifdef MATCH_TRACE
   printf("  current() -> sf=%Fp, lf=%Fp\n", s_r->sf,s_r->lf);
#endif
   return f;
   }

int get_flow_stdata(struct flow *f)
{
   int r = 1;
   mutex_lock("gfs", &flow_table_mutex);
   if (f->stdata == NULL) {
      if ((f->stdata = get_st_data()) == NULL)
         r = 0;  /* No free stdata blocks */
      }
   mutex_unlock("gfs", &flow_table_mutex);
   return r;
   }

int make_distrib_list(  /* From pattern stack */
   struct pme_env *pme, struct flow *f, struct stream *sp)
{
   int c_rtx = f->FlowRuleSet;
   struct rule *c_rt = ri[c_rtx-1].ri_rule_table;
   struct rule *rp;
   int j, a, t, st, del;
   struct distribution *r, *d, *ld;
   double N,D;
   char m_name[20];

   if (f->distrib_list != NULL) {
      scpos(0,scr_lrow);
      log_msg(LOG_ERR, 0, "Non-empty distribution list <<<");
      return 0;
      }

   for (j = 0; j != pme->p_s_depth; ++j) {
      if (stream_atr[a = pme->a_stack[j]]) {   /* Stream attribute */
         if (get_flow_stdata(f) == 0)
            return 0;
         }
      if (dist_atr[a]) {  /* Distribution-valued attribute */
#if PP_DEBUG_XX == 1
         log_msg(LOG_WARNING, 0, 
            "make_distrib_list(1): flow_nbr=%d, attrib=%d", flow_nbr(f), a);
#endif
         rp = &c_rt[pme->p_stack[j]-1];
         if ((d = get_dist()) == NULL) 
            return 0;
         memset(d, 0, sizeof(struct distribution));
#if MULTI
	 sprintf(m_name, "dyn_dst_%u_%u", flow_nbr(f), a);
         init_mutex(&d->dyn_dst_mutex, m_name);
#endif
         d->selector = a;
         memcpy(d->mask_params, &Masks[rp->RuleMaskVal], RULE_ADDR_LEN);
         memcpy(d->value_params, rp->RuleMatchedValue.rule, RULE_ADDR_LEN);

         d->Transform = d->mask_params[0];
         d->ScaleFactor = d->mask_params[1];
         d->lowerspec = d->LowerLimit =
            (d->mask_params[2] << 8) | d->mask_params[3];
         d->upperspec = d->UpperLim = d->UpperLimit =
            (d->mask_params[4] << 8) | d->mask_params[5];
         d->Scale = LO_SCALE(d->ScaleFactor);
         del = HI_SC_DELTA(d->ScaleFactor);
         if (del != 0) d->UpperLim = times_pwr_of_10(d->UpperLimit, del);

         d->Buckets = d->value_params[0];
         if (d->Buckets > MXBUCKETS) {
            scpos(0,scr_lrow);
            log_msg(LOG_ERR, 0,
               "Distribution with > %d buckets **", MXBUCKETS);
            d->Buckets = d->value_params[0] = MXBUCKETS;
	    }
         d->Parameter1 = d->value_params[1];
         d->Parameter2 = (d->value_params[2] << 8) | d->value_params[3];
         d->Parameter3 = (d->value_params[4] << 8) | d->value_params[5];

         N = d->Buckets - 1;
         switch (d->Transform) {
         case DS_LIN_RESET:
            d->dyn_dist = 1;
         case DS_LIN:
            D = d->UpperLim - d->LowerLimit;
            d->M = N/D;
            break;
         case DS_LOG_RESET:
            d->dyn_dist = 1;
         case DS_LOG:
            D = log10((double)d->UpperLim/(double)d->LowerLimit);
            d->M = N/D;
            break;
         case DS_DYN_REQ:
         case DS_DLOG_REQ:
            d->dyn_dist = 1;
            d->max_val = d->lowerspec;
            d->min_val = d->upperspec;
            d->n_values = 0;
            break;
            }

         if (f->distrib_list == NULL)
            f->distrib_list = d;
         else ld->next = d;
         ld = d;

         f->distrib_bits |= 1 << d->selector-FTTOPACKETSIZE;
         d->fp = f;  d->nrate = NULL;

         switch (a) {  /* Special handling for particular d-v attributes */
         case FTTOBITRATE:
         case FTFROMBITRATE:
         case FTTOPDURATE:
         case FTFROMPDURATE:
	    if (d->RateInterval != 0) link_distrib_to_event(d);
            else log_msg(LOG_WARNING, 0, "Rate distribution with zero interval !!!");
            break;

         case FTTOTCPSIZE:             /* TCP attributes: */
         case FTFROMTCPSIZE:           /*    (param1 = stream-size scale) */ 
         case FTTOTCPTIME:             /*    Must build streams, */
         case FTFROMTCPTIME:           /*    don't care about pp_type */
         case FTTOTCPRATE1:
         case FTFROMTCPRATE1:
         case FTTOTCPRATE2:
         case FTFROMTCPRATE2:
	   if (f->stdata != NULL)
               f->stdata->pp_match_reqd = 1;
            break;

         case FTTOBYTEDENSITY:
         case FTFROMBYTEDENSITY:
         case FTTOPKTDENSITY:
         case FTFROMPKTDENSITY:

         case FTTOSTREAMS:    /* 'To' CPB flows */
         case FTFROMSTREAMS:  /* 'From' CPB flows */
         case FTFLOWTIME:     /* NeTraMet streams */
# if STREAM_TO_FLOW
            st = d->Parameter2*100;  /* sf timeout, cs */
#  if STF_DEBUG
	    log_msg(LOG_WARNING, 0, "STF(1): flow=%x, sftime=%d (cs)",
               f, st);
#  endif
            if (st != 0) f->sftime = st;  /* Use last non-zero value */
# endif
            break;

         case FTTOINTERARRIVALTIME:    /* Inter-packet time attributes */
         case FTFROMINTERARRIVALTIME:
#if 0
         case FTTOTRAINLENGTH:         /* In train if <= FilterUpper */
         case FTFROMTRAINLENGTH:       /*   (and > FilterLower) */
#endif
         case FTTOTURNAROUNDTIME4:
         case FTFROMTURNAROUNDTIME4:

         case FTTOTURNAROUNDTIME1:     /* Turnaround time attributes: */
         case FTFROMTURNAROUNDTIME1:   /*    Must build streams, */
                                       /*    (param1 = pp_type selector) */  
	    if (d->PP_Type == PP_NONE) {
               f->stdata->use_flow_iats = 1;  /* Build flow iat distribution */
	       break;  /* No matching, just inter-packet times */
               }
            else if (d->PP_Type == PP_OTHER) {
               f->stdata->use_flow_iats = 0;  /* Build streams iat distribution */
	       break;  /* No matching, just inter-packet times */
               }
            else {
               f->stdata->pp_match_reqd = 1;
               }

            if (f->fk.PeerAddrType != AT_IP4
#if V6
                  && f->fk.PeerAddrType != AT_IP6
#endif
                  ) {
               log_msg(LOG_WARNING, 0, "PP test on non-IP flow !!!");
               break;
               }
            if ((d->PP_Type & PP_TCP) == PP_TCP) {
               if (f->fk.TransAddrType != PT_TCP) {
                  log_msg(LOG_WARNING, 0, "PP TCP test on non-TCP flow !!!");
                  break;
	          }
	       }
            else switch (d->PP_Type) {
            case PP_OTHER:    /* Don't try to match */
               break;
            case PP_UDP_DNS:
               if (f->fk.Low.TransAddress != htons(WNP_DOMAIN) &&
                     f->fk.High.TransAddress != htons(WNP_DOMAIN)) {
                  log_msg(LOG_WARNING, 0, 
                     "PP DNS test on non-DNS port (%04x->%04x) !!!",
                     f->fk.Low.TransAddress, f->fk.High.TransAddress);
                  break;
	          }
               if (f->fk.TransAddrType != PT_UDP) {
                  log_msg(LOG_WARNING, 0, "PP DNS test on non-UDP flow !!!");
                  break;
	          }
               break;
            case PP_ICMP_ECHO:
               if (f->fk.TransAddrType != PT_ICMP) {
                  log_msg(LOG_WARNING, 0, "PP ping test on non-ICMP flow !!!");
                  break;
	          }
               break;

            case PP_ICMP_TIMSTMP:
            case PP_UDP_ECHO:
            case PP_UDP_SNMP:
               break;
            default:
               log_msg(LOG_WARNING, 0, "Unknown PP value for flow !!!");
	       }
            break;
            }
         }
#if PP_DEBUG_XX == 1
      log_msg(LOG_WARNING, 0, 
         "make_distrib_list(2): flow_nbr=%d, bits=%x",
         flow_nbr(f), f->distrib_bits);
#endif

#if OAM
      if (a == FTOAMDATA) {
         if (get_flow_stdata(f) == 0)
            return 0;  /* No free stdata blocks */
         if (f->fk.OAMIdent != 0)
            log_msg(LOG_WARNING, 0, 
               "Flow %d has OAMIdent=%d; it can't be an OAM trigger <<<",
               flow_nbr(f), f->fk.OAMIdent);
         else {
            rp = &c_rt[pme->p_stack[j]-1];
            t = rp->RuleMatchedValue.rule[0];
            if (t > MXOAMIDS || t < 1)
               log_msg(LOG_WARNING, 0, 
                  "Flow %d: OAMtrigger must be from 1 to %d <<<",
                  flow_nbr(f), MXOAMIDS);
            else {
               f->stdata->OAMtrigger = t;
# if OAM_TRACE
               printf("Creating OAM trigger, flow=%d, trigger=%d\n",
                  flow_nbr(f),  f->stdata->OAMtrigger);
               fflush(stdout);
# endif
               }
            }
         }
#endif
      }
   return 1;
   }

#if TCP_TESTING || PKT_QUEUE_TEST || TCP_TAT_TRACE
void print_stream_name(FILE *f, char *msg, 
   struct stream *sp)
{
   fprintf(f, "str(%s) ", msg);
   fdaddr(f, sp->sfn.SrcPeerAddr);  fprintf(f, ", ");
   fdport(f, sp->sfn.SrcTransAddr);  fprintf(f, " -> ");
   fdaddr(f, sp->sfn.DestPeerAddr);  fprintf(f, ", ");
   fdport(f, sp->sfn.DestTransAddr);
   fprintf(f, ", TransType=%d, rs=%d: ", sp->sfn.TransType, sp->sfn.RuleSet);
   }
#endif

#if SCH_CHECK
int check_stream_hash(struct stream *isp, struct stream *l_n_h, char *msg)
{
   int n = 0;
   struct stream *sp, *lsp;
   Bit8 *np;  int j;  Bit32 hash;
   struct stream **sha;
   int sha_seen;

   mutex_lock("csh", &strhash_q_mutex);
   for (np = (Bit8 *)&isp->sfn, hash = 0, j = 0;
          j != sizeof(isp->sfn); ++np, ++j)
      /* hash += *np * primes[j]; */
      hash = (hash << 4) ^ (hash >> 28) ^ *np;
   sha = &sfht[hash % sfhashmod];

   if (isp->next_hc == (struct stream *)sha && l_n_h == isp) {
     mutex_unlock("csh", &strhash_q_mutex);
     return 1;  /* Just removed lone entry from hash chain */
     }

   for (sha_seen = 0, sp = isp; ; ) {
      if (sp->next_hc == NULL
	  /*	  && !(l_n_h == isp && sp == (struct stream *)sha) */
   ) {
         printf("check_stream_h(%s): next_hc == NULL, isp=%d, lsp=%d, sp=%d\n",
            msg, stream_nbr(isp), stream_nbr(lsp), stream_nbr(sp));
         mutex_unlock("csh", &strhash_q_mutex);
 quack(101);
         return 0;
         }
      lsp = sp;
      if ((sp = lsp->next_hc) == isp)
         /* Loop stops with lsp -> last stream in chain */
         break;
      if (sp == (struct stream *)sha) {
         if (sha_seen) break;  /* Entry removed leaving loop from sha */
         sha_seen = 1;
         }
      if (++n == 20) {
         printf("check_stream_h(%s): runaway next_hc chain, isp=%d\n",
            msg, stream_nbr(isp));
 quack(102);
         }
      }
   mutex_unlock("csh", &strhash_q_mutex);
   return 1;
   }
#endif  /* SCH_CHECK */

#if STHC_CHECK
int sh_ok(struct stream *sp, char *msg)  /* Verify stream hash chain is ok */
{
   struct stream *tp, *ltp;
   int n = 0;
   tp = sp;  do {
      if (tp == NULL) quack(999);
      if (++n > 8) quack(111);
      ltp = tp;
      } while ((tp = tp->next_hc) != sp);
   return 1;
   }
#else
#define sh_ok(x, y)
#endif

#if STH_CHECK
void check_stream_state(struct stream *sp,
   int in_ht, int exp_state, char *msg)
{
   if (sp->next_hc == NULL && in_ht || sp->next_hc != NULL && !in_ht) {
      printf("stream_state(1): sp->next_hc=%x, expected %d. %s\n",
	     sp->next_hc, in_ht, msg);  
      fflush(stdout);  quack(201);
      }
   if (sp->state != exp_state) {
      printf("stream_state(2): sp->in_state_q=%d, expected=%d. %s\n",
	     sp->state,exp_state, msg);  
      fflush(stdout);  quack(202);
      }
   }
#endif

int lost_to,lost_from;  /* Globals for update_LostPDUs */

void update_LostPDUs(struct flow *fp, struct stream *sp)
{
   struct stream_data *sdp;
   if ((sdp = fp->stdata) != NULL  /* Flow has distributions */
         && sdp->pp_match_reqd) {
      /* pktdatas still in queue are considered lost! */
      sdp->ToLostPDUs += lost_to;
      sdp->FromLostPDUs += lost_from;
      }
   }

void terminate_stream(struct stream *sp)
   /* Frees stream's pktdata blocks, removes it from stream hash table */
{
   struct pktdata *pk, *npk;
   struct stream *np, *tp, *ltp;
   struct stream_data *sdp;
   struct flow *fp;
#if SCH_CHECK
   struct stream *l_n_h;
#endif

#if STH_CHECK
   if (sp->next_hc == NULL) {
      quack(1111);
      return;  /* Already terminated! */
      }
#endif

#if STREAM_TO_FLOW
   if (sp->str_flowp != NULL) {  /* Stream has created a flow */
      sp->str_flowp->flow_strp = NULL;
      sp->str_flowp->sf_state = SF_INACTIVE;
         /* Flow can be garbage collected */
      }
#endif

   mutex_lock("ts", &pktdata_q_mutex);
   for (lost_to = lost_from = 0, pk = sp->pq; pk != NULL; pk = npk) {
      npk = pk->next;  /* Free stream's pktdata blocks */
      if (pk->direction == MFWD) ++lost_to;
      else ++lost_from;
      free_pktdata(pk);
      }
   mutex_unlock("ts", &pktdata_q_mutex);
   if (lost_to + lost_from != 0)
      update_flows_for_stream(sp, update_LostPDUs);

   if (sp->state == SS_MATCHED || sp->state == SS_IGNORE
         || sp->state == SS_UNKNOWN) {
//20      mutex_lock("ts", &strhash_q_mutex);
      np = sp->next_hc;  /* Remove stream from hash chain */
      tp = sp;  do {
         ltp = tp;
         } while ((tp = tp->next_hc) != sp);
#if SCH_CHECK
   l_n_h = ltp->next_hc;
#endif
      if (ltp == np)  /* Single flow in hash chain */
#if SCH_CHECK
         {
         check_stream_hash(sp, l_n_h, "term-middle");  /* ??? */
         np->next_hc = NULL;
         }
#else
         np->next_hc = NULL;
#endif
      else ltp->next_hc = np;
      sp->next_hc = NULL;
      sp->state = SS_TERM;
//20      mutex_unlock("ts", &strhash_q_mutex);
      }
   else {
      log_msg(LOG_WARNING, 0,
         "Couldn't terminate stream %x, state=%d, next_hc=%x!",
         sp, sp->state, sp->next_hc);
quack(86);
      }
#if SCH_CHECK
   check_stream_hash(sp, l_n_h, "term-after");  /* ??? */
#endif
   }

int unmatch_flow(struct flow *fp, int update, int match, struct stream *sp)
{
  if (match) {
      if (update) flow_stream_stats(fp, sp);
      return 1;
      }
   else return 0;
   }

int um_ruleset_nbr;  /* Global for unmatch_ruleset */

int unmatch_ruleset(struct flow *fp, int update, int match, struct stream *sp)
{
   if (fp->FlowRuleSet == um_ruleset_nbr) {
      if (update) flow_stream_stats(fp, sp);
      return 1;
      }
   else return 0;
   }

void unmatch_flows_for_stream(struct stream *sp,
   int (*unmatch_flow)(struct flow *fp, int u, int x, struct stream *sp))
{
   int j,k, f,r;
   struct flow *fp;
   Bit32 fx;

   for (f = j = 0; j != sp->f_pmi.n_flows; ++j) {  /* Forward flows */
     fp = flow_ptr(fx = sp->f_pmi.fx[j]);
     if (!unmatch_flow(fp, 1, sp->f_pmi.fx[j] == fx, sp))
         sp->f_pmi.fx[f++] = sp->f_pmi.fx[j];
      }
   sp->f_pmi.n_flows = f;
   for (r = j = 0; j != sp->r_pmi.n_flows; ++j) {  /* Reverse flows */
      fp = flow_ptr(fx = sp->r_pmi.fx[j]);
      for (k = 0; k != sp->f_pmi.n_flows; ++k) {
         if (sp->f_pmi.fx[k] == fx) break;
         }
      if (!unmatch_flow(fp, k == sp->f_pmi.n_flows,
                            /* Update if not also a forward flow */
            sp->r_pmi.fx[j] == flow_nbr(fp), sp))
         sp->r_pmi.fx[r++] = sp->r_pmi.fx[j];  /* Reverse flow to keep */
      }
   sp->r_pmi.n_flows = r;
   sp->state = f+r == 0 ? SS_UNKNOWN : SS_MATCHED;
   }

void unmatch_flow_streams(struct flow *fp)
   /* Takes fp out of any stream matched-flow lists it appears in */
{
   struct stream_data *sdp;
   struct stream *sp, *lsp;
   int j,k;

   if ((sdp = fp->stdata) == NULL)
      return;
   else if (sdp->active_streams == 0)
      return;  /* No streams */

   mutex_lock("tas", &stream_mutex);
//20   mutex_lock("tas", &stream_state_mutex);
//20   mutex_lock("tas", &stream_q_mutex);
#if SCC_CHECK > 1
   check_stream_chains("term_all_streams(before)");
#endif
   if ((sp = t_sq) != NULL) {
      do {
         lsp = sp;  sp = sp->next_sp;
         unmatch_flows_for_stream(lsp, unmatch_flow);
         } while (sp != NULL);
      }
#if SCC_CHECK
   check_stream_chains("term_all_streams: after");
#endif
//20   mutex_unlock("tas", &stream_q_mutex);
//20   mutex_unlock("tas", &stream_state_mutex);
   mutex_unlock("tas", &stream_mutex);
   }

void unmatch_ruleset_streams(int rs)
   /* Moves rs's streams from MATCHED and IGNORE back to UNKNOWN */
{
   struct stream *sp, *lsp;

   mutex_lock("tas2", &stream_mutex);
//20   mutex_lock("tas2", &stream_state_mutex);
//20   mutex_lock("tas2", &stream_q_mutex);
#if SCC_CHECK > 1
   check_stream_chains("term_all_streams(before)");
#endif
   if ((sp = t_sq) != NULL) {
      um_ruleset_nbr = rs;
      do {
         lsp = sp;  sp = sp->next_sp;
         unmatch_flows_for_stream(lsp, unmatch_ruleset);
         } while (sp != NULL);
      }
#if SCC_CHECK
   check_stream_chains("term_all_streams: after");
#endif
//20   mutex_unlock("tas2", &stream_q_mutex);
//20   mutex_unlock("tas2", &stream_state_mutex);
   mutex_unlock("tas2", &stream_mutex);
   }

void unmatch_all_streams(void)
   /* Moves all streams from MATCHED and IGNORE back to UNKNOWN,
         doesn't bump stream-based distrbutions */
{
   struct stream *sp, *lsp;
   int j;

   mutex_lock("tas3", &stream_mutex);
//20   mutex_lock("tas3", &stream_state_mutex);
//20   mutex_lock("tas3", &stream_q_mutex);
#if SCC_CHECK
   check_stream_chains("term_all_streams(before)");
#endif
   if ((sp = t_sq) != NULL) {
      do {
         lsp = sp;  sp = sp->next_sp;
         lsp->f_pmi.n_flows = 0;
         lsp->r_pmi.n_flows = 0;
         lsp->state = SS_UNKNOWN;
         } while (sp != NULL);
      }
#if SCC_CHECK > 1
   check_stream_chains("term_all_streams: after");
#endif
//20   mutex_unlock("tas3", &stream_q_mutex);
//20   mutex_unlock("tas3", &stream_state_mutex);
   mutex_unlock("tas3", &stream_mutex);
   }

#if LOG_STREAM_DATA
FILE *streams_log = NULL;

int small_total = 0, small_tcp = 0, small_udp = 0, small_icmp = 0;

void log_streams(struct stream *sp, double lifetime)
{
   int have_StreamData;

   if (streams_log == NULL) {
      if ((streams_log = fopen("streams.log", "w")) == NULL)
         log_msg(LOG_ERR, 1, "Couldn't open streams.log <<<<<");
      fprintf(streams_log, "## NeTraMet streams_log\n"
         "#Format: sourcetranstype  topdus frompdus tooctets fromoctets"
         "  lasttime lifetime <tcpstreamdata>\n"
         "#Note: lifetime in ms, %.3f\n");
      }

# if LOG_STREAM_DATA == 1
# define STREAM_MIN_PDUS  5
   if (sp->ToPDUs + sp->FromPDUs < STREAM_MIN_PDUS) {
      switch (sp->sfn.TransType) {
      case PT_TCP: ++small_tcp;
         break;
      case PT_UDP: ++small_udp;
         break;
      case PT_ICMP: ++small_icmp;
         break;
         }
      ++small_total;
      }
   else {
      if (small_total > 25) {
         fprintf(streams_log, "#SC: %u %u %u %u\n",
            small_total, small_tcp, small_udp, small_icmp);
         small_total = small_tcp = small_udp = small_icmp = 0;
         }
      fprintf(streams_log, "%u %u  %u %u %u %u  %u %u\n",
         sp->sfn.TransType, sp->flowp->fk.FlowKind, 
         sp->ToPDUs, sp->FromPDUs, sp->ToOctets, sp->FromOctets,
         (Bit32)((sp->StrFirstTime+lifetime)/10000.0),
         (Bit32)(lifetime/1000.0));  /* LastTime TimeTicks, lifetime ms */
      }

# else  /* Log RTT-data TCP streams */
   if (lifetime < 1000.0)  /* Ignore streams < 1 ms lifetime */
      return;

   if (sp->sfn.TransType != PT_TCP)  /* Not TCP, no RTT data */
      have_StreamData = 0;
   else
      have_StreamData = !(sp->ToSynRTT == 0 && sp->FromSynRTT == 0 &&
        sp->ToDataPoints == 0 &&  sp->FromDataPoints == 0);

   if (!have_StreamData) return;

   fprintf(streams_log, " %u  %u %u %u %u  %u %u",
      sp->sfn.TransType,
      sp->ToPDUs, sp->FromPDUs, sp->ToOctets, sp->FromOctets,
      (Bit32)((sp->StrFirstTime+lifetime)/10000.0), (Bit32)(lifetime/1000.0));
      /* LastTime TimeTicks, lifetime ms */

   fprintf(streams_log, " <%u %u %u %u %u %u>\n",
      sp->ToSynRTT,sp->FromSynRTT, sp->ToDataPoints,sp->FromDataPoints,
      sp->ToDataPoints == 0 ? 0 :
         (Bit32)(sp->ToTotalRTT/sp->ToDataPoints),
      sp->FromDataPoints == 0 ? 0 :
         (Bit32)(sp->FromTotalRTT/sp->FromDataPoints));  /* us */
# endif
   }
#endif

void flow_stream_stats(struct flow *fp, struct stream *sp)
   /* Computes flow-based statistics for terminating streams */
{
   struct distribution *d;
   double delta;

//--   printf("1", flow_nbr(fp));  //--@@

   for (d = fp->distrib_list; d != NULL; d = d->next) {
      switch (d->selector) {
      case FTTOFLOWOCTETS:
         bump_dist(d, sp->ToOctets);
         break;
      case FTFROMFLOWOCTETS:
         bump_dist(d, sp->FromOctets);
         break;
      case FTTOFLOWPDUS:
         bump_dist(d, sp->ToPDUs);
         break;
      case FTFROMFLOWPDUS:
         bump_dist(d, sp->FromPDUs);
         break;
      case FTTOBYTEDENSITY:
         if (sp->ToSeen) {
            delta = sp->LastToTime - sp->FirstToTime;
            accum_dist(d, delta, sp->ToOctets);
            }
         break;
      case FTTOPKTDENSITY:
         if (sp->ToSeen) {
            delta = sp->LastToTime - sp->FirstToTime;
            accum_dist(d, delta, sp->ToPDUs);
            }
         break;
      case FTTOSTREAMS:
         if (sp->ToSeen) {
            delta = sp->LastToTime - sp->FirstToTime;
            bump_dist(d, delta);  /* us */
            }
         break;
      case FTFROMBYTEDENSITY:
         if (sp->FromSeen) {
            delta = sp->LastFromTime - sp->FirstFromTime;
            accum_dist(d, delta, sp->FromOctets);
            }
         break;
      case FTFROMPKTDENSITY:
         if (sp->FromSeen) {
            delta = sp->LastFromTime - sp->FirstFromTime;
            accum_dist(d, delta, sp->FromPDUs);
            }
         break;
      case FTFROMSTREAMS:
         if (sp->FromSeen) {
            delta = sp->LastFromTime - sp->FirstFromTime;
            bump_dist(d, delta);  /* us */
            }
         break;
      case FTFLOWTIME:
             /* Use highest estimate for total stream time */
//--   printf("2", flow_nbr(fp));  //--@@
         if (sp->ToSeen && sp->FromSeen) {
            delta = sp->LastFromTime > sp->LastToTime ?
               sp->LastFromTime : sp->LastToTime;
            delta -= sp->StrFirstTime;
            }
         else if (sp->ToSeen)  /* Only 'to' packets seen */
            delta = sp->LastToTime - sp->StrFirstTime;
         else if (sp->FromSeen)  /* Only 'from' packets seen */
            delta = sp->LastFromTime - sp->StrFirstTime;
         else continue;  /* Didn't see any packets! */
//--   printf("3", flow_nbr(fp));  //--@@
         bump_dist(d, delta);  /* us */
#if LOG_STREAM_DATA
         log_streams(sp, delta);
#endif
         break;
         }
      }
   }

void tcp_stream_stats(struct flow *fp, struct stream *sp)
   /* Computes TCP-based statistics for terminating streams */
{
#if TCP_PKT_DATA
   struct distribution *d;
   Bit32 bytes, b2;
   double delta;
   int n;

   for (d = fp->distrib_list; d != NULL; d = d->next) {
      switch (d->selector) {
      case FTTOTCPSIZE:
      case FTTOTCPTIME:
      case FTTOTCPRATE1:
      case FTTOTCPRATE2:
         bytes = sp->ToSeqOctets > sp->FromAckOctets ? 
            sp->ToSeqOctets : sp->FromAckOctets;
         break;
      case FTFROMTCPSIZE:
      case FTFROMTCPTIME:
      case FTFROMTCPRATE1:
      case FTFROMTCPRATE2:
         bytes = sp->FromSeqOctets > sp->ToAckOctets ? 
            sp->FromSeqOctets : sp->ToAckOctets;
         break;
         }
      if (bytes == 0) continue;  /* No data for stream! */

      if (d->FilterLower != 0 || d->FilterUpper != 0) {
         n = scale(d, bytes);  /* Divide by 10**FilterScale */
# define TCPTIME_TEST 0
# if TCPTIME_TEST
         if (d->selector == FTFROMTCPTIME)
            printf(" TCPFromTime: FromB=%lu|ToB=%lu, n=%u, L=%u|U=%u\n",
               bytes,b2, n, d->FilterLower,d->FilterUpper);
# endif
         if (d->FilterLower != 0 && n <= (Bit32)d->FilterLower) continue;
         if (d->FilterUpper != 0 && n > (Bit32)d->FilterUpper) continue;
         }

          /* Use highest estimate for total stream time */
      if (sp->ToSeen && sp->FromSeen) {
         delta = sp->LastFromTime > sp->LastToTime ?
            sp->LastFromTime : sp->LastToTime;
         delta -= sp->StrFirstTime;
         }
      else if (sp->ToSeen)  /* Only 'to' packets seen */
         delta = sp->LastToTime - sp->StrFirstTime;
      else if (sp->FromSeen)  /* Only 'from' packets seen */
         delta = sp->LastFromTime - sp->StrFirstTime;
      else continue;  /* Didn't see any packets! */

      switch (d->selector) {
      case FTTOTCPSIZE:
      case FTFROMTCPSIZE:
         bump_dist(d, bytes);  /* Bytes */
         break;
      case FTTOTCPTIME:
      case FTFROMTCPTIME:
         bump_dist(d, delta);  /* us */
         break;
      case FTFROMTCPRATE1:
      case FTFROMTCPRATE2:
         if (delta == 0) delta = 1;  /* Don't divide by 0 */
         bump_dist(d, bytes*1000000.0/delta);  /* B/s */
         break;
      case FTTOTCPRATE1:
      case FTTOTCPRATE2:
         if (delta == 0) delta = 1;  /* Don't divide by 0 */
         bump_dist(d, bytes*1000000.0/delta);  /* B/s */
         break;
         }
      }

# if TCP_TESTING
         print_stream_name(stdout, "fin", sp);
         printf(": active=%d, To %lu/%lu, From %lu/%lu",
            str->active_streams,
            str->ToTCPSentOctets.low,str->ToTCPSeqOctets.low,
            str->FromTCPSentOctets.low,str->FromTCPSeqOctets.low);
            printf("\n");)
# endif
#endif  /* TCP_PKT_DATA */
   }

void update_stream_stats(struct flow *fp, struct stream *sp)
{
   if (fp->stdata != NULL) {
      ++fp->stdata->n_streams;
      if (sp->ToPDUs != 0 && sp->FromPDUs != 0)
	{  quack(222);
         ++fp->stdata->n_twoway_streams;
	}  //30
      if (sp->sfn.TransType == PT_TCP) ++fp->stdata->n_tcp_streams;
      else if (sp->sfn.TransType == PT_UDP) ++fp->stdata->n_udp_streams;
      }

   if (fp->distrib_list) {
      flow_stream_stats(fp, sp);
      if (fp->fk.TransAddrType == PT_TCP ) {
         tcp_stream_stats(fp, sp);
         if (fp->stdata != NULL)
            --fp->stdata->terminating_streams;
         }
      }
   }

int check_idle_streams(void)
   /* Streams time out after being idle for OTHER_STREAM_TIMEOUT cs */
{
   struct stream *sp, *nsp;
   Bit32 uc, IDLE_time, pkts;
   struct flow *fp;
   double av_int;
   Bit32 sto, inact;
   int j, keep, terminated;
#if STR_TO_TRACE
   int n_streams, n_kept, pq;
#endif

   uc = OTHER_STREAM_TIMEOUT;  /* Centiseconds */
   if (last_t_cs <= uc)
      return;  /* Avoid -ve IDLE_time (nothing can have aged out yet) */

   IDLE_time = last_t_cs - uc;  /* centiseconds */
#if STR_TO_TRACE
   n_streams = n_kept = pq = 0;
#endif
#if STR_TO_TRACE > 1
   printf("o_i(%d) now=%d, active=%d: ",
      IDLE_time, uptime_cs(), t_active_streams);
#endif

   mutex_lock("o_is", &stream_mutex);
//20   mutex_lock("o_is", &stream_state_mutex);
//20   mutex_lock("o_is", &stream_q_mutex);
      /* Tried moving lock/unlock inside stream check loop.
         Dismal failure, got lots of lctr=0xFFFF messages!
                                        Nevil, 17 Aug 03 */
   terminated = 0;

#if SCC_CHECK > 1
   check_stream_chains("check_idle_streams: before");
#endif
   for (sp = t_sq;  ; ) {
      if (sp == NULL) break;
      nsp = sp->next_sp;  /* Next stream to check */
#if STR_TO_TRACE
      ++n_streams;
#endif

#if SCC_CHECK > 1
   check_stream_chains("check_idle_streams: during");
#endif
      if (sp->LastTime > IDLE_time) {
         break;  /* Hasn't been queued long enough */
         }
      else {  /* Could be inactive, can we recover it? */
         if (sp->n_pkts_queued != 0) {
            keep = 1;
#if STR_TO_TRACE
            ++pq;
#endif
            }
         else if ((pkts = sp->ToPDUs+sp->FromPDUs - 1) == 0)
            keep = 0;  /* Only saw one packet */
         else if (sp->FirstTime < sp->LastTime) {
            av_int = (double)(sp->LastTime - sp->FirstTime)/(double)pkts;
            sto = av_int*STO_MULTIPLIER;
            inact = last_t_cs - sp->LastTime;
#if STR_TO_TRACE
            if (keep = (inact < sto)) ++n_kept;
#else
            keep = (inact < sto);
#endif
            }
	 else {  /* Trace files can have odd time jumps in them */
            keep = 0;  /* No data to improve sto estimate */
            }
         }

      if (nsp == NULL) {  /* Remove stream from queue */
         if (t_active_streams != 1) {
            log_msg(LOG_WARNING, 0,
               "NULL link within stream chain!");
 quack(111);
//20            mutex_unlock("o_is", &stream_q_mutex);
//20            mutex_unlock("o_is", &stream_state_mutex);
            mutex_unlock("o_is", &stream_mutex);
            return;
            }
         t_sq = t_sq_tail = NULL;
         }
      else {
         t_sq = nsp;  nsp->prev_sp = NULL;
         }
      --t_active_streams;

      if (keep) {
         if (t_sq == NULL) {  /* Add it to tail of queue */
            /* We won't waste time tesing it until it drifts back
               to the head of the queue.  That means we no longer
               have the queue in (approximnate) LastTime order */
            t_sq = sp;  sp->prev_sp = NULL;
            }
         else {
            t_sq_tail->next_sp = sp;  sp->prev_sp = t_sq_tail;
            }
         t_sq_tail = sp;  sp->next_sp = NULL;
         ++t_active_streams;
         }
      else {  /* Timed-out stream */
#if LOG_P2P_TAGS
         if (sp->p2pProto != 0)
            log_p2p(sp, NULL, NULL);  /* End of stream */
#endif

         update_flows_for_stream(sp, update_stream_stats);
         terminate_stream(sp);  /* Sets sp->state = SS_TERM */
#if STH_CHECK
   check_stream_state(sp, 0,SS_TERM, "check_idle(term)");
#endif
#if STR_ANAL
         if (sp->hlp != NULL) {
            if (sp->hlp->n_hosts >= 10)
               log_hostlist(sp, 2);
            free_hlist(sp->hlp);
            }
         else if (sp->sfn.PeerType < AT_P2P) {  /* Ordinary stream */
            update_host_list(0, 1, sp);
            update_host_list(1, 1, sp);
            }
#endif
         ++terminated;
         free_stream(sp);
#if SCC_CHECK > 1
     check_stream_chains("check_o_i_s: after 1");
#endif
/* printf("66 other_idle() timed out sp=%d, pkts=%d\n",
      stream_nbr(sp), sp->ToPDUs+sp->FromPDUs);  /**/
         }
      sp = nsp;            
      }

#if SCC_CHECK
   check_stream_chains("check_idle_streams: after");
#endif

//20   mutex_unlock("o_is", &stream_q_mutex);
//20   mutex_unlock("o_is", &stream_state_mutex);
   mutex_unlock("o_is", &stream_mutex);
#if STR_TO_TRACE > 1
   printf("%d tested, %d kept, %d pq\n", n_streams,n_kept,pq);
#endif
   return terminated;
   }

#define MX_SEQ_DELTA   65535  /* Max TCP window size */
/* #define MX_SEQ_DELTA   29200  /* 20 Ethernet TCP packets */

void tcp_stream_update(struct count_env *cep, 
   struct stream *sp, const struct pktdata *pd)
   /* Updates TCP state for flows of stream sp.
      pd points to packet in stream queue (if we saved it),
      or to packet in free queue (if we didn't) */
{
#if TCP_PKT_DATA
   int j, flow_order = pd->direction;
   Bit32 tlen = pd->pi.tcp.d_len;
   Bit32 Bytes;

   if (pd->pi.tcp.flags & tcp_FlagRST)
      sp->flags_seen |= RSTseen;
   else if (flow_order == MREV) {
      if (pd->pi.tcp.flags & tcp_FlagSYN) sp->flags_seen |= FromSYNseen;
      if (pd->pi.tcp.flags & tcp_FlagFIN) sp->flags_seen |= FromFINseen;
      }
   else {
      if (pd->pi.tcp.flags & tcp_FlagSYN) sp->flags_seen |= ToSYNseen;
      if (pd->pi.tcp.flags & tcp_FlagFIN) sp->flags_seen |= ToFINseen;
      }

   if (!sp->terminating && (sp->flags_seen & RSTseen ||
         (sp->flags_seen & BothFINseen) == BothFINseen)) {
      sp->terminating = 1;
      }

   if (!sp->terminating) {
      if (flow_order == MREV) {

         if (sp->FromSeq_seen) {
            if (scmp32_ge(pd->pi.tcp.exp_seq, sp->FromLastSeq)) {  /* Normal */
               Bytes = pd->pi.tcp.exp_seq - sp->FromLastSeq;
               if (Bytes <= MX_SEQ_DELTA) {
                  sp->FromSeqOctets += Bytes;
                  }
               else {
# if TRACE_SEQ_DECR
                  print_pkt_trace(pt_file, 
                     "FROM-SEQ", sp, ": SeqBytes=%lu <<<", Bytes);
# endif
                  }
               sp->FromLastSeq = pd->pi.tcp.exp_seq;
               }
            }
         else if (pd->pi.tcp.exp_seq != 0) {
               /* tlen data bytes in this packet */
            sp->FromSeqOctets = tlen;
            sp->FromLastSeq = pd->pi.tcp.exp_seq;
            sp->FromSeq_seen = 1;
            }

         if (sp->ToAck_seen) {
            if (scmp32_ge(pd->pi.tcp.ack, sp->ToLastAck)) {
               Bytes = pd->pi.tcp.ack - sp->ToLastAck;
               sp->ToAckOctets += Bytes;
               }
            sp->ToLastAck = pd->pi.tcp.ack;
            }
         else if (pd->pi.tcp.ack != 0) {
              /* Assume these will be acked (!) */
            sp->ToAckOctets = tlen;
            sp->ToLastAck = pd->pi.tcp.ack;
            sp->ToAck_seen = 1;
            }
         }
      else {  /* MFWD */
         if (sp->ToSeq_seen) {
            if (scmp32_ge(pd->pi.tcp.exp_seq, sp->ToLastSeq)) {  /* 'Normal' */
               Bytes = pd->pi.tcp.exp_seq - sp->ToLastSeq;
               if (Bytes <= MX_SEQ_DELTA)
                  sp->ToSeqOctets += Bytes;
               else {
# if TRACE_SEQ_DECR
                  print_pkt_trace(pt_file, 
                     "TO-SEQ", sp, ": SeqBytes=%lu <<<", Bytes);
# endif
                  }
               sp->ToLastSeq = pd->pi.tcp.exp_seq;
               }
            }
         else if (pd->pi.tcp.exp_seq != 0) {
               /* tlen data bytes in this packet */
            sp->ToSeqOctets = tlen;
            sp->ToLastSeq = pd->pi.tcp.exp_seq;
            sp->ToSeq_seen = 1;
            }

         if (sp->FromAck_seen) {
            if (scmp32_ge(pd->pi.tcp.ack, sp->FromLastAck)) {
               Bytes = pd->pi.tcp.ack - sp->FromLastAck;
               sp->FromAckOctets += Bytes;
               } 
            sp->FromLastAck = pd->pi.tcp.ack;
            }
         else if (pd->pi.tcp.ack != 0) {
            sp->FromAckOctets = tlen;
               /* Assume these will be acked (!) */
            sp->FromLastAck = pd->pi.tcp.ack;
            sp->FromAck_seen = 1;
            }
         }
      }

# if 0
      if (flow_order == MFWD) {
         sp->FromWindow = pd->pi.tcp.send_win = ntohs(bp->pi.tcp.th.window);
         pd->pi.tcp.recv_win = sp->ToWindow;
         }
      else {
         sp->ToWindow = pd->pi.tcp.send_win = ntohs(bp->pi.tcp.th.window);
         pd->pi.tcp.recv_win = sp->FromWindow;
         }
# endif

# if TCP_TRACE
      fprintf(stderr,
      "%5d %9.5f %6.3f %s %s%s len=%4d, ack=%lu, seq=%lu, flg=%04x\n", 
         sp->sf_nbr,
         microseconds(bp->arrival_time)/1000000.0, 
         (microseconds(bp->arrival_time)-
            (flow_order == MREV ? sp->at.LastToTime : sp->at.LastFromTime)
            )/1000000.0,
         flow_order == MREV ? "<-" : "->",
	 flow_order != sp->at.LastDirection ? " " : "=",
         sp->at.TimeOKrev ? "R" : " ",
         tlen, ThisAck, ThisSeq, ThisFlags
         );
# endif
#endif  /* TCP_PKT_DATA */
   }

void tcp_flow_update(struct count_env *cep, struct flow *fp, 
   struct stream *sp, const struct pktdata *pd)
   /* Updates TCP counters for flow fp.
      pd points to current packet's data */
{
#if  TCP_PKT_DATA
   struct stream_data *sdp = fp->stdata;
   int j, flow_order = pd->direction;
   Bit32 tlen = pd->pi.tcp.d_len;
   Bit32 Bytes;

   if (!sp->terminating && (sp->flags_seen & RSTseen ||
         (sp->flags_seen & BothFINseen) == BothFINseen)) {
      ++sdp->terminating_streams;
      }

   if (pd->direction == MFWD) {
      sdp->ToLostPDUs += cep->lostPDUs;
      sdp->ToPQOverflows += cep->ofloPDUs;
      }
   else {
      sdp->FromLostPDUs += cep->lostPDUs;
      sdp->FromPQOverflows += cep->ofloPDUs;
      }
   if (!sp->terminating) {
      if (flow_order == MREV) {
         if (sp->FromSeq_seen) {
            if (scmp32_ge(pd->pi.tcp.exp_seq, sp->FromLastSeq)) {  /* Normal */
               Bytes = pd->pi.tcp.exp_seq - sp->FromLastSeq;
               if (Bytes <= MX_SEQ_DELTA)
                  add_Bit32_to_counter64(sdp->FromTCPSeqOctets, Bytes);
               }
            else ++sdp->FromTCPDecrSeq;  /* Backward jump in seq */
            }
         else if (pd->pi.tcp.exp_seq != 0)
            add_Bit32_to_counter64(sdp->FromTCPSeqOctets, tlen);

         if (sp->ToAck_seen) {
            if (scmp32_ge(pd->pi.tcp.ack, sp->ToLastAck)) {
               Bytes = pd->pi.tcp.ack - sp->ToLastAck;
               add_Bit32_to_counter64(sdp->ToTCPAckOctets, Bytes);
               }
            }
         else if (pd->pi.tcp.ack != 0)
            add_Bit32_to_counter64(sdp->ToTCPAckOctets, tlen);

         if (tlen > 0)
            add_Bit32_to_counter64(sdp->FromTCPLenOctets, tlen);

#define  SU_DEBUG  0
#if SU_DEBUG
printf("  MREV(%u): seq=%u, len=%u, Ack=%u -> FSS=%d, FLen=%u, Fseq=%u, TAS=%d, Tack=%u\n",
    sdp, pd->pi.tcp.exp_seq, tlen, pd->pi.tcp.ack,
    sp->FromSeq_seen, sdp->FromTCPLenOctets.low, sdp->FromTCPSeqOctets.low, 
    sp->ToAck_seen, sdp->ToTCPAckOctets.low);
#endif
         }
      else {  /* MFWD */
         if (sp->ToSeq_seen) {
            if (scmp32_ge(pd->pi.tcp.exp_seq, sp->ToLastSeq)) {  /* 'Normal' */
               Bytes = pd->pi.tcp.exp_seq - sp->ToLastSeq;
               if (Bytes <= MX_SEQ_DELTA)
                  add_Bit32_to_counter64(sdp->ToTCPSeqOctets, Bytes);
               }
            else ++sdp->ToTCPDecrSeq;  /* Backward jump in seq */
            }
         else if (pd->pi.tcp.exp_seq != 0)
            add_Bit32_to_counter64(sdp->ToTCPSeqOctets, tlen);

         if (sp->FromAck_seen) {
            if (scmp32_ge(pd->pi.tcp.ack, sp->FromLastAck)) {
               Bytes = pd->pi.tcp.ack - sp->FromLastAck;
               add_Bit32_to_counter64(sdp->FromTCPAckOctets, Bytes);
               } 
            sp->FromLastAck = pd->pi.tcp.ack;
            }
         else if (pd->pi.tcp.ack != 0)
            add_Bit32_to_counter64(sdp->FromTCPAckOctets, tlen);

         if (tlen > 0)
            add_Bit32_to_counter64(sdp->ToTCPLenOctets, tlen);

#if SU_DEBUG
printf("MFWD(%u): seq=%u, len=%u, Ack=%u -> TSS=%d, TLen=%u, Tseq=%u, FAS=%d, Fack=%u\n",
    sdp, pd->pi.tcp.exp_seq, tlen, pd->pi.tcp.ack,
    sp->ToSeq_seen, sdp->ToTCPLenOctets.low, sdp->ToTCPSeqOctets.low, 
    sp->FromAck_seen, sdp->FromTCPAckOctets.low);
#endif
         }
      }
#endif  /* TCP_PKT_DATA */
   }

FILE *pt_file;
#define ptf_title "ntm.trace"

#if TCP_TAT_TRACE
void print_tcp_pktdata(struct pktdata *pd)
{
   printf("%.3f %s %c%c%c%c%c %lu:%lu(%u) ack %lu, pkt %u\n",
      pd->ts/1000.0,
      pd->direction == MFWD ? "->" : "<-",
      pd->pi.tcp.flags & tcp_FlagACK  ? 'A' : ' ',
      pd->pi.tcp.flags & tcp_FlagPUSH ? 'P' : ' ',
      pd->pi.tcp.flags & tcp_FlagRST  ? 'R' : ' ',
      pd->pi.tcp.flags & tcp_FlagSYN  ? 'S' : ' ',
      pd->pi.tcp.flags & tcp_FlagFIN  ? 'F' : ' ',
      pd->pi.tcp.exp_seq - pd->pi.tcp.d_len,
      pd->pi.tcp.exp_seq,  pd->pi.tcp.d_len,
     pd->pi.tcp.ack, pd->pi.tcp.pkt_nbr);
   /*      pd->pi.tcp.send_win, pd->pi.tcp.recv_win); */
   }

void print_tcp_list(struct pktdata *pd, char *msg)
{
   int n = 0;
   printf("   %s:\n",msg);
   for (; pd != NULL; pd = pd->next) {
      printf("      ");  print_tcp_pktdata(pd);
      }
   }
#endif  /* TCP_TAT_TRACE */

#define PKT_QUEUE_TEST 0

#if PKT_QUEUE_TEST
int q_len(struct pktdata *pq)
{
   struct pktdata *fp;  int n;
   for (n = 0, fp = pq; fp != NULL; fp = fp->next) ++n;
   return n;
   }
#endif

int tcp_turnaround(struct count_env *cep,
   struct stream *sp, struct pktdata *tp)
   /* Compares incoming TCP packet with those in stream's packet queue */
{
#if !TCP_PKT_DATA
  return 0;
#else
   struct pktdata *pd, *lpd, *npd;
   int acked, resent, delete, prev_delete, overlap;
   Bit32 a,b,c,d;
   int flow_order = tp->direction;
   int delta_OK = 0;
   int TaT_OK = 0;
/*
   tp = new packet
   pd = queue of old packets, kept in ascending exp_seq order

   For opposite-direction pkts:
      if ack matches exp_seq, set TaT_OK flags
      Delete pkts with seq <= ack
   For same-direction packets:
      Test for overlaps in seq:
         Overlap -> bump lostPDUs
   Save new pkt if it has data, SYN or FIN,
      i.e. if it affects exp_ack
   */

#if PKT_QUEUE_TEST
   int ql = q_len(sp->pq);
   if (ql != sp->pq_len) {
      print_stream_name(stdout, "tcp_ta(1)", sp);
      printf(" ql=%d, pq_len=%d\n", ql,sp->pq_len);
      }
#endif
   b = tp->pi.tcp.exp_seq;  a = b - tp->pi.tcp.d_len;  /* For overlap test */
   acked = resent = prev_delete = 0;
   cep->type_pp = PP_NO_TEST;
   cep->lostPDUs = cep->ofloPDUs = 0;
   mutex_lock("tt1", &pktdata_q_mutex);
   for (pd = sp->pq; pd != NULL; ) {
      delete = 0;
      if (pd->direction != flow_order) {  /* Opposite-direction packet */
         if ((tp->pi.tcp.flags & tcp_FlagACK) == tcp_FlagACK &&  /* Ack OK */
               tp->pi.tcp.ack == pd->pi.tcp.exp_seq &&  /* Ack matched */
               !acked) {  /* Only ack the oldest copy! */
            if ((pd->pi.tcp.flags & (tcp_FlagACK | tcp_FlagSYN))
                     == tcp_FlagSYN &&
                  (tp->pi.tcp.flags & (tcp_FlagACK | tcp_FlagSYN))
                     == (tcp_FlagACK | tcp_FlagSYN)) {
	       TaT_OK |= PP_OK_SYNACK;
	       }
            else if ((pd->pi.tcp.flags & (tcp_FlagACK | tcp_FlagSYN))
                     == (tcp_FlagACK | tcp_FlagSYN) &&
                  (tp->pi.tcp.flags & (tcp_FlagACK | tcp_FlagSYN))
                     == tcp_FlagACK) {
	       TaT_OK |= PP_OK_SYNDATA;
               }
            if (pd->pi.tcp.d_len != 0) {  /* Acking sent data */
               if (prev_delete)  /* Acking 2nd (or later) packet of queue */
                  TaT_OK |= PP_OK_MULTI;
               else if (pd->next == NULL) /* Only one pkt in queue */
                  TaT_OK |= PP_OK_SINGLE;
	       else TaT_OK |= PP_OK_INGROUP;  /* Other pkts in queue */
	       }
            if (TaT_OK) {  /* Compute turnaround time */
               cep->us_delta = tp->ts - pd->ts;  delta_OK = 1;
               }
#  if TCP_TAT_TRACE == 1
            if (TaT_OK & PP_OK_SYN) {
               print_stream_name(stdout, "SYN/RST", sp);
               printf("\n   TaT_OK=%02x, delta_OK=%d, delta=%.0f us\n", 
                  TaT_OK, delta_OK, cep->us_delta);
               }
#  endif
            acked = delete = 1;  /* Delete acked packet */
	    }
         else if (tp->pi.tcp.ack > pd->pi.tcp.exp_seq)
	    delete = 1;  /* Delete pkts with exp_seq < ack */
         }
      else {  /* Same-direction packet */
         /*               a---b    New packet */
         /*     c---------d        Old packet */
         d = pd->pi.tcp.exp_seq;  c = d - pd->pi.tcp.d_len;
         overlap = pd->pi.tcp.flags & tp->pi.tcp.flags &
               (tcp_FlagSYN | tcp_FlagFIN) ?  /* Two SYN/FIN packets */
            a <= d && b >= c :  /* SYN/FIN packets may not overlap at all */
            a < d && b > c;  /* Data packets may overlap by one byte */
         if (overlap) {
#  if TCP_TAT_TRACE == 2
            printf("--- c=%lu,d=%lu, a=%lu,b=%lu, overlap=%d, resent=%d\n", 
               c,d, a,b, overlap, a ==c && b == d);
            print_stream_name(stdout, "TCP overlap", sp);  printf("\n      ");
            print_tcp_pktdata(tp);
            print_tcp_list(sp->pq.packet, "Overlap queue");  printf("\n");
#  endif
            if (b >= d) {   /* Don't count 'lower-part-only' overlap */
               if (!resent) {  /* Only count resent packet once! */
                  ++cep->lostPDUs;
                  resent = (a == c && b == d);
                  }
	       }
            }
         }
      if (delete) {
         prev_delete = 1;  /* At least one packet deleted from queue */
         npd = pd->next;
         if (pd == sp->pq) sp->pq = npd;
         else lpd->next = npd;
         --sp->pq_len;
         free_pktdata(pd);
         pd = npd;
#if PKT_QUEUE_TEST
   ql = q_len(sp->pq);
   if (ql != sp->pq_len) {
      print_stream_name(stdout, "tcp_ta(2)", sp);
      printf(" ql=%d, pq_len=%d\n", ql,sp->pq_len);
      }
#endif
         }
      else {
         lpd = pd;  pd = pd->next;
         }
      }
   mutex_unlock("tt1", &pktdata_q_mutex);
#if PKT_QUEUE_TEST
   ql = q_len(sp->pq);
   if (ql != sp->pq_len) {
      print_stream_name(stdout, "tcp_ta(3)", sp);
      printf(" ql=%d, pq_len=%d\n", ql,sp->pq_len);
      }
#endif

   if (tp->pi.tcp.d_len != 0 ||  /* Enqueue packet if it affects seq */
         tp->pi.tcp.flags & (tcp_FlagSYN | tcp_FlagFIN)) {
      mutex_lock("tt2", &pktdata_q_mutex);
      if ((pd = sp->pq) == NULL) {  /* Empty queue */
         sp->pq = tp;  tp->next = NULL;
         }
      else if (pd->direction == flow_order &&
            tp->pi.tcp.exp_seq < pd->pi.tcp.exp_seq) {
         sp->pq = tp;  tp->next = pd;  /* Enqueue at head */
         }
      else for (; ; pd = pd->next) {
         if ((npd = pd->next) == NULL) {  /* Enqueue at tail */
            pd->next = tp;  tp->next = NULL;
            break;
            }
	 if (pd->direction == flow_order &&
               tp->pi.tcp.exp_seq < npd->pi.tcp.exp_seq) {
            pd->next = tp;  tp->next = npd;
            break;
            }
         }
      if (++sp->pq_len > sp->mx_pq_len) {
         for (pd = sp->pq; pd != NULL; ) {  /* Discard pkts from head of queue */
            ++cep->lostPDUs;  ++cep->ofloPDUs;
            npd = pd->next;  free_pktdata(pd);
            if (--sp->pq_len <= sp->mx_pq_len) break;
            pd = npd;
            }
         sp->pq = npd;
         }
      mutex_unlock("tt2", &pktdata_q_mutex);

#  if TCP_TAT_TRACE == 3
      if (sp->pq_len >= MX_PQ_LEN_TCP) {
         print_stream_name(stdout, "TCP queue too big", sp); 
         printf("\n      ");
         print_tcp_pktdata(tp);
         print_tcp_list(sp->pq.packet, "Queue");  printf("\n");
         }
#  endif
#if PKT_QUEUE_TEST
      ql = q_len(sp->pq);
      if (ql != sp->pq_len) {
         print_stream_name(stdout, "tcp_ta(4)", sp);
         printf(" ql=%d, pq_len=%d\n      ", ql,sp->pq_len);
#  if TCP_TAT_TRACE == 3
         print_tcp_pktdata(tp);
         print_tcp_list(sp->pq, "Queue");  printf("\n");
#  endif
         }
#endif
      }
   else free_pktdata(tp);  /* Don't keep packets which don't affect seq */

   cep->type_pp = PP_TCP | TaT_OK;
   return delta_OK;
#endif  /* TCP_PKT_DATA */
   }

void show_stream_hash(int display)
{
# define N_HCLENS     60
# define N_FILEKINDS  16
   int j, inuse_slots,  entries, cl, hc_len[N_HCLENS+1], mx_cl;
   struct stream **sha, *sp;
   struct pkt_match_info *pmi;
   struct flow *fp;
#define STREAM_KINDS  0
#if STREAM_KINDS
   double av_int;
   Bit32 sto, inact, pkts, keep, fk, lifetime;

   int str_count[N_FILEKINDS+1];  /* One for each FlowKind value */
   int str_1m[N_FILEKINDS+1], str_2m[N_FILEKINDS+1], str_5m[N_FILEKINDS+1],
      expired[N_FILEKINDS+1];
   memset(str_count, 0, sizeof(str_count));
   memset(str_1m, 0, sizeof(str_1m));
   memset(str_2m, 0, sizeof(str_2m));
   memset(str_5m, 0, sizeof(str_5m));
   memset(expired, 0, sizeof(expired));
#endif

   mx_cl = inuse_slots = entries = 0;
   memset(hc_len, 0, sizeof(hc_len));
   for (j = 0; j != sfhashmod; ++j) {
      if (sfht[j] != NULL) {
//20         mutex_lock("ssh", &strhash_q_mutex);
         mutex_lock("ssh", &stream_mutex);
         sha = &sfht[j];  sp = *sha;
         for (cl = 1; ; ++cl, ++entries) {
#if STREAM_KINDS
            if (display) {
               pmi = &sp->f_pmi;
               if (pmi->n_flows != 0) {
                  fp = flow_ptr(pmi->fx[0]);
                  fk = fp->fk.FlowKind;
	          if (fk > N_FILEKINDS) fk = N_FILEKINDS;  /* Oflo bucket */
                  lifetime = sp->LastTime - sp->FirstTime;
                  }
               else fk = lifetime = 0;
               ++str_count[fk];
               if (lifetime > 6000) {
                  if (sp->FirstTime < sp->LastTime) {
                     if ((pkts = sp->ToPDUs + sp->FromPDUs) == 0) pkts = 1;
                     av_int = (double)(sp->LastTime - sp->FirstTime)/(double)pkts;
                     sto = av_int*STO_MULTIPLIER;
                     inact = sp->LastTime - sp->LastTime;
                     keep = inact < sto;
                     }
	          else {  /* Trace files can have odd time jumps in them */
                     keep = 0;  /* No data to improve sto estimate */
                     }
                  if (!keep) ++expired[fk];
                  }
               if (lifetime > 30000)      ++str_5m[fk];   /* 300 s */
               else if (lifetime > 12000) ++str_2m[fk];   /* 120 s */
               else if (lifetime >  6000) ++str_1m[fk];   /*  60 s */
               }
#endif
            if ((sp = sp->next_hc) == (struct stream *)sha) break;
               /* Loop stops with lsp -> last stream in chain */
	    }
//20         mutex_unlock("ssh", &strhash_q_mutex);
         mutex_unlock("ssh", &stream_mutex);
         ++inuse_slots;  ++entries;
         if (cl > mx_cl) {
            if (cl > N_HCLENS) cl = mx_cl = N_HCLENS;
            else mx_cl = cl;
            }
         ++hc_len[cl];
         }
      }

   if (display) {
      printf("lookup_stream_helper() hash performance:\n");
      printf("   %d hash slots, %d in use (%.2f %%)\n",
         sfhashmod, inuse_slots, inuse_slots*100.0/sfhashmod);

      if (mx_cl >= 0) {
         ++mx_cl;
         printf("   %d entries in table, average %.2f entries/inuse_slot\n",
            entries, (float)entries/(inuse_slots == 0 ? 1 : inuse_slots));
         printf("   chain lengths:");
         for (j = 1; j <= mx_cl; ++j) printf(" %d", hc_len[j]);
         printf(" (max %d)\n", mx_cl);
         }
#if STREAM_KINDS
      for (mx_cl = sizeof(str_count)/sizeof(int)-1; mx_cl >=0; --mx_cl)
         if (str_count[mx_cl] != 0) break;
      if (mx_cl >= 0) {
         ++mx_cl;
         printf("   Streams per FlowKind value: count 1m 2m 5m expired\n");
         for (j = 0; j <= N_FILEKINDS; ++j) {
            if (str_count[j] != 0)
               printf("      %2d  %7d %7d %6d %6d  %5d\n", j, str_count[j],
                  str_1m[j], str_2m[j], str_5m[j], expired[j]);
            }
         }
#endif
      }
   else log_msg(LOG_INFO, 0,
      "stream_hash: %d slots, %d in use, %d entries",
      sfhashmod, inuse_slots, entries);
   }

void lookup_stream_helper(struct stream **spp, int *wire_order,
   struct pkt *p_p, int *create_new)
{
   Bit32 hash;
#if !QBYTE_PEER
   Bit32 hs, hd, *qp;  int j;
#endif
   int wo;
   struct stream **sha;
   struct stream *sp, *lsp, *nsp;
   struct stream_data *sdp;

   if (p_p->has_ports) {
#if QBYTE_PEER
      hash = ((p_p->Low.TransAddress * 16381) ^ p_p->Low.PeerAddress.qbyte) ^
         ((p_p->High.TransAddress * 16381) ^ p_p->High.PeerAddress.qbyte) ^
         (p_p->TransAddrType * 211 + p_p->PeerAddrType * 251);
#else
      hs = p_p->Low.TransAddress * 16381;
      qp = p_p->Low.PeerAddress.peer32;
      for (j = 0;  j != (PEER_ADDR_LEN+3)/4;  ++qp, ++j)
         hs = (hs << 4) ^ (hs >> 28) ^ *qp;
      hd = p_p->High.TransAddress * 16381;
      qp = p_p->High.PeerAddress.peer32;
      for (j = 0;  j != (PEER_ADDR_LEN+3)/4;  ++qp, ++j)
         hd = (hd << 4) ^ (hd >> 28) ^ *qp;
      hash = hs ^ hd ^ (p_p->TransAddrType * 211 + p_p->PeerAddrType * 251);
#endif
      }
   else {
#if QBYTE_PEER 
      hash = p_p->Low.PeerAddress.qbyte ^ p_p->High.PeerAddress.qbyte ^ 
         (p_p->TransAddrType * 211 + p_p->PeerAddrType * 251); 
#else 
      hs = 0;
      qp = p_p->Low.PeerAddress.peer32; 
      for (j = 0;  j != (PEER_ADDR_LEN+3)/4;  ++qp, ++j) 
          hs = (hs << 4) ^ (hs >> 28) ^ *qp; 
      hd = 0;
      qp = p_p->High.PeerAddress.peer32; 
      for (j = 0;  j != (PEER_ADDR_LEN+3)/4;  ++qp, ++j) 
         hd = (hd << 4) ^ (hd >> 28) ^ *qp; 
      hash = hs ^ hd ^ (p_p->TransAddrType * 211 + p_p->PeerAddrType * 251); 
#endif 
      }

//20   if (lock) mutex_lock("lsh", &stream_state_mutex);
//20   mutex_lock("lsh", &strhash_q_mutex);
   sha = &sfht[hash % sfhashmod];
      /* Address of stream hash table entry */
   if ((sp = *sha) == NULL) {
      wo = 1;
      lsp = NULL;  /* Empty hash chain */
      }
   else {
      sh_ok(sp,"lsh: before");
      for (;;) {
         wo = 1;
#if QBYTE_PEER
         if (p_p->Low.PeerAddress.qbyte == sp->sfn.SrcPeerAddr &&
               p_p->High.PeerAddress.qbyte == sp->sfn.DestPeerAddr &&
#else
         if (memcmp(p_p->Low.PeerAddress.peer32, 
                  sp->sfn.SrcPeerAddr, PEER_ADDR_LEN) == 0 &&
               memcmp(p_p->High.PeerAddress.peer32, 
                  sp->sfn.DestPeerAddr, PEER_ADDR_LEN) == 0 &&
#endif
               ( p_p->has_ports ?
                  p_p->Low.TransAddress == sp->sfn.SrcTransAddr &&
                  p_p->High.TransAddress == sp->sfn.DestTransAddr
                  : 1 ) &&
               p_p->TransAddrType == sp->sfn.TransType &&
               p_p->PeerAddrType == sp->sfn.PeerType)
            break;  /* Matched */
         wo = 0;
#if QBYTE_PEER
         if (p_p->Low.PeerAddress.qbyte == sp->sfn.DestPeerAddr &&
               p_p->High.PeerAddress.qbyte == sp->sfn.SrcPeerAddr &&
#else
         if (memcmp(p_p->Low.PeerAddress.peer32, 
                  sp->sfn.DestPeerAddr, PEER_ADDR_LEN) == 0 &&
               memcmp(p_p->High.PeerAddress.peer32, 
                  sp->sfn.SrcPeerAddr, PEER_ADDR_LEN) == 0 &&
#endif
               ( p_p->has_ports ?
                  p_p->Low.TransAddress == sp->sfn.DestTransAddr &&
                  p_p->High.TransAddress == sp->sfn.SrcTransAddr
                  : 1 ) &&
               p_p->TransAddrType == sp->sfn.TransType &&
               p_p->PeerAddrType == sp->sfn.PeerType)
            break;  /* Matched */

         lsp = sp;
         if ((sp = lsp->next_hc) == (struct stream *)sha) {
            wo = 1;  sp = NULL;  break;
            /* Loop stops with lsp -> last stream in chain */
	    }
         }
      }
   if (sp != NULL) {  /* Existing stream */
//20      mutex_unlock("lsh", &strhash_q_mutex);
//20      if (lock) mutex_unlock("lsh", &stream_state_mutex);
      sh_ok(sp,"fsh: existing");
      *spp = sp;  *create_new = 0;
      }
   else if (!*create_new) {  /* Not in stream table */
      wo = 1;  /* New stream */
      *spp = NULL;
//20      mutex_unlock("lsh", &strhash_q_mutex);
//20      if (lock) mutex_unlock("lsh", &stream_state_mutex);
      }
   else {  /* Make new stream */
      nsp = get_stream();
      if (nsp == NULL) {
         *spp = NULL;  *create_new = 0;
//20         mutex_unlock("lsh", &strhash_q_mutex);
//20         if (lock) mutex_unlock("lsh", &stream_state_mutex);
         }
      else  {
         memset(nsp, 0, sizeof(struct stream)-sizeof(nsp->sfn));
#if QBYTE_PEER
         nsp->sfn.SrcPeerAddr = p_p->Low.PeerAddress.qbyte;
         nsp->sfn.DestPeerAddr = p_p->High.PeerAddress.qbyte;
#else
         memcpy(nsp->sfn.SrcPeerAddr, 
            p_p->Low.PeerAddress.peer32, PEER_ADDR_LEN);
         memcpy(nsp->sfn.DestPeerAddr, 
            p_p->High.PeerAddress.peer32, PEER_ADDR_LEN);
#endif
         if (nsp->has_ports = p_p->has_ports) {
            nsp->sfn.SrcTransAddr = p_p->Low.TransAddress;
            nsp->sfn.DestTransAddr = p_p->High.TransAddress;
            }
	 else
            nsp->sfn.SrcTransAddr = nsp->sfn.DestTransAddr = 0;
         nsp->sfn.TransType = p_p->TransAddrType;
         nsp->sfn.PeerType = p_p->PeerAddrType;

         if (lsp == NULL) *sha = nsp;  /* Add stream to hash chain */
         else lsp->next_hc = nsp;
         nsp->next_hc = (struct stream *)sha;
         nsp->state = SS_GOT;  /* get_stream() sets SS_GOT, was zeroed above */
//20         mutex_unlock("lsh", &strhash_q_mutex);
//20         if (lock) mutex_unlock("lsh", &stream_state_mutex);

         sh_ok(nsp,"lsh: new");
         *spp = nsp;

#if STH_CHECK
         check_stream_state(nsp, 1, SS_GOT, "fsh(new)");
#endif
         }
      }
   *wire_order = wo;
   }

int lookup_stream(struct pkt *p_p, int create_new, int requeue)
{
   int make_new, wo, r;
   struct stream *sp;

   make_new = create_new;
//?   if (!lock) return 0;  /* terminating */

//20   if (lock) mutex_lock("lus", &stream_mutex);  //20
   lookup_stream_helper(&sp, &wo, p_p, &make_new);
      /* wo = 1  same order as first packet seen */

   if (sp == NULL) {
      p_p->sp = NULL; return 0;  /* Ignored || no free pkt */
      }

   sp->LastTime = p_p->arrival_time/10000.0;
   if (make_new) {
      sp->FirstTime = sp->LastTime;  r = 2;
      }
  else r = 1;
   sp->StrLastTime = p_p->arrival_time;  /* us */

   if (p_p->PeerAddrType < AT_P2P)
      last_t_cs = sp->LastTime;

#if 0  /* ^_^ */
//10   if (sp->state != SS_IGNORE) {
   if (sp->state != SS_IGNORE && (
       ntohs(sp->sfn.SrcTransAddr) == 53 ||
       ntohs(sp->sfn.DestTransAddr) == 53 )
       ) {
      printf("lookup_stream(1): sp=%d, dirn=%d: ", stream_nbr(sp), wo);
      daddr(sp->sfn.SrcPeerAddr);  printf(wo ? "->" : "<-");
      daddr(sp->sfn.DestPeerAddr);  printf(" ");
      dport(sp->sfn.SrcTransAddr);  printf(wo ? "->" : "<-");
      dport(sp->sfn.DestTransAddr);
      printf("  pt=%d, tt=%d.  pkts=%u\n",
         sp->sfn.PeerType,  sp->sfn.TransType,  sp->ToPDUs+sp->FromPDUs);
      fflush(stdout);
      }
#endif

   p_p->opp_dir = !wo;

   ++sp->n_pkts_queued;

//20   if (lock) {
//20      mutex_lock("lsh2", &stream_state_mutex);
//20      mutex_lock("lsh2", &stream_q_mutex);
//20      }
   if (sp->state == SS_GOT) {
      queue_new_stream(sp);
      sp->state = SS_UNKNOWN;
      }
   else {
      if (requeue)  /* Don't do this while check_idle_streams
                          has the stream chain locked */
         requeue_active_stream(sp);
      }
//20   if (lock) {
//20      mutex_unlock("lsh2", &stream_q_mutex);
//20      mutex_unlock("lsh2", &stream_state_mutex);
//20      }

//20   if (lock) mutex_unlock("lus", &stream_mutex);  //20

   if (sp->state == SS_IGNORE && sp->oa_count == 0) {
      --sp->n_pkts_queued;
      p_p->sp = NULL;  return 0;  /* Ignored */
      }
   p_p->sp = sp;
   return r;
   }

void queue_new_stream(struct stream *sp)
{  /* CAUTION: caller must lock stream_q_mutex */
  if (sp->state != SS_GOT) {
      log_msg(LOG_WARNING, 0,
         "Couldn't queue new stream %x, state=%d!", sp, sp->state);
   quack(302);
      return;
      }

#if SCC_CHECK > 1
   check_stream_chains("queue_new_stream(before)");
#endif
   if (t_sq == NULL) {  /* Put it at tail of queue */
      t_sq = sp;  sp->prev_sp = NULL;
      }
   else {
      t_sq_tail->next_sp = sp;  sp->prev_sp = t_sq_tail;
      }
   t_sq_tail = sp;  sp->next_sp = NULL;
   ++t_active_streams;

#if SCC_CHECK
   check_stream_chains("queue_new_stream(after)");
#endif
   }

void requeue_active_stream(struct stream *sp)
{  /* CAUTION: caller must lock stream_q_mutex */
   if (sp->state != SS_UNKNOWN && sp->state != SS_MATCHED &&
         sp->state != SS_IGNORE) {
      log_msg(LOG_WARNING, 0,
         "Couldn't requeue stream %x, state=%d!", sp, sp->state);
   quack(301);
      return;
      }

#if SCC_CHECK > 1
   check_stream_chains("queue_active_stream(before)");
#endif

   if (sp->prev_sp != NULL)  /* Remove stream from queue */
      sp->prev_sp->next_sp = sp->next_sp;
   else t_sq = sp->next_sp;
   if (sp->next_sp != NULL)
      sp->next_sp->prev_sp = sp->prev_sp;
   else t_sq_tail = sp->prev_sp;

   if (t_sq == NULL) {  /* Put it back at tail of queue */
      t_sq = sp;  sp->prev_sp = NULL;
      }
   else {
      t_sq_tail->next_sp = sp;  sp->prev_sp = t_sq_tail;
      }
   sp->next_sp = NULL;  t_sq_tail = sp;

#if SCC_CHECK
   check_stream_chains("queue_active_stream(after)");
#endif
   }

struct flow *create(struct search_result *s_r,
   int rtx, struct pme_env *pme,
   struct flow_key *search_key, struct pkt *p_p)
{
   struct flow *f, *ff, *lf;
   Bit32 fx;
   Bit32 rs_first_flow;
   struct flow *pf;
#if NEW_ATR || defined(MATCH_TRACE)
   int j;
#endif
#if OAM
   char m_name[20];
#endif

   if (FloodMode == TV_TRUE)  /* Running in flood mode */
      return NULL;  /* Don't garbage collect for every new flow! */
   if ((fx = alloc_flow_nbr()) == 0) {  /* Find a flow data row */
      bump_noflowpkts(ntohs(p_p->ntm_interface));
      return NULL;  /* No unused rows in flow table */
      }

   mutex_lock("P", &flow_table_mutex);
   f = &fa[fx-1];
#if FLOW_INDEX
   flow_ix[fx-1] = f;  /* 1-org */
#endif
   ++nflows;  /* Nbr of active accounting flows */
   ++FlowsCreated;
   ++ri[rtx-1].ri_FlowRecords;  /* 1-org */

   if (f == s_r->lf) {  /* Garbage collector recovered the last
               flow of the hash chain.  (Yes, it can happen)! */
      if ((ff = s_r->sf[0]) == NULL) {
         lf = NULL;  /* Empty hash chain */
         }
      else {
         do {
            lf = ff;
            } while ((ff = lf->ht_next) != (struct flow *)s_r->sf);
         }  /* Loop stops with lf -> last flow in chain */
      s_r->lf = lf;
      }

   f->ht_next = (struct flow *)(s_r->sf);  /* Link f into hash chain */
   if (s_r->lf == NULL) {  /* Empty hash chain */
      s_r->sf[0] = f;  ++n_hash_ents;
      }
   else (s_r->lf)->ht_next = f;

   rs_first_flow = ri[rtx-1].ri_flow_chain;  /* Link f into rs chain */
   if (rs_first_flow == 0) {  /* First flow for this ruleset */
      f->rs_next = 0;
      ri[rtx-1].ri_flow_chain = fx;
      }
   else if (fx < rs_first_flow) {  /* Add to front of chain */
      f->rs_next = rs_first_flow;
      ri[rtx-1].ri_flow_chain = fx;
      }
   else {  /* Link into ruleset chain */
      pf = f;  do {
         --pf;
         } while (pf->FlowRuleSet != rtx);
      f->rs_next = pf->rs_next;
      pf->rs_next = fx;
      }
   mutex_unlock("P", &flow_table_mutex);

# if RS_CH_DEBUG
   check_rs_chain(rtx, "create");
# endif

   memcpy(&f->fk,  /* Copy values from search_key */
      search_key, sizeof(struct flow_key));
#ifdef MATCH_TRACE
   printf("new flow, hash=%u :  f=%Fp\n", search_hash,f);
   printf("Count: %u+%u",pme->a_stack[0],pme->p_stack[0]);
   for (j = 1; j != pme->p_s_depth; ++j)
      printf(",%u+%u",pme->a_stack[j],pme->p_stack[j]);
   printf(": ");
   daddr(search_key->Low.PeerAddress);  printf("->");
   daddr(search_key->High.PeerAddress);  printf("\n");
#endif

   f->FlowRuleSet = rtx;
   setzero64(f->UpOctets);
   setzero64(f->UpPDUs);
   setzero64(f->DownOctets);
   setzero64(f->DownPDUs);

#if NF_CISCO_DATA
   f->LastTime = f->FirstTime = p_p->dFirst;
#elif LFAP_METER
   f->LastTime = f->FirstTime = p_p->FirstTime;
#else
   f->FirstTime = p_p->arrival_time/10000.0;  /* cs */
#endif
#if NEW_ATR
   f->stdata = NULL;
# if OAM
   if (f->fk.OAMIdent != 0 && 
         get_flow_stdata(f) != 0) {  /* Make sure flow has stdata */
      if (f->stdata->OAMtrigger != 0)
         log_msg(LOG_WARNING, 0, 
            "Flow %d has OAMtrigger=%d; its OAMIdent must be zero <<<",
            flow_nbr(f), f->stdata->OAMtrigger);
      else if (f->fk.OAMIdent > MXOAMIDS || f->fk.OAMIdent < 1)
         log_msg(LOG_WARNING, 0, 
            "Flow %d: OAMIdent must be from 1 to %d <<<",
            flow_nbr(f), MXOAMIDS);
      else {
#  if OAM_TRACE
          printf("Creating OAM flow, flow=%d, Ident = %d, stdata=%x\n",
            flow_nbr(f), f->fk.OAMIdent, f->stdata);
         fflush(stdout);
#  endif
         f->stdata->OAMtrigger = 0;  f->stdata->OAM_head = NULL;
         mutex_lock("cr", &oam_q_mutex);
         f->stdata->OAM_next_flow = oaf[f->fk.OAMIdent-1];
            /* Add to head of list */
         oaf[f->fk.OAMIdent-1] = f;
         mutex_unlock("cr", &oam_q_mutex);
#  if MULTI
	 sprintf(m_name, "OAM_%u_%u", flow_nbr(f), f->fk.OAMIdent);
         init_mutex(&f->stdata->OAM_mutex, m_name);
#  endif
         }
      }
# endif  /* OAM */

   f->distrib_list = NULL;
   f->distrib_bits = 0;
   make_distrib_list(pme, f, p_p->sp);
   setzero64(f->LastUpOctets);
   setzero64(f->LastUpPDUs);
   setzero64(f->LastDownOctets);
   setzero64(f->LastDownPDUs);
#endif  /* NEW_ATR */
   return f;
   }

#if NEW_ATR
static double pwr_of_10[10] = {
   1.0,
   0.1,
   0.01,
   0.001,
   0.0001,
   0.00001,
   0.000001,
   0.0000001,
   0.00000001,
   0.000000001 };

Bit32 times_pwr_of_10(Bit16 x, int p)
{
   return (Bit32)((double)x)/pwr_of_10[p > 9 ? 9 : p];
   }

Bit32 scale(struct distribution *d, double y)
{
   double x;
   x = y*pwr_of_10[d->FilterScale > 9 ? 9 : d->FilterScale];
   return (Bit32)floor(x);
   }

double inv_scale(struct distribution *d, Bit16 j)
{
   return ((double)j)/pwr_of_10[d->Scale > 9 ? 9 : d->Scale];
   }

#define CHECK_DIST_SCALE  1
#if CHECK_DIST_SCALE
int n_bad_t = 0;  //03
int n_bad_n = 0;  //03
#endif
void bump_dist(struct distribution *d, double y)
{
   double x, T;
   int n;
   Bit32 ix;

#if CHECK_DIST_SCALE
   if (d->Transform < 1 || d->Transform > DS_LOG_RESET ) {
      if (n_bad_t % 10 == 0)
      log_msg(LOG_ERR, 0, "bump_dist(1): flow %d, transform=%d, mask_params[0]=%d",
         flow_nbr(d->fp), d->Transform, d->mask_params[0]);
      ++n_bad_t;
      return;
      }
#endif
   if (d->dyn_dist)
      mutex_lock("bd", &d->dyn_dst_mutex);
   if ((d->Transform == DS_DYN_REQ || d->Transform == DS_DLOG_REQ)
         && d->n_values <= d->Buckets) {
      if (d->n_values == d->Buckets) {
            /* Too many actual values, start using bins */
         counter64 values[MXBUCKETS+1];
         double R, N,D;
         int m, j;

         N = d->Buckets - 1;
         switch (d->Transform) {
         case DS_DYN_REQ:
            R = d->max_val == d->min_val ? 8  /* Bug fix, 25 Jun 02 */
              : d->max_val - d->min_val;  /* Leave room at both ends */
            m = d->min_val - 0.125*R;  /* May be negative */
            d->LowerLimit = m < d->lowerspec ? d->lowerspec : m;
            d->UpperLim = d->max_val + 0.25*R;
            D = d->UpperLim - d->LowerLimit;
            d->M = N/D;
            break;
         case DS_DLOG_REQ:
            m = d->min_val*0.7;  /* Leave room at both ends */
            d->LowerLimit = m < d->lowerspec ? d->lowerspec : m;
            d->UpperLim = d->max_val*3.0;
            D = log10((double)d->UpperLim/(double)d->LowerLimit);
            d->M = N/D;
            break;
            }
#if 0  /* +-+ */
char nbr[20], buf[2000], *p_p;
   log_msg(LOG_WARNING, -1,
      "DD: lower=%d, upper=%d, min_val=%d, max_val=%d, Lower=%d, Upper=%d",
      d->lowerlim,d->upperlim, d->min_val,d->max_val,
      d->LowerLimit, d->UpperLim);
   for (p_p = buf, j = 0; j != d->Buckets; ++j) {
      sprintf(nbr," %llu", d->counts[j]);
      p_p = strmov(p_p, nbr);
      }
   *p_p = '\0';
   log_msg(LOG_WARNING, -1,  /* %% */
      "  counts=%s", buf);
   }
#endif

         memcpy(values, d->counts, sizeof(d->counts));
         memset(d->counts, 0, sizeof(d->counts));
         for (j = 0; j != d->Buckets; ++j) {
            ix = Bit32from64(values[j]);
               /* Assume Bit32 values to determine dynamic limits */
            if (ix <= d->LowerLimit) n = 0;
            else if (ix > d->UpperLim) n = d->Buckets;
            else switch (d->Transform) {
            case DS_DYN_REQ:
               T = ((double)ix - (double)d->LowerLimit) * d->M;
               n = 1 + (int)floor(T);
               break;
            case DS_DLOG_REQ:
               T = log10((double)ix/(double)d->LowerLimit) * d->M;
               n = 1 + (int)floor(T);
               break;
               }
#if CHECK_DIST_SCALE
            if (n < 0 || n > d->Buckets || n > MXBUCKETS) {
               log_msg(LOG_ERR, 0, "bump_dist(2): flow %d, n=%d, transform=%d",
                  flow_nbr(d->fp), n, d->Transform);
               }
            else
#endif
               incr_counter64(d->counts[n]);
            }

         d->n_values = d->Buckets+1;
         }
      else {  /* d->n_values < d->Buckets */
         ix = y*pwr_of_10[d->Scale > 9 ? 9 : d->Scale];
         if (ix < d->min_val && ix >= d->lowerspec) d->min_val = ix;
         if (ix > d->max_val && ix <= d->upperspec) d->max_val = ix;

         assign_Bit32_to_c64(d->counts[d->n_values++], ix);
         d->non_empty = 1;
         if (d->dyn_dist)
            mutex_unlock("bd", &d->dyn_dst_mutex);
         return;
         }
      }

   x = y*pwr_of_10[d->Scale > 9 ? 9 : d->Scale];
   if (x <= (double)d->LowerLimit) n = 0;
   else if (x > (double)d->UpperLim) {
      n = d->Buckets;
      }
   else switch (d->Transform) {
   case DS_LIN:
   case DS_DYN_REQ:
   case DS_LIN_RESET:
      T = (x - (double)d->LowerLimit) * d->M;
      n = 1 + (int)floor(T);
      break;
   case DS_LOG:
   case DS_DLOG_REQ:
   case DS_LOG_RESET:
      T = log10(x/(double)d->LowerLimit) * d->M;
      n = 1 + (int)floor(T);
      break;
      }
#if CHECK_DIST_SCALE
   if (n < 0 || n > d->Buckets || n > MXBUCKETS) {
      if (n_bad_n % 10 == 0)
         log_msg(LOG_ERR, 0, "bump_dist(3): flow %d, n=%d, transform=%d",
            flow_nbr(d->fp), n, d->Transform);
      ++n_bad_n;
      }
   else
#endif
      incr_counter64(d->counts[n]);
   d->non_empty = 1;
#if PP_DEBUG_XX >= 2
   log_msg(LOG_WARNING, 0,
      "  bump_dist(4): d=%x, attrib=%d, n=%d", d, d->selector, n);
#endif
   if (d->dyn_dist)
      mutex_unlock("bd", &d->dyn_dst_mutex);
   }

void accum_dist(struct distribution *d, double y, Bit32 v)
{
   double x, T;
   int n;

   if (d->dyn_dist)
      mutex_lock("ad", &d->dyn_dst_mutex);
   x = y*pwr_of_10[d->Scale > 9 ? 9 : d->Scale];
   if (x <= (double)d->LowerLimit) n = 0;
   else if (x > (double)d->UpperLim) {
      n = d->Buckets;
      }
   else switch (d->Transform) {
   case DS_LIN:
   case DS_DYN_REQ:
   case DS_LIN_RESET:
      T = (x - (double)d->LowerLimit) * d->M;
      n = 1 + (int)floor(T);
      break;
   case DS_LOG:
   case DS_LOG_RESET:
      T = log10(x/(double)d->LowerLimit) * d->M;
      n = 1 + (int)floor(T);
      break;
      }
   add_Bit32_to_counter64(d->counts[n], v);
   d->non_empty = 1;
   if (d->dyn_dist)
      mutex_unlock("ad", &d->dyn_dst_mutex);
   }

void other_stream_update(struct count_env *cep, 
   struct stream *sp, struct pktdata *tp)
   /* Checks for timed-out pktdatas, adds tp to tail of sp's pq */
{
   double reply_timeout;
   struct pktdata *pd, *lpd, *npd;
   struct stream_data *sdp;

   reply_timeout = tp->ts - OTHER_STREAM_TIMEOUT*10000.0;  /* U centiseconds */
# if DNS_ROOT_DEBUG
   log_msg(LOG_WARNING, -1, /* Keep log file open */
      "other_st_update(): sp=%d, pq_len=%d, mx_len=%d, tp->ts=%f",
      stream_nbr(sp), sp->pq_len,sp->mx_pq_len, tp->ts);
   log_msg(LOG_WARNING, 0, ""); /* Close log file */
#endif
   lpd = NULL;
   mutex_lock("osu", &pktdata_q_mutex);
#if PP_DEBUG_XX >= 2
   printf("o_s_update(%x):", tp);
#endif
   for (pd = sp->pq; pd != NULL; ) {
         /* Look for pkts to discard */
#if PP_DEBUG_XX >= 2
   printf(" %x", pd);
#endif
      if (sp->pq_len >= sp->mx_pq_len ||  /* Too many packets in queue */
            pd->ts < reply_timeout) {     /* Older than upper limit */
         if (pd->direction == MFWD) ++cep->lost_to;
         else ++cep->lost_from;
# if DNS_ROOT_DEBUG
   log_msg(LOG_WARNING, 0,
      "Lost packet: sp=%d, pq_len=%d, mx_len=%d, r_to=%f, pd->ts=%f",
      stream_nbr(sp), sp->pq_len,sp->mx_pq_len, reply_timeout, pd->ts);
#endif
         if (sp->pq_len >= sp->mx_pq_len) {
            if (pd->direction == MFWD) ++cep->oflo_to;
            else ++cep->oflo_from;
            }
         --sp->pq_len;
         npd = pd->next;
         if (lpd == NULL) sp->pq = npd;
         else lpd->next = npd;
         free_pktdata(pd);  pd = npd;
         }
      else {
         lpd = pd;  pd = pd->next;
         }
      }

   if (lpd == NULL) sp->pq = tp;  /* Add new pkt to tail of queue */
   else lpd->next = tp;
   tp->next = NULL;
   mutex_unlock("osu", &pktdata_q_mutex);
   ++sp->pq_len;  /* Was missing, causing pktdatas to be
      wrongly discarded as lost packets.  Nevil, 27 Oct 00 */
#if PP_DEBUG_XX >= 2
   printf("\n");  fflush(stdout);
#endif
   }

void other_flow_update(struct count_env *cep, 
      struct flow *fp, struct stream *sp, struct pktdata *tp)
{
   struct stream_data *sdp = fp->stdata;

   sdp->ToLostPDUs += cep->lost_to;
   sdp->FromLostPDUs += cep->lost_from;
   sdp->ToPQOverflows += cep->oflo_to;
   sdp->FromPQOverflows += cep->oflo_from;
   }
#endif  /* NEW_ATR */

#define CHECK_UPTIME_DECR  !MULTI
   /* Don't need this when using pkts from merged input streams */

#if CHECK_UPTIME_DECR
struct uptime_check {
   Bit32 lll_Time, lll_good, lll_bad;
   long long lll_llL;  /*LL*/
   };
struct uptime_check utc_data[N_PME_THREADS] = {
# if N_PME_THREADS == 4
  { 0,0,0,0LL }, { 0,0,0,0LL }, { 0,0,0,0LL }, {0,0,0,0LL} };
# elif N_PME_THREADS == 2
  { 0,0,0,0LL }, {0,0,0,0LL} };
# else
  {0,0,0,0LL} };
# endif
#endif

#if LOG_DNS_PKTS
FILE *dns_log = NULL;
# if DAG_DIRECT  /* Really need fmt_start_ts() in meter_ux.c <<<<< */
extern long long meter_start_time;  /* Dag (64-bit) timestamp */
# else
extern struct timeval meter_start_time;
# endif

int print_dn(Bit8 *nbp, Bit8 *dnp, int n_len, int level)
{
   int j, n, k;
   int lj[20], ids;

   if (level > 2) level = 2;
   for (ids = j = 0; ;) {
      if ((n = dnp[j]) == 0) break;
      lj[ids++] = j;  ++j;
      if ((j += n) >= n_len) {
         *nbp = '\0';   return j;
         }
      }
   if (ids == 0) *nbp++ = '.';
   else {
      k = ids-level;
      if (k < 0) k = 0;
      for ( ; k != ids; ++k) {
         j = lj[k];
         for (n = dnp[j++] ; n != 0; --n)
            *nbp++ = dnp[j++];
         *nbp++ = '.';
         }
      }
   *nbp = '\0';
   return j+1;
   }

void log_dns(struct flow *fp, struct pkt *p_p, double at_us)
{
   int j;
   Bit8 nbuf[500];

   if (fp == NULL || p_p == NULL) {
      log_msg(LOG_ERR, 0, "log_dns(): fp or p_p == NULL!");
      return;
      }

   if (dns_log == NULL) {
      if ((dns_log = fopen("dns.log", "w")) == NULL)
         log_msg(LOG_ERR, 1, "Couldn't open dns.log <<<<<");
      fprintf(dns_log, "%lu.%06lu meter_start_time\n",
# if DAG_DIRECT
         DAGT_S(meter_start_time), DAGT_US(meter_start_time));
# else
         meter_start_time.tv_sec, meter_start_time.tv_usec);
# endif
      }
   fprintf(dns_log, "%13.6f %d %d  ",
      at_us/1000000.0,  fp->fk.FlowKind, fp->fk.FlowClass);
   if (p_p->pi.udp.dns.p0 & dns_resp)  /* DNS response */
#if QBYTE_PEER
      fdaddr(dns_log, p_p->High.PeerAddress.qbyte);
#else
      fdaddr(dns_log, *(Bit32 *)&p_p->High.PeerAddress);
#endif
   else  /* DNS request */
#if QBYTE_PEER
      fdaddr(dns_log, p_p->Low.PeerAddress.qbyte);
#else
      fdaddr(dns_log, *(Bit32 *)&p_p->Low.PeerAddress);
#endif
   fprintf(dns_log, "  %04x %02x%02x ", 
      p_p->pi.udp.dns.ident,  p_p->pi.udp.dns.p0,p_p->pi.udp.dns.p1);
//15   fprintf(dns_log, " %d %d %d %d  ",
//15      ntohs(p_p->pi.udp.dns.qdcount), ntohs(p_p->pi.udp.dns.ancount),
//15      ntohs(p_p->pi.udp.dns.nscount), ntohs(p_p->pi.udp.dns.arcount));
   j = 12 + print_dn(nbuf, &p_p->pi.udp.d.data[12], DATA_SZ-12, fp->fk.FlowClass+1);
//02   fprintf(dns_log, " %s %u %u", nbuf,
//02      (p_p->pi.udp.d.data[j] << 8) | p_p->pi.udp.d.data[j+1], j+2);
   fprintf(dns_log, " %s %u", nbuf,
      (p_p->pi.udp.d.data[j] << 8) | p_p->pi.udp.d.data[j+1]);
   /* DNS Resource Record (RR) values in
         http://www.iana.org/assignments/dns-parameters */
   fprintf(dns_log, "\n");
   }
#endif

#if NEW_ATR && STREAM_TO_FLOW
void create_flow_from_stream(
   struct stream *sp, struct flow *ofp)
{
   Bit32 fx, new_rs_next;
   struct flow *fp;
   struct stream_name *sfn;
   Bit32 rs_first_flow;
   struct flow *pfp;
   int rtx = ofp->FlowRuleSet;  /* Parent flow's ruleset ix */

#  if STF_DEBUG
   log_msg(LOG_WARNING, 0,
      "STF(2): stream=%x, flow=%x, about to create flow", sp, ofp);
#  endif
   if ((fx = alloc_flow_nbr()) == 0) {  /* Find a flow data row */
#  if STF_DEBUG
      log_msg(LOG_WARNING, 0,
         "STF(2a): couldn't allocate flow nbr ???");
#  endif
      return;  /* No unused rows in flow table */
      }
   fp = &fa[fx-1];  /* New flow for stream */
#  if STF_DEBUG
   log_msg(LOG_WARNING, 0,
      "STF(2b): stream=%x, flow=%x, using flow fx=%d, rtx=%d",
      sp, ofp, fx, rtx);
#  endif
#if FLOW_INDEX
   flow_ix[fx-1] = fp;  /* 1-org */
#endif
   ++nflows;  /* Nbr of active flows */
   ++FlowsCreated;
   ++ri[rtx-1].ri_FlowRecords;  /* 1-org */
   rs_first_flow = ri[rtx-1].ri_flow_chain;
      /* Link f into rs chain */
   if (rs_first_flow == 0) {  /* First flow for this ruleset */
      new_rs_next = 0;
      ri[rtx-1].ri_flow_chain = fx;
      }
   else if (fx < rs_first_flow) {  /* Add to front of chain */
      new_rs_next = rs_first_flow;
      ri[rtx-1].ri_flow_chain = fx;
      }
   else {  /* Link into ruleset chain */
      pfp = fp;  do {
         --pfp;
         } while (pfp->FlowRuleSet != rtx);
      new_rs_next = pfp->rs_next;
      pfp->rs_next = fx;
      }

   memset(fp, 0, sizeof(struct flow));
   fp->rs_next = new_rs_next;  /* Nevil, 16 Jun 03 */
   memcpy(&fp->fk, &ofp->fk, sizeof(struct flow_key));
   /* Copy original flow key, this gets attributes saved by ruleset */

   fp->FlowRuleSet = ofp->FlowRuleSet;
   fp->FirstTime = sp->FirstTime;

   /* Overwrite original attributes with values from stream */
   sfn = &sp->sfn;
#if QBYTE_PEER
   fp->fk.Low.PeerAddress.qbyte = sfn->SrcPeerAddr;
   fp->fk.High.PeerAddress.qbyte = sfn->DestPeerAddr;
#else
   memcpy(&fp->fk.Low.PeerAddress.byte,
      sfn->SrcPeerAddr, sizeof(sfn->SrcPeerAddr));
   memcpy(&fp->fk.High.PeerAddress.byte,
      sfn->DestPeerAddr, sizeof(sfn->DestPeerAddr));
#endif
   fp->fk.Low.TransAddress = sfn->SrcTransAddr;
   fp->fk.High.TransAddress = sfn->DestTransAddr;
   fp->fk.TransAddrType = sfn->TransType;

   fp->fk.ECNCodeBits = sp->ToECN | sp->FromECN;
      /* Use whichever direction we see first */

   assign_Bit32_to_c64(fp->DownOctets, sp->FromOctets);
   assign_Bit32_to_c64(fp->UpOctets, sp->ToOctets);
   assign_Bit32_to_c64(fp->DownPDUs, sp->FromPDUs);
   assign_Bit32_to_c64(fp->UpPDUs, sp->ToPDUs);

   fp->sf_state = SF_ACTIVE;
   fp->flow_strp = sp;
   sp->str_flowp = fp;
   }
#endif

#if NEW_ATR && OAM
void build_oam_recs(int OAMtrigger, struct pkt *p_p, int rs)
{
   int r;
   Bit32 readers;
   struct flow *fp;
   struct OAMrec *noap, *loap, *oap;

   for (readers = r = 0; r != MXCLCTRS; ++r) {
      if (ci[r].ci_RuleSet == rs) readers |= 1 << r;
      }
   mutex_lock("bor", &oam_q_mutex);
   for (fp = oaf[OAMtrigger-1]; fp != NULL; fp = fp->stdata->OAM_next_flow) {
# if OAM_TRACE
      printf("Building OAM record for flow=%d, trigger=%d, readers=%x\n",
         flow_nbr(fp), fp->stdata->OAMtrigger, readers);
      fflush(stdout);
# endif
      mutex_lock("bor", &fp->stdata->OAM_mutex);
      if ((noap = get_oamrec()) != NULL) {
         noap->trigger_ip_id = htons(p_p->ip_id);
         noap->trigger_ts = unix_ts(p_p->arrival_time);
         noap->ToOctets = fp->UpOctets;
         noap->ToPDUs = fp->UpPDUs;  
         noap->FromOctets = fp->DownOctets;
         noap->FromPDUs = fp->DownPDUs;
         noap->reader_bits = readers;
         noap->next = NULL;
	 if ((loap = fp->stdata->OAM_head) == NULL)  /* Empty queue */
            fp->stdata->OAM_head = noap;
         else {
            for (oap = loap->next; oap != NULL; ) {  /* Find tail */
               loap = oap;  oap = oap->next;
               }
            loap->next = noap;  /* Add to tail */
            }
         }
      mutex_unlock("bor", &fp->stdata->OAM_mutex);
      }
   mutex_unlock("bor", &oam_q_mutex);
   }
#endif

void actual_count(struct count_env *cep,
   struct flow *fp, int flow_order, struct pkt *p_p)
{
   struct search_result *s_r;
   Bit32 ThisUptime;
#if NEW_ATR
   struct stream *sp = p_p->sp;
   struct pktdata *pd, *lpd;
   struct distribution *d;
   int tlen;

   Bit16 ThisFlags;
   Bit32 ThisAck, ThisSeq;

   double iat_delta_us;
   int in_train;
#endif
#if DNS_ROOT_DEBUG
   int matched = 0;
#endif
#if APP_DATA
   struct flow *ofp;
#endif

   if (!ri[fp->FlowRuleSet-1].ri_running) {  /* No longer running */
      ntm_semaphore_up("ac", &r_rts_sem);
      return;
      }   

#if NF_CISCO_DATA
   if (flow_order == MREV) {
      add_Bit32_to_counter64(fp->DownOctets,p_p->dOctets);
      add_Bit32_to_counter64(fp->DownPDUs,p_p->dPkts);
      }
   else {
      add_Bit32_to_counter64(fp->UpOctets,p_p->dOctets);
      add_Bit32_to_counter64(fp->UpPDUs,p_p->dPkts);
      }
   if (p_p->dFirst < fp->FirstTime)
      fp->FirstTime = p_p->dFirst;
   if (p_p->dLast > fp->LastTime)
      fp->LastTime = p_p->dLast;
   fp->ntm_LastTime = uptime_cs();  /* NetFlowMet meter time! */
#elif LFAP_METER
   if (flow_order == MREV) {
      add_Bit32_to_counter64(fp->DownOctets,p_p->dOctets_sent);
      add_Bit32_to_counter64(fp->DownPDUs,p_p->dPackets_sent);
      add_Bit32_to_counter64(fp->UpOctets,p_p->dOctets_recvd);
      add_Bit32_to_counter64(fp->UpPDUs,p_p->dPackets_recvd);
      }
   else {
      add_Bit32_to_counter64(fp->UpOctets,p_p->dOctets_sent);
      add_Bit32_to_counter64(fp->UpPDUs,p_p->dPackets_sent);
      add_Bit32_to_counter64(fp->DownOctets,p_p->dOctets_recvd);
      add_Bit32_to_counter64(fp->DownPDUs,p_p->dPackets_recvd);
      }
/*	done in create(...);
   fp->FirstTime = p_p->FirstTime; */
   fp->LastTime  = p_p->LastTime;
#else  /* !NF_CISCO_DATA && !LFAP_METER */

   cep->at_us = p_p->arrival_time;
   cep->at_cs = sp->LastTime;

# if APP_DATA
   switch(sp->pm_state) {
   case PM_NOTEST:
      break;
   case PM_EARLY:
      sp->p2p_fx = flow_nbr(fp);  /* Remember flow for p2pProto zero */
      sp->pm_state = PM_WAITING;
      break;
   case PM_WAITING:
      break;
   case PM_MATCH_1:
      if (sp->p2p_fx != 0) {  /* Switch to different flow */
         ofp = &fa[sp->p2p_fx-1];
         move_stream_counts(fp, ofp, p_p);
         }
      sp->p2p_fx = flow_nbr(fp);
      sp->pm_state = PM_MATCHED;
      break;
   case PM_MATCHED:
      break;
   case PM_COMPLETE:  /* p2p protocol determined */
      if (sp->p2p_fx != 0) {  /* p2p match on earlier packet */
         ofp = &fa[sp->p2p_fx-1];
         move_stream_counts(fp, ofp, p_p);
         }
#if P2P_TEST_TAGS
      tag_p2p_host(sp);
#endif
      sp->pm_state = PM_COUNTING;
      break;
   case PM_COUNTING:
      break;
      }
# endif

   if (flow_order == MREV) {
      add_Bit32_to_counter64(fp->DownOctets,p_p->p_len);
      incr_counter64(fp->DownPDUs);
      }
   else {
      add_Bit32_to_counter64(fp->UpOctets,p_p->p_len);
      incr_counter64(fp->UpPDUs); 
     }
   fp->ntm_LastTime = cep->at_cs;
      /* Also sets LastTime (see flowhash.h) */
#endif  /* !NF_CISCO_DATA etc. */

   /* !!!!! fp->fk.ECNCodeBits |= p_p->ECNCodeBits;
      /* OR the bits together; 0 -> no ECN */
   /* ECN is part of flow key, can't change it once flow is created !!! */

#if 0  /* CR_DATA || DAG_DIRECT  /*P*/
   last_ts = p_p->arrival_time;  /* Used by uptime() routines */
#endif

#if CHECK_UPTIME_DECR
   if (fp->ntm_LastTime < utc_data[p_p->hash_group].lll_Time) {
      log_msg(LOG_WARNING, 0,
         "LastTime jump backward: if=%d, hg=%d, lll=%u, Last=%u, dif=%d, good=%u, bad=%u", 
         p_p->Low.Interface, p_p->hash_group, utc_data[p_p->hash_group].lll_Time,
         fp->ntm_LastTime, utc_data[p_p->hash_group].lll_Time-fp->ntm_LastTime,
         utc_data[p_p->hash_group].lll_good, ++utc_data[p_p->hash_group].lll_bad);
      if (utc_data[p_p->hash_group].lll_bad > 100) log_msg(LOG_ERR, 1, "Too many backward time jumps!");
      }
   else  ++utc_data[p_p->hash_group].lll_good;
   utc_data[p_p->hash_group].lll_Time = fp->ntm_LastTime;
#endif

/* printf("66 count(1): stdata=%0x, match_reqd=%d\n",
     fp->stdata, fp->stdata->pp_match_reqd); /**/

#if NEW_ATR && !NF_CISCO_DATA && !LFAP_METER /* Can see individual packets */

   if (fp->stdata != NULL &&  /* Flow has stream_data (>= 1 stream attrib) */
      sp != NULL) {  /* Use stream looked up via tuple */

      if (fp->stdata->pp_match_reqd) {

#if PP_DEBUG_XX >= 2
   log_msg(LOG_WARNING, 0,
      "count(1) stream: sp=%d, fp=%d, dbits=%x, order=%d, dirn=%d, d[0]=%d",
      stream_nbr(sp), flow_nbr(fp), fp->distrib_bits, flow_order,
      p_p->opp_dir, sp->flow_order[0]);
#endif

         if (p_p->TransAddrType == PT_TCP) {
            if (cep->type_pp == 0) {
               cep->delta_ok = 0;
               cep->type_pp = PP_OTHER;
#if TCP_PKT_DATA
               if ((pd = get_pktdata()) != NULL) {
                  ThisFlags = ntohs(p_p->pi.tcp.th.flags);
                  tlen = p_p->pi.tcp.tcp_len - tcp_dataoffset(ThisFlags);
                  ThisSeq = ntohl(p_p->pi.tcp.th.seqnum);
                  ThisAck = ntohl(p_p->pi.tcp.th.acknum);
                  pd->ts = cep->at_us;  pd->pkt_len = p_p->p_len;
                  pd->pp_type = PP_TCP;
                  pd->direction = flow_order;
                  pd->pi.tcp.ack = ThisAck;
                  pd->pi.tcp.exp_seq = ThisSeq + tlen;
                  if (ThisFlags & (tcp_FlagSYN | tcp_FlagFIN))
                     ++pd->pi.tcp.exp_seq;  /* SYN, FIN consume one seq nbr */
                  pd->pi.tcp.d_len = tlen;
                  pd->pi.tcp.pkt_nbr = ++sp->pkt_nbr;
                  pd->pi.tcp.flags = ThisFlags & tcp_FlagALL;

                  cep->delta_ok = tcp_turnaround(cep, sp, pd);
                  tcp_stream_update(cep, sp, pd);
                  }
#endif  /* TCP_PKT_DATA */
               }

            tcp_flow_update(cep, fp, sp, pd);
            }

         else if (p_p->TransAddrType == PT_UDP) {
            if (p_p->High.TransAddress == htons(WNP_DOMAIN) ||
                  p_p->Low.TransAddress == htons(WNP_DOMAIN)) {
               if (cep->type_pp == 0) {
#if PP_DEBUG_XX >= 2
   log_msg(LOG_WARNING, 0,
   "  count(2) DNS pkt: sp=%d, info_sz=%d, ident=%x, info=%02x%02x, dir=%d",
      stream_nbr(sp), p_p->pktinfo_sz, p_p->pi.udp.dns.ident,
      p_p->pi.udp.dns.p0,p_p->pi.udp.dns.p1, flow_order);
#endif

#if LOG_DNS_PKTS
   log_dns(fp,p_p,cep->at_us);
#endif
                  cep->type_pp = PP_UDP_DNS;
                  cep->delta_ok = 0;
                  cep->lost_to = cep->lost_from =
                     cep->oflo_to = cep->oflo_from = 0;
                  cep->type_pp = PP_UDP_DNS;
                  if (p_p->pi.udp.dns.p0 & dns_resp) {  /* DNS response */
#if PP_DEBUG_XX >= 2
   log_msg(LOG_WARNING, 0,
      "  count(3) DNS response: sp=%d", stream_nbr(sp));
#endif
#if PP_DNS_TRACE
   log_msg(LOG_WARNING, 0, "DNS response(1): sp=%d, id=%04x, order=%d",
      stream_nbr(sp), p_p->pi.udp.dns.ident, flow_order);
#endif
                     mutex_lock("cd1", &pktdata_q_mutex);
                     for (pd = sp->pq; pd != NULL; pd = pd->next) {
                        if (pd->pi.udp.dns.ident == p_p->pi.udp.dns.ident &&
                              pd->direction != flow_order) {
                           cep->us_delta = cep->at_us - pd->ts;
                           if (pd == sp->pq) sp->pq = pd->next;
                           else lpd->next = pd->next;
                           --sp->pq_len;
# if PP_DNS_TRACE || DNS_ROOT_DEBUG
   log_msg(LOG_WARNING, 0,
      "   --- DNS response matched, stream %x, pq_len=%d, dir=%d", 
     sp, sp->pq_len, flow_order);
   matched = 1;
   quack(99);
#endif
                           free_pktdata(pd);
                           cep->delta_ok = 1;
                           break;
                           }
                        lpd = pd;
#if PP_DEBUG_XX >= 2
   if (cep->delta_ok)
      log_msg(LOG_WARNING, 0,"    count(4) DNS match: sp=%d", stream_nbr(sp));
#endif
                        }
                     mutex_unlock("cd1", &pktdata_q_mutex);
                     if (pd == NULL) {  /* Unrequested response; count as lost */
                        if (flow_order == MFWD) ++cep->lost_to;
                        else ++cep->lost_from;
                        }
# if DNS_ROOT_DEBUG
   if (!matched) log_msg(LOG_WARNING, 0,
      "   ??? DNS response NOT matched, stream %x, pq_len=%d, dir=5D",
      sp, sp->pq_len, flow_order);
#endif
                     }
                  else {  /* DNS query */
#if PP_DEBUG_XX >= 2
   log_msg(LOG_WARNING, 0, "  count(3) DNS query: sp=%d", stream_nbr(sp));
#endif
#if PP_DNS_TRACE
   log_msg(LOG_WARNING, 0, "DNS query(1): sp=%d, id=%04x, order=%d",
      stream_nbr(sp), p_p->pi.udp.dns.ident, flow_order);
#endif
                     if ((pd = get_pktdata()) != NULL) {
                        pd->ts = cep->at_us;  pd->pkt_len = p_p->p_len;
                        pd->direction = flow_order;
                        pd->pi.udp.dns.ident = p_p->pi.udp.dns.ident;
# if DNS_ROOT_DEBUG
   log_msg(LOG_WARNING, 0,
      "   +++ DNS request queued, stream %x, pq_len=%d,", sp, sp->pq_len);
#endif

                        other_stream_update(cep, sp, pd);
                        }

                     }
                  }

               }
            other_flow_update(cep, fp, sp, pd);

            /* else if (fp->fk.High.TransAddress[1] != .. Other UDP pairs */
            }
         else if (p_p->TransAddrType == PT_ICMP) {
            if (cep->type_pp == 0) {
#if PP_DEBUG_XX >= 2
   log_msg(LOG_WARNING, 0,
      "count(5) ICMP pkt: sp=%d, info_sz=%d, ident=%04x, sequence=%04x",
      stream_nbr(sp), p_p->pktinfo_sz, p_p->pi.icmp.ident, p_p->pi.icmp.sequence);
#endif

            cep->delta_ok = 0;
            cep->type_pp = PP_ICMP_ECHO;
            cep->lost_to = cep->lost_from =
               cep->oflo_to = cep->oflo_from = 0;
            if (p_p->Low.TransAddress == 0) {  /* Echo response */
#if PP_DEBUG_XX >= 2
if (p_p->pi.icmp.sequence != 0)
   log_msg(LOG_WARNING, 0,
      "  count(6) ICMP echo response: sp=%d, info_sz=%d, ident=%04x, sequence=%04x, dirn=%d",
      stream_nbr(sp), p_p->pktinfo_sz, p_p->pi.icmp.ident, p_p->pi.icmp.sequence, flow_order);
#endif
                  mutex_lock("ci1", &pktdata_q_mutex);
  quack(2223);
                  for (pd = sp->pq; pd != NULL; pd = pd->next) {
#if PP_DEBUG_XX >= 2
   log_msg(LOG_WARNING, 0,
      "  count(7) ICMP queue: ident=%04x, sequence=%04x, dirn=%d",
	   pd->pi.icmp.ident, pd->pi.icmp.sequence, pd->direction);
#endif
                     if (pd->pi.icmp.ident == p_p->pi.icmp.ident &&
                           pd->pi.icmp.sequence == p_p->pi.icmp.sequence &&
                           pd->direction != flow_order) {
                        cep->us_delta = cep->at_us - pd->ts;
#if PP_DEBUG_XX >= 2
   log_msg(LOG_WARNING, 0,
      "  count(8) ICMP echo response MATCHED <<<<<<<: sp=%d, cep->at_us=%.0f, delta=%.0f",
      stream_nbr(sp), cep->at_us, cep->us_delta);
#endif
                        if (pd == sp->pq) sp->pq = pd->next;
                        else lpd->next = pd->next;
                        --sp->pq_len;
                        free_pktdata(pd);
                        cep->delta_ok = 1;
                        break;
                        }
                     lpd = pd;
                     }
                  mutex_unlock("ci1", &pktdata_q_mutex);
                  }
               else if (p_p->Low.TransAddress == 8) {  /* Echo request */
#if PP_DEBUG_XX >= 2
if (p_p->pi.icmp.sequence != 0)
   log_msg(LOG_WARNING, 0,
      "  count(9) ICMP echo request: sp=%d, info_sz=%d, ident=%04x, sequence=%04x, dirn=%d, cep->at_us=%.0f",
      stream_nbr(sp), p_p->pktinfo_sz, p_p->pi.icmp.ident, p_p->pi.icmp.sequence, flow_order, cep->at_us);
#endif
                  if ((pd = get_pktdata()) != NULL) {
                     pd->ts = cep->at_us;  pd->pkt_len = p_p->p_len;
                     pd->direction = flow_order;
                     pd->pi.icmp.ident = p_p->pi.icmp.ident;
                     pd->pi.icmp.sequence = p_p->pi.icmp.sequence;

                     other_stream_update(cep, sp, pd);
                     }
                  }
               /* else if (fp->fk.High.TransAddress[1] ==  .. Other ICMP pairs */
               }

            other_flow_update(cep, fp, sp, pd);
            }
         /* else if (p_p->TransAddrType == .. other protocols */

         if (cep->delta_ok) ++n_delta_OKs;  /* To verify pp matching */
         }
      else {  /* Not pp matching */
         }
      }

#if PP_DEBUG_XX >= 2
if (fp->stdata)
   log_msg(LOG_WARNING, 0,
      "  count(10): order=%d, match_rq=%d, pp_type=%d, delta_ok=%d",
      flow_order, fp->stdata->pp_match_reqd, cep->type_pp, cep->delta_ok);
#endif

#if PP_DEBUG_XX == 1
 if (fp->distrib_list == NULL && flow_nbr(fp) != 1)
   log_msg(LOG_WARNING, 0,
      "  count(10a): flow_nbr=%d, fd->distrib == NULL ???", flow_nbr(fp));
#endif

   for (d = fp->distrib_list; d != NULL; d = d->next) {  /* Update distribs */
      if (flow_order == MREV) {
         switch (d->selector) {
         case FTFROMPACKETSIZE:
            bump_dist(d, p_p->p_len);
            break;
         case FTFROMINTERARRIVALTIME:
            if (fp->stdata && fp->stdata->use_flow_iats && fp->stdata->FromSeen) {
               iat_delta_us = cep->at_us - fp->stdata->LastFromTime;
               bump_dist(d, iat_delta_us);
               }
            else if (sp->FromSeen) {
               iat_delta_us = cep->at_us - sp->LastFromTime;
               bump_dist(d, iat_delta_us);
               }
            break;
# if STREAM_TO_FLOW
         case FTTOSTREAMS:
         case FTFROMSTREAMS:
         case FTFLOWTIME:
            if (sp == NULL) continue;  /* No stream yet */
            if (sp->str_flowp != NULL) {  /* Already created a flow */
               sp->str_flowp->LastTime = fp->ntm_LastTime;
               add_Bit32_to_counter64(sp->str_flowp->DownOctets,p_p->p_len);
               incr_counter64(sp->str_flowp->DownPDUs); 
               }
            else if (fp->sftime != 0 &&  /* Making flows for LRS */
                  sp->LastTime-sp->FirstTime >= fp->sftime) {
               create_flow_from_stream(sp, fp);
               if (sp->str_flowp != NULL) {  /* Bump counts */
                  sp->str_flowp->LastTime = fp->ntm_LastTime;
                  add_Bit32_to_counter64(sp->str_flowp->DownOctets,p_p->p_len);
                  incr_counter64(sp->str_flowp->DownPDUs); 
                  }
               }
            break;
# endif
         case FTTOTURNAROUNDTIME1:
/*         case FTTOTURNAROUNDTIME4: */
#if PP_DEBUG_XX >= 2
if (cep->delta_ok)
   log_msg(LOG_WARNING, 0,
      "  count(11) MREV: delta_ok=%d, pp_type=%02X, d->PP_type=%02X, count=%02X",
      cep->delta_ok, cep->type_pp, d->PP_Type, d->PP_Type & cep->type_pp);
#endif
            if (cep->delta_ok) {
               if ((p_p->TransAddrType == PT_TCP &&
                     d->PP_Type & cep->type_pp & ~PP_TCP) ||  /* Bits in common */
                     (p_p->TransAddrType != PT_TCP &&
                     d->PP_Type == cep->type_pp)) {  /* Exact match */
                  bump_dist(d, cep->us_delta);
#if PP_DEBUG_XX >= 2
   log_msg(LOG_WARNING, 0,
      "  count(12): counted, attrib=%d, delta=%.0f, fp=%x, dbits=%x\n",
      d->selector, cep->us_delta, fp, fp->distrib_bits);
#endif
                  }
#if PP_DEBUG_XX >= 2
   else log_msg(LOG_WARNING, 0,
      "  count(13): NOT counted, attrib=%d, delta=%.0f\n", d->selector, cep->us_delta);
#endif
	       }
            break;
/*          case FTFROMTRAINLENGTH: */
         case FTFROMTURNAROUNDTIME4:
            if (d->FilterUpper != 0 && p_p->p_len > d->FilterUpper)
               in_train = 0;
            else if (d->FilterLower != 0 && p_p->p_len <= d->FilterLower)
               in_train = 0;
            else in_train = 1;

            if (d->last_was_in_train) {
               if (in_train) ++d->trainlength;
               else {  /* End of train */
                  bump_dist(d, d->trainlength);
                  d->trainlength = 0;
                  }
               }
            d->last_was_in_train = in_train;
            break;
            }
         }
      else {  /* MFWD */
         switch (d->selector) {
         case FTTOPACKETSIZE:
            bump_dist(d, p_p->p_len);
            break;
         case FTTOINTERARRIVALTIME:
            if (fp->stdata && fp->stdata->use_flow_iats && fp->stdata->ToSeen) {
               iat_delta_us = cep->at_us - fp->stdata->LastToTime;
#if 0
if (iat_delta_us > 1.0e6)  //03
   printf("  count(MFWD): ToSeen=%d, at_us=%f, LastTo=%f, iat_delta_us=%f\n",
      sp->ToSeen, cep->at_us, fp->stdata->LastToTime, iat_delta_us);  //03
#endif
               bump_dist(d, iat_delta_us);
               }
            else if (sp->ToSeen) {
               iat_delta_us = cep->at_us - sp->LastToTime;
               bump_dist(d, iat_delta_us);
               }
            break;
# if STREAM_TO_FLOW
         case FTTOSTREAMS:
         case FTFROMSTREAMS:
         case FTFLOWTIME:
            if (sp == NULL) continue;  /* No stream yet */
            if (sp->str_flowp != NULL) {  /* Already created a flow */
               sp->str_flowp->LastTime = fp->ntm_LastTime;
               add_Bit32_to_counter64(sp->str_flowp->UpOctets,p_p->p_len);
               incr_counter64(sp->str_flowp->UpPDUs); 
               }
            else if (fp->sftime != 0 &&  /* Making flows for LRS */
                  sp->LastTime-sp->FirstTime >= fp->sftime) {
               create_flow_from_stream(sp, fp);
               if (sp->str_flowp != NULL) {  /* Bump counts */
                  sp->str_flowp->LastTime = fp->ntm_LastTime;
                  add_Bit32_to_counter64(sp->str_flowp->UpOctets,p_p->p_len);
                  incr_counter64(sp->str_flowp->UpPDUs); 
                  }
               }
            break;
# endif
         case FTFROMTURNAROUNDTIME1:
/*         case FTFROMTURNAROUNDTIME4: */
#if PP_DEBUG_XX >= 2
if (cep->delta_ok)
   log_msg(LOG_WARNING, 0,
      "  count(14) MFWD: delta_ok=%d, pp_type=%02X, d->PP_type=%02X, count=%02X",
      cep->delta_ok, cep->type_pp, d->PP_Type, d->PP_Type & cep->type_pp);
#endif
            if (cep->delta_ok) {
               if ((p_p->TransAddrType == PT_TCP &&
                     d->PP_Type & cep->type_pp & ~PP_TCP) ||  /* Bits in common */
                     (p_p->TransAddrType != PT_TCP &&
                     d->PP_Type == cep->type_pp)) {  /* Exact match */
                  bump_dist(d, cep->us_delta);
                  }
	       }
            break;
/*         case FTTOTRAINLENGTH: */
         case FTTOTURNAROUNDTIME4:
            if (d->FilterUpper != 0 && p_p->p_len > d->FilterUpper)
               in_train = 0;
            else if (d->FilterLower != 0 && p_p->p_len <= d->FilterLower)
               in_train = 0;
            else in_train = 1;

            if (d->last_was_in_train) {
               if (in_train) ++d->trainlength;
               else {  /* End of train */
                  bump_dist(d, d->trainlength);
                  d->trainlength = 0;
                  }
               }
            d->last_was_in_train = in_train;
            break;
            }
         }
      }

   if (flow_order == MREV) {  /* Set new LastTimes after updating distribs */
# if TCP_PKT_DATA
      if (cep->delta_ok && p_p->TransAddrType == PT_TCP &&
            (cep->type_pp & PP_TCP) != 0) {
         if (cep->type_pp & (PP_OK_SYNACK | PP_OK_SYNDATA)) {
            sp->FromSynRTT = cep->us_delta;
#  if TCP_TAT_TRACE == 1
            print_stream_name(stdout, "SYN+ACK|DATA", sp);
            printf("\n   To: delta=%.0f, FromSynRTT=%u, OK=%02x\n",
                  cep->us_delta, sp->FromSynRTT, pp_type);
#  endif
            }
         else {
            ++sp->FromDataPoints;  sp->FromTotalRTT += cep->us_delta;
/*       printf("+++From: dp=%u, tr=%.03f\n", sp->FromDataPoints, sp->FromTotalRTT/1000.0);  /**/
            }
         }
# endif  /* TCP_PKT_DATA */
      if (fp->stdata) {
         fp->stdata->LastFromTime = cep->at_us;  fp->stdata->FromSeen = 1;
         }
      }
   else {  /* MFWD */
# if TCP_PKT_DATA
      if (cep->delta_ok && p_p->TransAddrType == PT_TCP &&
            (cep->type_pp & PP_TCP) != 0) {
         if (cep->type_pp & (PP_OK_SYNACK | PP_OK_SYNDATA)) {
            sp->ToSynRTT = cep->us_delta;
#  if TCP_TAT_TRACE == 1
            print_stream_name(stdout, "SYN+ACK|DATA", sp);
            printf("\n   To: delta=%.0f, ToSynRTT=%u, OK=%02x\n",
               cep->us_delta, sp->ToSynRTT, cep->type_pp);
#  endif
            }
         else{
            ++sp->ToDataPoints;  sp->ToTotalRTT += cep->us_delta;
/*	 printf("+++To: dp=%u, tr=%.03f\n", sp->ToDataPoints, sp->ToTotalRTT/1000.0);  /**/
            }
         }
#endif  /* TCP_PKT_DATA */
      if (fp->stdata) {
         fp->stdata->LastToTime = cep->at_us;  fp->stdata->ToSeen = 1;
         }
      }

#endif  /* NEW_ATR && !NF_CISCO_DATA   Individual packets */

#if OAM
   if (fp->stdata != NULL && fp->stdata->OAMtrigger != 0) {
# if OAM_TRACE
      printf("OAM trigger observed, flow=%d, trigger=%d\n",
         flow_nbr(fp), fp->stdata->OAMtrigger);
      fflush(stdout);
# endif
      build_oam_recs(fp->stdata->OAMtrigger, p_p, fp->FlowRuleSet);
      }
#endif
   }

#define count(sp, cfp, cdir)  {                       \
   pmi->fx[pmi->n_flows] = flow_nbr(cfp);             \
   pmi->flow_order[pmi->n_flows++] = (cdir == MFWD);  \
   }

void new_pkt_monitor(struct pkt *p_p)
{
   int x, r;
   struct rtinf_rec *rip;
   struct flow_key sd_key, ds_key;
   Bit32 sd_hash, ds_hash;
   struct search_result sd_pointers, ds_pointers;
   struct pme_env pme;
   struct flow *f;
   struct stream *sp = p_p->sp;
   struct pkt_match_info *pmi;

   /* printf("pkt_mon(): ");  dump_bp(p_p);  /**/

#if PKT_DEBUG
  if (p_p->pkt_state != 1) {
    printf("pkt_monitor(1): not got, pbp=%x\n", p_p);  fflush(stdout);
    }
#endif

   pmi = p_p->opp_dir ? &sp->r_pmi : &sp->f_pmi;
   pmi->n_flows = 0;
   ntm_semaphore_down("npm", &r_rts_sem);

   for (x = running_rts; x != 0; x = rip->next_running_rt) {
      rip = &ri[x-1];

      r = match(1, p_p, x,&pme, &p_p->Low,&p_p->High);
      if (r == MIGNORE)
         continue;  /* Try next rule set */
         /* match() builds search_key from p_p */
      else if (r == MFAIL) {  /* S->D Fail */
         r = match(0, p_p, x,&pme, &p_p->High,&p_p->Low);
         if (r != MSUC)
            continue;  /* Ignore or fail */
         else { 
            build_search_key(&ds_hash,&ds_key,
               x,&pme, p_p, &p_p->High,&p_p->Low);
            if ((f = current(&ds_pointers, ds_hash,&ds_key)) != NULL) {
               count(sp, f,MREV);
               }
            else {
               f = create(&ds_pointers, x,&pme, &ds_key, p_p);
               if (f == NULL) continue;  /* Coudn't create flow */
               f->owisFWD = 0;
               count(sp, f,MREV);
               }
            }
         }
      else {  /* S->D Suc */
         build_search_key(&sd_hash,&sd_key,
            x,&pme, p_p, &p_p->Low,&p_p->High);
         if ((f = current(&sd_pointers, sd_hash,&sd_key)) != NULL) {
            count(sp, f,MFWD);
            }
         else {
            build_search_key(&ds_hash,&ds_key,
               x,&pme, p_p, &p_p->High,&p_p->Low);
            if ((f = current(&ds_pointers, ds_hash,&ds_key)) != NULL) {
               count(sp, f,MREV);
               }
            else {
               f = create(&sd_pointers, x,&pme, &sd_key, p_p);
               if (f == NULL) continue;  /* Coudn't create flow */
               f->owisFWD = 1;
               count(sp, f,MFWD);
               }
            }
         }
      }

   ntm_semaphore_up("npm", &r_rts_sem);
   }

void process_pkt(struct pkt *p_p)
{
   struct stream *sp = p_p->sp;
   struct pkt_match_info *pmi;
   struct flow *fp;
   struct stream_data *sdp;
   struct count_env ce;
   int j;

   if (p_p->opp_dir) {
      pmi = &sp->r_pmi;
      ++sp->FromPDUs;
      sp->FromOctets += p_p->p_len;
      if (sp->FromSeen == 0) {
         sp->FirstFromTime = sp->StrLastTime;  sp->FromSeen = 1;
         }
      sp->LastFromTime = sp->StrLastTime;
      sp->FromECN = p_p->ECNCodeBits;
      }
   else {
      pmi = &sp->f_pmi;
      ++sp->ToPDUs;
      sp->ToOctets += p_p->p_len;
      if (sp->ToSeen == 0) {
         sp->FirstToTime = sp->StrLastTime;  sp->ToSeen = 1;
         }
      sp->LastToTime = sp->StrLastTime;
      sp->ToECN = p_p->ECNCodeBits;
      }
#if 0  /* ^_^ */
   if (sp->state != SS_IGNORE && (
       ntohs(sp->sfn.SrcTransAddr) == 53 ||
       ntohs(sp->sfn.DestTransAddr) == 53 )
       ) {
      printf("process_pkt(1): sp=%d, dirn=%d: ",
         stream_nbr(sp), !p_p->opp_dir);
      daddr(sp->sfn.SrcPeerAddr);  printf(p_p->opp_dir ? "<-" : "->");
      daddr(sp->sfn.DestPeerAddr);  printf(" ");
      dport(sp->sfn.SrcTransAddr);  printf(p_p->opp_dir ? "<-" : "->");
      dport(sp->sfn.DestTransAddr);
      printf(" pt=%d, tt=%d.  pkts=%u\n",
         sp->sfn.PeerType,  sp->sfn.TransType, sp->ToPDUs+sp->FromPDUs);
      fflush(stdout);
      }
#endif
//20   mutex_lock("pkp", &stream_state_mutex);
   mutex_lock("pkp", &stream_mutex);
   --sp->n_pkts_queued;

#if PKT_DEBUG
   if (p_p->pkt_state != 1) {
      printf("process_pkt(1): not got, pme_pp=%x, state=%d\n", p_p, p_p->pkt_state);
      quack(901);
      }
#endif

   if (sp->state == SS_IGNORE || (  /* Existing stream, ignored */
      sp->state == SS_MATCHED &&    /* or matched and */
      pmi->n_flows != 0 &&          /* seen flows in this direction and */
         (sp->oa_count == 0 ||  /* No 'other address' attribs or */
         sp->pm_state == PM_NOTEST ||  /* No app testing for this flow or */
         sp->pm_state == PM_COUNTING)) ) {  /* Flow won't change now */
      ce.type_pp = 0;  /* Initialise pp_matching globals */
      for (j = 0; j != pmi->n_flows; ++j)
         actual_count(&ce, flow_ptr(pmi->fx[j]), pmi->flow_order[j], p_p);
//20      mutex_unlock("pkp", &stream_state_mutex);
      mutex_unlock("pkp", &stream_mutex);
      n_matches += pmi->n_flows;  n_cached_matches += pmi->n_flows;
#if PKT_DEBUG
      if (p_p->pkt_state != 1) {
         printf("process_pkt(2): not got, pme_pp=%x, state=%d\n",
            p_p, p_p->pkt_state);
         quack(902);
      }
#endif
      free_pkt(p_p);
      return;
      }
   else {  /* Must match packet */
      if (sp->state != SS_UNKNOWN && sp->state != SS_MATCHED
            && sp->state != SS_IGNORE) {
//20         mutex_unlock("pkp", &stream_state_mutex);
         mutex_unlock("pkp", &stream_mutex);
         if (sp->state != SS_IGNORE)  /* Queued pkt(s) for new stream */
            log_msg(LOG_WARNING, 0,
               "Attempt to match packet in state %d, stream %x, next_hc=%x!",
                sp->state, sp, sp->next_hc);
         free_pkt(p_p);
         return;
         }
//20      mutex_unlock("pkp", &stream_state_mutex);
      mutex_unlock("pkp", &stream_mutex);
      p_p->Low.ObsInterface = p_p->if_nbr;
      p_p->High.ObsInterface = 0;

      if (p_p->n_ports == 2) {  /* What will we do for 4-port Dag cards? */
	 if (p_p->port_nbr == 0) {
            p_p->Low.Interface = htons(p_p->if_nbr);
            p_p->High.Interface = htons(p_p->if_nbr + n_interfaces);
	    }
         else {
            p_p->Low.Interface = htons(p_p->if_nbr + n_interfaces);
            p_p->High.Interface = htons(p_p->if_nbr);
            }
         }
      else {
         p_p->Low.Interface = htons(p_p->if_nbr);
         if (n_interfaces == 2)  /* Assume they're a two-dir pair */ 
            p_p->High.Interface = p_p->if_nbr == 1 ? htons(2) : htons(1);
         else p_p->High.Interface = p_p->Low.Interface;
            /* NETFLOW uses Low.Interface for ntm_interface */
            /* Reading trace files will need a kludge too !!! */
         }
      p_p->Low.AdjType = p_p->adj_type;  p_p->High.AdjType = 0;

#if APP_DATA
      p_p->p2ptested = 0;
#endif
#if PKT_DEBUG
      if (p_p->pkt_state != 1) {
         printf("process_pkt(3): not got, pme_pp=%x, state=%d\n",
            p_p, p_p->pkt_state);
         quack(903);
         }
#endif
      new_pkt_monitor(p_p);  /* Makes array of flows in sp->fp[] */
#if 0  /* ^-^ */
      printf("process_pkt(4): sp=%d: n_flows=%d, order=(%d",
         stream_nbr(sp), pmi->n_flows, pmi->flow_order[0]);
      for (j = 1; j < pmi->n_flows; ++j) printf(" %d", pmi->flow_order[j]);
      printf("), fx=(%d", pmi->fx[0]);
      for (j = 1; j < pmi->n_flows; ++j) printf(" %d", pmi->fx[j]);
      printf(")\n");
#endif
      n_matches += pmi->n_flows;
#if PKT_DEBUG
      if (p_p->pkt_state != 1) {
         printf("process_pkt(5): not got, pme_pp=%x, state=%d\n",
            p_p, p_p->pkt_state);
         quack(904);
         }
#endif

      mutex_lock("pnp", &stream_mutex);
      if (sp->state == SS_MATCHED) {  /* Existing stream, sp->oa_count != 0 */
         ce.type_pp = 0;
         for (j = 0; j != pmi->n_flows; ++j)
            actual_count(&ce, flow_ptr(pmi->fx[j]), pmi->flow_order[j], p_p);
#if PKT_DEBUG
         if (p_p->pkt_state != 1) {
            printf("process_pkt(6): not got, pme_pp=%x, state=%d\n",
               p_p, p_p->pkt_state);
            quack(905);
            }
#endif
         mutex_unlock("pnp", &stream_mutex);
         free_pkt(p_p);
         return;
         }

//20      mutex_lock("pnp", &stream_state_mutex);
      if (pmi->n_flows == 0) {  /* No matches */
         sp->state = SS_IGNORE;
//20         mutex_unlock("pnp", &stream_state_mutex);
         mutex_unlock("pnp", &stream_mutex);
         free_pkt(p_p);
         return;
         }
      else {  /* At least one match */
         sp->state = SS_MATCHED;
         if (p_p->PeerAddrType == AT_IP4
# if V6
               || p_p->PeerAddrType == AT_IP6
# endif
               ) {
            sp->mx_pq_len = p_p->TransAddrType == PT_TCP ?
               MX_PQ_LEN_TCP : MX_PQ_LEN_OTHER;
            }
         sp->StrFirstTime = p_p->arrival_time;

# if TCP_TRACE
         sp->sf_nbr = ++sf_count;
# endif
# if TCP_TESTING
         print_stream_name(stdout, "start", sp);
         printf("\n");
# endif
//20         mutex_unlock("pnp", &stream_state_mutex);

         ce.type_pp = 0;
         for (sp->oa_count = j = 0; j != pmi->n_flows; ++j) {
            fp = flow_ptr(pmi->fx[j]);
            sp->oa_count += ri[fp->FlowRuleSet-1].oa_count;

            if ((sdp = fp->stdata) != NULL) {
               if (sdp->active_streams > sdp->mx_active_streams)
                  sdp->mx_active_streams = sdp->active_streams;
               }

            actual_count(&ce, fp, pmi->flow_order[j], p_p);
            }
         mutex_unlock("pnp", &stream_mutex);

#if PKT_DEBUG
         if (p_p->pkt_state != 1) {
            printf("process_pkt(7): not got, pme_pp=%x, state=%d\n",
               p_p, p_p->pkt_state);
            quack(906);
            }
#endif
         free_pkt(p_p);
         return;
         }
      }
   log_msg(LOG_ERR, 0, ">>> Reached bottom of process_pkt <<<");
   }


#define MAXPKTLEN  1526

#if !NF_CISCO_DATA && !LFAP_METER
   /* Front-end calls the following routines to get attrib values */

# if DECNET
int dn_node(Bit8 *dn_addr)
{
   return dn_addr[1]<<8 | dn_addr[0];
   }

#  define dn_area(node) (node >> 10)
#  define dn_host(node) (node & 0x3FF)
# endif

void unpack_dn_node(Bit8 *ap, const Bit8 *dn_addr)
{
   Bit8 dl = dn_addr[1];
   ap[0] = dl >> 2;    /* DECnet area */
   ap[1] = dl & 0x03;  ap[2] = dn_addr[0];  /* DECnet host */
   ap[3] = 0;
   }

void unpack_at_node(Bit8 *ap, const Bit8 *at_net, const Bit8 at_node)
{
   ap[0] = at_net[0];  ap[1] = at_net[1];
   ap[2] = at_node;
   ap[3] = 0;
   }

# define BAD_ADJ    1
# define BAD_PEER   2
# define BAD_TRANS  3

void zero_pkt_fields(int which, struct pkt *p_p)
{
   switch (which) {
   case BAD_ADJ:
      p_p->Low.AdjAddr_ms4 = p_p->High.AdjAddr_ms4 = 0;
      p_p->Low.AdjAddr_ls2 = p_p->High.AdjAddr_ls2 = 0;
      p_p->invalid_addrs |= INV_ADJ;
      break;
   case BAD_PEER:
# if QBYTE_PEER
      p_p->Low.PeerAddress.qbyte = p_p->High.PeerAddress.qbyte = 0;
# else
      memset(&p_p->Low.PeerAddress.byte,  0, PEER_ADDR_LEN+TRANS_ADDR_LEN);
      memset(&p_p->High.PeerAddress.byte, 0, PEER_ADDR_LEN+TRANS_ADDR_LEN);
# endif
      p_p->invalid_addrs |= INV_PEER;
      break;
   case BAD_TRANS:
      p_p->Low.TransAddress = p_p->High.TransAddress = 0;
      p_p->invalid_addrs |= INV_TRANS;
      break;
      }
   }

# if V6
int find_v6_type(int *prot, int *hx, const Bit8 *p, const int hl)
{
   int nh, nx;

   nh = p[6];  nx = 40;
   for (;;) {
      switch (nh) {
      case PT_6FRAG:  /* Fragment */
      case PT_6ESP:   /* Encapsulating Security Payload (ESP) RFC 2406 */
      case PT_6NONH:  /* No next header */
         *prot = nh;
         return 0;  /* Can't see into this packet */

      case PT_6ICMP:  /* ICMPv6 */
#  if V6_TRACE
         printf("ICMPv6 packet filtered and extracted\n");
#  endif
      default:
         *prot = nh;  *hx = nx;
         return 1;  /* Can see into this one */

      case PT_6HOP:   /* Hop-by-hop options */
      case PT_6ROUT:  /* Routing (Type 0) */
      case PT_6DEST:  /* Destination options */
      case PT_6AH:    /* Authentication (AH) RFC 2402 */
#  if V6_TRACE
         printf("??? v6 header type %d found ???\n", nh);
#  endif

         break;  /* Keep looking */
         }
      if (hl < nx) {
         *prot = -1;  /* Not enough header bytes to get Next Header */
         return 0;
         }
      nh = p[nx];
/*    nx += (p[nx]+1) * 8;  /* Original, WRONG!  RUS, 22 Jul 03 */
      nx += (p[nx+1] * 8);
      }
   }
# endif

int pkt_extract(struct pkt *p_p,
   const int type, const Bit8 *hdrp, const int hl)
{
   int prot, dtype, dx,tx, j;
   union decnet *dp;
# if DECNET
   int dlen, snode, dnode, area, host;
# endif
# if NF_OCX_BGP
   Subnet *my_subnet;
   Bit32 my_IP4addr;
   Bit16 my_ASN;
# endif
# if V6_TRACE
   char sa[IP6_ADDR_LEN*4], da[IP6_ADDR_LEN*4];


# if PKT_DEBUG
   if (p_p->pkt_state != 1) {
      printf("pkt_extract(1): not got, p_p=%x\n", p_p);  fflush(stdout);
      quack(201);
      }
# endif
   if (type == AT_IP6) printf(" IPv6 packet extracted: \n");
/*   else if (type == AT_IP4) printf(" IPv4 packet extracted:\n"); /**/
# endif
   p_p->invalid_addrs = 0;
   p_p->PeerAddrType = type;

   switch (type) {
# if V6
   case AT_IP6:  /* RFC 2460 is 'IPv6 Specification' */
      p_p->DSCodePoint = (ntohs(*(Bit16 *)&hdrp[0]) & 0x0FC0) >> 6;
      p_p->ECNCodeBits = (hdrp[1] & 0x30) >> 4;
      if (hl < 8+IP6_ADDR_LEN*2) {  /* Didn't get the peer addresses! */
         zero_pkt_fields(BAD_PEER, p_p);  return 0;
         }
      memcpy(&p_p->Low.PeerAddress, &hdrp[8], IP6_ADDR_LEN);  /* IPv6 source */
      memcpy(&p_p->High.PeerAddress,&hdrp[24],IP6_ADDR_LEN);  /* IPv6 dest */
# if V6_TRACE
      sfmt_v6addr(sa,(Bit8 *)&p_p->Low.PeerAddress);
      sfmt_v6addr(da,(Bit8 *)&p_p->High.PeerAddress);
      printf("   %s -> %s\n", sa,da);
# endif
      if (!find_v6_type(&prot,&dx, hdrp,hl)) {
         if (prot < 0)  /* Not enough bytes to determine protocol */
            p_p->TransAddrType = 0;
         else  /* Can't see into packet body */
            p_p->TransAddrType = prot;
         zero_pkt_fields(BAD_TRANS, p_p);  return 0;
         }
      
      switch (p_p->TransAddrType = prot) {
      case PT_TCP:
         if (hl < dx+4) {  /* Didn't get the IP ports! */
            zero_pkt_fields(BAD_TRANS, p_p);  return 0;
            }
         p_p->has_ports = 1;
         p_p->Low.TransAddress = *(Bit16 *)&hdrp[dx];  /* Source port */
         p_p->High.TransAddress = *(Bit16 *)&hdrp[dx+2];  /* Dest port */
         /* In **network** order within Bit16, to match rule values */

# if TCP_PKT_DATA
         if (hl < dx+4+sizeof(p_p->pi.tcp.th)) {  /* Didn't get tcp header */
            p_p->pi.tcp.tcp_len = 0;
            memset(&p_p->pi.tcp.th, 0, sizeof(p_p->pi.tcp.th));
            p_p->pktinfo_sz = 0;
            }
         else {
            memcpy(&p_p->pi.tcp.th, &hdrp[dx+4],
               p_p->pktinfo_sz = sizeof(p_p->pi.tcp.th));
            p_p->pi.tcp.tcp_len = ntohs(*(Bit16 *)&hdrp[2]) - dx;
            }
# endif
# if APP_DATA
         tx = dx + ((hdrp[dx+12] & 0xF0) >> 2);  /* TCP HLEN */
	 j = hl - dx;  /* Bytes of TCP data */
         if (j < 4) {
            /* Didn't get enough TCP data */
            p_p->pktinfo_sz = 0;
            }
         else {
            if (j > DATA_SZ) j = DATA_SZ;
            memcpy(p_p->pi.tcp.data, &hdrp[tx], p_p->pktinfo_sz = j);
            }
# endif
         break;
      case PT_UDP:
         if (hl < dx+4) {  /* Didn't get the IP ports! */
            zero_pkt_fields(BAD_TRANS, p_p);  return 0;
            }
         p_p->has_ports = 1;
         p_p->Low.TransAddress = *(Bit16 *)&hdrp[dx];  /* Source port */
         p_p->High.TransAddress = *(Bit16 *)&hdrp[dx+2];  /* Dest port */
         /* In **network** order within Bit16, to match rule values */

         if (hl < dx+8) /* Can't see any data bytes */
            p_p->pktinfo_sz = 0;
         else {
            j = hl - (dx+8);  /* Nbr of UDP data bytes */
            if (j < sizeof(p_p->pi.udp.dns)) {  /* Always want DNS */
               p_p->pktinfo_sz = 0;
               }
            else {
               if (j > DATA_SZ) j = DATA_SZ;
               memcpy(p_p->pi.udp.d.data, &hdrp[dx+8], p_p->pktinfo_sz = j);
               }
            }
         break;
      case PT_6ICMP:  /* RFC 2463 */
      case PT_ICMP:
         if (hl < dx+2) {  /* Didn't get the type,code fields */
            zero_pkt_fields(BAD_TRANS, p_p);  return 0;
            }
         p_p->has_ports = 0;
         p_p->Low.TransAddress = hdrp[dx];  /* Type field */
         p_p->High.TransAddress = hdrp[dx+1];  /* Code field */
         break;
      default:
         p_p->has_ports = 0;
         p_p->Low.TransAddress = p_p->High.TransAddress = 0;
         }
      return 1;
# endif  /* V6 */

   case AT_IP4:
      p_p->DSCodePoint = hdrp[1] >> 2;  /* Old TOS field, RFC 2474 */
      p_p->ECNCodeBits = hdrp[1] && 0x03;  /* RFC 3168, ECN for IP */
      if (hl < 12+IP4_ADDR_LEN*2) {  /* Didn't get the peer addresses! */
         zero_pkt_fields(BAD_PEER, p_p);  return 0;
         }
      prot = p_p->TransAddrType = hdrp[9];  /* IPv4 Protocol Type */
# if QBYTE_PEER
      p_p->Low.PeerAddress.qbyte = *(Bit32 *)&hdrp[12];  /* IPv4 source */
      p_p->High.PeerAddress.qbyte = *(Bit32 *)&hdrp[16];  /* IPv4 dest */
# else
      memcpy(&p_p->Low.PeerAddress,&hdrp[12],IP4_ADDR_LEN);  /* IPv4 source */
      memcpy(&p_p->High.PeerAddress,&hdrp[16],IP4_ADDR_LEN);  /* IPv4 dest */
# endif
# if PKT_DEBUG
   if (p_p->pkt_state != 1) {
      printf("pkt_extract(1.5): not got, p_p=%x\n", p_p);  fflush(stdout);
      quack(2015);
      }
# endif
      if ((ntohs(*(Bit16 *)&hdrp[6]) & 0x1FFF) != 0) {  /* Frag offset != 0 */
         zero_pkt_fields(BAD_TRANS, p_p);  /* Can't see into fragment */
         return 0;
         }
      dx = (hdrp[0] & 0x0F) << 2;  /* HLEN */
      /* Data offset.  HLEN = IPv4 header length in 32-bit units */

      switch (p_p->TransAddrType = prot) {
      case PT_TCP:
         if (hl < dx+4) {  /* Didn't get the IP ports! */
            zero_pkt_fields(BAD_TRANS, p_p);  return 0;
            }
         p_p->has_ports = 1;
         p_p->Low.TransAddress = *(Bit16 *)&hdrp[dx];  /* Source port */
         p_p->High.TransAddress = *(Bit16 *)&hdrp[dx+2];  /* Dest port */
         /* In **network** order within Bit16, to match rule values */
# if PKT_DEBUG
   if (p_p->pkt_state != 1) {
      printf("pkt_extract(2): not got, p_p=%x\n", p_p);  fflush(stdout);
      quack(202);
      }
# endif

#  if TCP_PKT_DATA
         if (hl < dx+4+sizeof(p_p->pi.tcp.th)) {  /* Didn't get tcp header */
            p_p->pi.tcp.tcp_len = 0;
            memset(&p_p->pi.tcp.th, 0, sizeof(p_p->pi.tcp.th));
            p_p->pktinfo_sz = 0;
            }
         else {
            memcpy(&p_p->pi.tcp.th, &hdrp[dx],
               p_p->pktinfo_sz = sizeof(p_p->pi.tcp.th));
            p_p->pi.tcp.tcp_len = ntohs(*(Bit16 *)&hdrp[2]) - dx;
            }
#  endif
# if APP_DATA
         tx = dx + ((hdrp[dx+12] & 0xF0) >> 2);  /* TCP HLEN */
	 j = hl - dx;  /* Bytes of TCP data */
         if (j < 4) /* Didn't get enough UDP data */
            p_p->pktinfo_sz = 0;
         else {
            if (j > DATA_SZ) j = DATA_SZ;
            memcpy(p_p->pi.tcp.data, &hdrp[tx], p_p->pktinfo_sz = j);
            }
# endif
         break;
      case PT_UDP:
# if PKT_DEBUG
   if (p_p->pkt_state != 1) {
      printf("pkt_extract(2.5): not got, p_p=%x\n", p_p);  fflush(stdout);
      quack(2025);
      }
# endif
         if (hl < dx+4) {  /* Didn't get the IP ports! */
            zero_pkt_fields(BAD_TRANS, p_p);  return 0;
            }
         p_p->has_ports = 1;
         p_p->Low.TransAddress = *(Bit16 *)&hdrp[dx];  /* Source port */
         p_p->High.TransAddress = *(Bit16 *)&hdrp[dx+2];  /* Dest port */
         /* In **network** order within Bit16, to match rule values */

         if (hl < dx+8) /* Can't see any data bytes */
            p_p->pktinfo_sz = 0;
         else {
            j = hl - (dx+8);  /* Nbr of UDP data bytes */
            if (j < sizeof(p_p->pi.udp.dns)) /* Didn't get enough UDP data */
               p_p->pktinfo_sz = 0;
            else {
               if (j > DATA_SZ) j = DATA_SZ;
               memcpy(p_p->pi.udp.d.data, &hdrp[dx+8], p_p->pktinfo_sz = j);
               }
            }
# if PKT_DEBUG
   if (p_p->pkt_state != 1) {
      printf("pkt_extract(3): not got, p_p=%x\n", p_p);  fflush(stdout);
      quack(203);
      }
# endif
# if PP_DEBUG_XX >= 2
   printf("\npkt_extract(2) UDP pkt: info_sz=%d, ",  p_p->pktinfo_sz);
   for (j = 0; j != p_p->pktinfo_sz; ++j) printf(" %02x", ((char *)&p_p->pi.udp)[j]);
   printf("\n");
# endif
         break;
      case PT_ICMP:
         if (hl < dx+2) {  /* Didn't get the type,code fields */
            zero_pkt_fields(BAD_TRANS, p_p);  return 0;
            }
         p_p->has_ports = 0;
         p_p->Low.TransAddress = hdrp[dx];  /* Type field */
         p_p->High.TransAddress = hdrp[dx+1];  /* Code field */

         if (hl < dx+4+sizeof(p_p->pi.icmp)) {  /* Not enough ICMP data */
            zero_pkt_fields(BAD_TRANS, p_p);
            memset(&p_p->pi.icmp, 0, sizeof(p_p->pi.icmp));
            p_p->pktinfo_sz = 0;
            return 0;
            }
         else {
            memcpy(&p_p->pi.icmp, &hdrp[dx+4],
               p_p->pktinfo_sz = sizeof(p_p->pi.icmp));
            }
         break;
      default:
         p_p->has_ports = 0;
         p_p->Low.TransAddress = p_p->High.TransAddress = 0;
         }

#  if OAM
      p_p->ip_id = *(Bit16 *)&hdrp[4];  /* IPv4 identification */
      p_p->High.PeerAddress.qbyte = *(Bit32 *)&hdrp[16];  /* IPv4 dest */
#  endif

#  if NF_OCX_BGP
      if (asn_lookup) {
         my_IP4addr = ntohl(p_p->Low.PeerAddress);
         if ((my_subnet = FindSubnet(ntohl(my_IP4addr))) != NULL) {
            my_ASN = use_owner_asns ? my_subnet->src_as : my_subnet->exit_as;
            p_p->Low.nf_ASN[0] = my_ASN >> 8;
            p_p->Low.nf_ASN[1] = my_ASN & 0xff;
            p_p->Low.nf_mask = bits_in_mask(my_subnet->mask);
            }
         else p_p->Low.nf_ASN[0] = p_p->Low.nf_ASN[1] = 0;
         my_IP4addr = ntohl(p_p->High.PeerAddress);
         if ((my_subnet = FindSubnet(ntohl(my_IP4addr))) != NULL) {
            my_ASN = use_owner_asns ? my_subnet->src_as : my_subnet->exit_as;
            p_p->High.nf_ASN[0] = my_ASN >> 8;
            p_p->High.nf_ASN[1] = my_ASN & 0xff;
            p_p->High.nf_mask = bits_in_mask(my_subnet->mask);
            }
         else p_p->High.nf_ASN[0] = p_p->High.nf_ASN[1] = 0;
         }
#  endif

# if PKT_DEBUG
   if (p_p->pkt_state != 1) {
      printf("pkt_extract(4): not got, p_p=%x\n", p_p);  fflush(stdout);
      quack(204);
      }
# endif
      return 1;
   case AT_NOVELL:
      p_p->TransAddrType = hdrp[5];  /* XNS packet type */
      if (hl < 18+IPX_ADDR_LEN) {  /* Didn't get the peer addresses! */
         zero_pkt_fields(BAD_PEER, p_p);  return 0;
         }
      memcpy(&p_p->High.PeerAddress,&hdrp[6],IPX_ADDR_LEN);  /* IPX dest net */
      memcpy(&p_p->Low.PeerAddress,&hdrp[18],IPX_ADDR_LEN);  /* IPX soure net */
      if (hl < 30) {  /* Didn't get the IPX sockets! */
         zero_pkt_fields(BAD_TRANS, p_p);  return 0;
         }
      p_p->has_ports = 1;
      p_p->High.TransAddress = *(Bit16 *)&hdrp[16];
         /* IPX dest socket */
      p_p->Low.TransAddress = *(Bit16 *)&hdrp[28];
         /* IPX source socket */
      return 1;
   case AT_ETHERTALK:
      if (hl < 10) {  /* Didn't get the peer addresses! */
         zero_pkt_fields(BAD_PEER, p_p);  return 0;
         }
      unpack_at_node(&p_p->Low.PeerAddress.byte, &hdrp[6],hdrp[9]);  /* AT source */
      unpack_at_node(&p_p->High.PeerAddress.byte, &hdrp[4],hdrp[8]);  /* AT source */
      if (hl < 13) {  /* Didn't get the AT sockets! */
         zero_pkt_fields(BAD_TRANS, p_p);  return 0;
         }
      p_p->TransAddrType =hdrp[12];  /* AT DDP protocol type */
      p_p->has_ports = 1;
      p_p->Low.TransAddress = (Bit16)hdrp[11];  /* AT source socket */
      p_p->High.TransAddress = (Bit16)hdrp[10];  /* AT dest socket */
      return 1;
   case AT_CLNS:
      p_p->TransAddrType = hdrp[5] & 0x1F;  /* CLNS packet type */
      p_p->has_ports = 0;
      p_p->Low.TransAddress = p_p->High.TransAddress = 0;
      memcpy(&p_p->High.PeerAddress,&hdrp[11],j = hdrp[10]);  /* Dest */
      if (j < NSAP_ADDR_LEN)
         memset(&p_p->High.PeerAddress + j, 0, NSAP_ADDR_LEN-j);
      memcpy(&p_p->Low.PeerAddress,&hdrp[12+j],dx = hdrp[11+j]);  /* Source */
      if (dx < NSAP_ADDR_LEN)
         memset(&p_p->Low.PeerAddress + dx, 0, NSAP_ADDR_LEN-dx);
      return 1;
   case AT_DECNET:
      p_p->has_ports = 0;
      p_p->Low.TransAddress = p_p->High.TransAddress = 0;
      dtype = hdrp[2];  /* DECnet packet type */
      if (dtype != 0x81) dp = (union decnet *)&hdrp[3];
      else {
         dtype = hdrp[3];
         dp = (union decnet *)&hdrp[4];
         }
      switch (p_p->TransAddrType = dtype & 0x0F) {
      case 0x06:  /* Data */
      case 0x0E:  /* Data + discard flag */
         unpack_dn_node(&p_p->Low.PeerAddress.byte, dp->d.src_dn_addr);
         unpack_dn_node(&p_p->High.PeerAddress.byte, dp->d.dest_dn_addr);
# if DECNET
         dnode = dn_node(dp->d.dest_dn_addr);
         snode = dn_node(dp->d.src_dn_addr);
         fprintf(logfl,"Data to %d.%d from %d.%d\n",
            dn_area(dnode),dn_host(dnode),
            dn_area(snode),dn_host(snode) );
# endif
         break;
      case 0x07:  /* Level 1 routing */
         unpack_dn_node(&p_p->Low.PeerAddress.byte, dp->l1r.src_dn_addr);
# if DECNET
         snode = dn_node(dp->l1r.src_dn_addr);
         fprintf(logfl,"Level 1 routing from %d.%d\n",
            dn_area(snode),dn_host(snode));
# endif
         break;
      case 0x09:  /* Level 2 routing */
         unpack_dn_node(&p_p->Low.PeerAddress.byte, dp->l2r.src_dn_addr);
# if DECNET
         snode = dn_node(dp->l2r.src_dn_addr);
         fprintf(logfl,"Level 2 routing from %d.%d\n",
            dn_area(snode),dn_host(snode));
# endif
         break;
      case 0x0B:  /* Router hello */
         unpack_dn_node(&p_p->Low.PeerAddress.byte, dp->rh.src_dn_addr);
# if DECNET
         snode = dn_node(dp->rh.src_dn_addr);
         dnode = dn_node(dp->rh.rtr_dn_addr);
         fprintf(logfl,"Router hello from %d.%d, other router %d.%d\n",
            dn_area(snode),dn_host(snode),
            dn_area(dnode),dn_host(dnode) );
# endif
         break;
      case 0x0D:  /* Endnode hello */
         unpack_dn_node(&p_p->Low.PeerAddress.byte, dp->eh.src_dn_addr);
# if DECNET
         snode = dn_node(dp->eh.src_dn_addr);
         dnode = dn_node(dp->eh.rtr_dn_addr);
         fprintf(logfl,"Endnode hello from %d.%d, designated router %d.%d\n",
            dn_area(snode),dn_host(snode),
            dn_area(dnode),dn_host(dnode) );
# endif
         break;
      default:  /* Unknown DECnet type */
         scpos(0,scr_lrow);
         printf("\nDN pkt type %02x: ", dtype);
         for (j=0; j != 16; ++j) printf(" %02x",hdrp[j]);
         printf("\n");
         for (j=16; j != 34; ++j) printf(" %02x",hdrp[j]);
         printf("\n");
# ifdef TESTING
         if (!logfl) open_log();
         fprintf(logfl, "\nDN pkt type %02x: ", dtype);
         for (j=17; j != 34; ++j) fprintf(logfl," %02x",hdrp[j]);
         fprintf(logfl, "\n");
# endif
         }
      return 0;
      }
   return 0;
   }

void other_extract(struct pkt *p_p,
   const Bit16 ethertype, const Bit16 lsap)
{
   int bad_ip = ethertype == 0x0800 || ethertype == 0x86DD;  /* v4 or v6 */
   p_p->PeerAddrType = AT_OTHER;
   memset(&p_p->Low.PeerAddress.byte, 0, sizeof(p_p->Low.PeerAddress));
   memset(&p_p->High.PeerAddress.byte, 0, sizeof(p_p->High.PeerAddress));
   p_p->Low.PeerAddress.trans = htons((Bit16)ethertype);  /* Blue Book type */
   p_p->High.PeerAddress.trans = htons(bad_ip ? (Bit16)p_p->invalid_addrs
      : (Bit16)lsap);  /* 802.2 LSAP */
   p_p->has_ports = p_p->TransAddrType = 0;
   p_p->Low.TransAddress = p_p->High.TransAddress = 0;
   }

int handle_pkt(struct pkt *p_p,
   Bit16 ether_type, Bit16 lsap, const Bit8 *ethp, 
   Bit8 *p, int pl)  /* What's left after stripping encapsulation */
{
   struct snap_hdr *sh;
   int r = 0, first_vlan = 1;
   Bit32 t_seconds;

# if V6_DEBUG
   printf("handle packet\n");
# endif

   /* Handle any data-link-layer headers .. */
   while (ether_type == 0x8100) {  /* 802.1q frame (can be nested) */
      if (first_vlan) {
         p_p->Priority = p[0] >> 5;
         p_p->VLANid = *(Bit16 *)&p[0] & 0x0FFF;
            /* In *network* order within Bit16, to match rule values */
         first_vlan = 0;
         }
      p += 2;  /* Move past the 802.1q tag */
      ether_type = ntohs(*(Bit16 *)p);
      p += 2;  pl -= 4;  /* Move past the Len/Etype */
      if (ether_type <= 1500) {  /* 802.3 packet */
         sh = (struct snap_hdr *)p;
         if ((lsap = ntohs(sh->lsap)) == 0xAAAA) {  /* SNAP packet */
            ether_type = ntohs(sh->etype);
            p += sizeof(struct snap_hdr);  pl -= sizeof(struct snap_hdr);
            }
         else ether_type = 0;  /* Other 802.2 packet */
         }
      else lsap = 0;  /* Blue book packet */
      /*printf("====> VLANid=%04x, Priority=%d, ether_type=%04x\n",
         p_p->VLANid, p_p->Priority, ether_type);  fflush(stdout);  /**/
      }

   /* Handle any MPLS shim .. */
   if (ether_type == 0x8847) {  /* MPLS unicast */
      /* ether_type == 0x8848  /* MPLS multicast (!) */
      while (p[2] & 0x01 == 0) {  /* Find bottom of label stack */
         p += 4; pl -= 4;
         }
      p += 4; pl -= 4;
      ether_type = (p[0] & 0xF0) == 0x40 ? 0x0800 : 0x08DD;
      }

   switch (ether_type) {
   case 0x0800:
      if ((p[0] & 0xF0) == 0x40) {
         if (use_ip_length) p_p->p_len = ntohs(*(Bit16 *)&p[2]);
         r = pkt_extract(p_p, AT_IP4, p,pl);
         }
      break;
# if V6
   case 0x86DD:  /* RFC 2464, IPv6 packets over Ethernet */
      if ((p[0] & 0xF0) == 0x60) {
         if (use_ip_length) p_p->p_len = ntohs(*(Bit16 *)&p[4]);
         r = pkt_extract(p_p, AT_IP6, p,pl);
         }
      break;
# endif
   case 0x809B:
      if (proto_reqd[AT_ETHERTALK]) r = pkt_extract(p_p, AT_ETHERTALK, p,pl);
      break;
   case 0x8137:
      if (proto_reqd[AT_NOVELL]) r = pkt_extract(p_p, AT_NOVELL, p,pl);
      break;
   case 0x6003:
      if (proto_reqd[AT_DECNET]) r = pkt_extract(p_p, AT_DECNET, p,pl);
      break;
   case 0x0000:  /* 802.3 packet other than SNAP */
      switch (lsap) {
      case 0xFFFF:
         if (proto_reqd[AT_NOVELL]) r = pkt_extract(p_p, AT_NOVELL, p,pl);
         break;  /* Novell "Raw 802.2" doesn't use LLC */
# ifdef CLNS
      case 0xFEFE:  /* CLNS */
         if (proto_reqd[AT_CLNS]) r = pkt_extract(p_p, AT_CLNS, p+2,pl-2);
         break;  /* ISO 8473-1 uses 1-org, byte 0 is LLC control byte */
# endif
      case 0xE0E0:  /* Novell 802.2 */
         if (proto_reqd[AT_NOVELL]) r = pkt_extract(p_p, AT_NOVELL, p+3,pl-3);
         break;
/*    case 0xF0F0:  NetBIOS */
         }
      }

   if (r == 0 && proto_reqd[AT_OTHER]) {
      other_extract(p_p, ether_type,lsap);  r = 1;
      }

   if (adj_reqd && p_p->PeerAddrType != AT_DUMMY) {
      p_p->Low.AdjAddr_ms4 = *(Bit32 *)&ethp[6];  /* Source */
      p_p->Low.AdjAddr_ls2 = *(Bit16 *)&ethp[10];
      p_p->High.AdjAddr_ms4 = *(Bit32 *)&ethp[0];  /* Dest */
      p_p->High.AdjAddr_ls2 = *(Bit16 *)&ethp[4];
      }

# if MULTI
#  if N_PME_THREADS == 4
#    define HG_MASK 0x03
#  elif N_PME_THREADS == 2
#    define HG_MASK 0x01
#  elif N_PME_THREADS != 1
#    error N_PME_THREADS not 1, 2 or 4 <<<<<
#  endif

#  if N_PME_THREADS > 1
   p_p->hash_group =
#   if QBYTE_PEER
      (p_p->Low.PeerAddress.qbyte ^ p_p->High.PeerAddress.qbyte) & HG_MASK;
#   else
      (p_p->Low.PeerAddress.peer32[(PEER_ADDR_LEN+3)/4 - 1] ^
         p_p->Low.PeerAddress.peer32[(PEER_ADDR_LEN+3)/4 - 1]) & HG_MASK;
#   endif
#  else
   p_p->hash_group = 0;  /* Only run one thread */
#  endif
# endif  /* MULTI */

#if !NETFLOW && !LFAP_METER  /* These use time() instead */
   /* We know the pkts are in timestamp order here! */
   t_seconds = p_p->arrival_time/1000000.0;
   if (t_seconds > last_t_sec) {
      last_t_sec = t_seconds;
# if MULTI
      ntm_semaphore_up("osp", &osp_semaphore);   /* Trigger osp() */
# else
      one_second_process();
# endif
      }
#endif

   if (r) {
      mutex_lock("hpk", &stream_mutex);
      r = lookup_stream(p_p, 1, 1);
      mutex_unlock("hpk", &stream_mutex);
#if !STR_ANAL
      return r;
#else
      if (r) {
         if (r == 2) {  /* New stream */
            update_host_list(0, 0, p_p->sp);
            update_host_list(1, 0, p_p->sp);
            }
         return 1;
         }
      else return 0;
#endif
      }
   return 0;
   }

#endif  /* !NF_CISCO_DATA && !LFAP_METER */


char *u_fmt(char *buf, Bit32 n)
{
   if (n < 1024)
      sprintf(buf, "[%u B]", n);
   else if (n < 1024*1024)
      sprintf(buf, "[%u kB]", (n+512)/1024);
   else
      sprintf(buf, "[%u MB]", (n+512*1024)/(1024*1024));
   return buf;
   }

unsigned show_memory(int show)
{
   char msg[90], nb[30];
   unsigned t, r,rh, f,fh, s,sd, d,de, p, o, k, hl;

   k = mxpkts*sizeof(struct pkt);
   r = mxrules*sizeof(struct rule);
   rh = mxrulehash*sizeof(Bit16);
   f = mxflows*sizeof(struct flow);
   fh = fthashmod*sizeof(struct flow *);
#if NEW_ATR
   s = mxstreams*sizeof(struct stream) + sfhashmod*sizeof(struct stream *);
   sd = mxstr_blocks*sizeof(struct stream_data);
   d = mxdistribs*sizeof(struct distribution);
   de = mxdistevents*sizeof(struct event);
   p = mxpktdatas*sizeof(struct pktdata);
# if STR_ANAL
   hl = mxhlists*sizeof(struct host_list);
# endif
# if OAM
   o = mxoamrecs*sizeof(struct OAMrec);
# endif
#else
   s = sd = 0;
   d = de = p = 0;
#endif
   t = k+r+rh + f+fh + s+sd + d+de + p;

   if (show) {
      display_msg(1,"Memory Usage Info ..");
      sprintf(msg, "Rules: %d max, %d free %s", 
         mxrules, mxrules-inuse_rules, u_fmt(nb,r));
      display_msg(0,msg);
      sprintf(msg, "RuleHash: %d max, %d free %s", 
         mxrulehash, mxrulehash-inuse_rule_hash, u_fmt(nb,rh));
      display_msg(0,msg);
      sprintf(msg, "Flows: %d max (%d B/flow) %s", 
         mxflows, sizeof(struct flow), u_fmt(nb,f));
      display_msg(0,msg);
      sprintf(msg, "FlowHash: %d slots %s",
         fthashmod, u_fmt(nb,fh));
      display_msg(0,msg);
#if NEW_ATR
      sprintf(msg, "TCP streams: %d max, %d free %s",
         mxstreams, n_free_streams, u_fmt(nb,s));
      display_msg(0,msg);
      sprintf(msg, "Flows with streams: %d alloc, %d free (%d B/str) %s",
         mxstr_blocks, n_free_str_blocks, sizeof(struct stream_data),
         u_fmt(nb,sd));
      display_msg(0,msg);
      sprintf(msg, "Distribs: %d max, %d free (%d B/dist) %s",
      mxdistribs, n_free_distribs, sizeof(struct distribution), u_fmt(nb,d));
      display_msg(0,msg);
      sprintf(msg, "Events: %d alloc, %d free %s",
         mxdistevents, n_free_events, u_fmt(nb,de));
      display_msg(0,msg);
      sprintf(msg, "Packet data: %d alloc, %d free (%d B/pkt) %s",
         mxpktdatas, n_free_pktdata, sizeof(struct pktdata), u_fmt(nb,p));
      display_msg(0,msg);
# if STR_ANAL
      sprintf(msg, "Host_lists: %d alloc, %d free (%d B/rec) %s",
         mxhlists, n_free_hlists, sizeof(struct host_list), u_fmt(nb,hl));
      display_msg(0,msg);
# endif
# if OAM
      sprintf(msg, "OAM records: %d alloc, %d free (%d B/rec) %s",
         mxoamrecs, n_free_oamrecs, sizeof(struct OAMrec), u_fmt(nb,o));
      display_msg(0,msg);
# endif
#endif
      sprintf(msg, "Pkt structs: %d alloc, %d free (%d B/rec) %s",
         mxpkts, n_free_pkts, sizeof(struct pkt), u_fmt(nb,k));
      display_msg(0,msg);

      sprintf(msg, "Total allocated data: %s", u_fmt(nb,t));
      display_msg(0,msg);
      }

   return t;
   }

#if NEW_ATR
void show_new_attrib(void)
{
   char msg[60];
   display_msg(1,"New Attribute Info ..");
   sprintf(msg, "Streams (-t): %d in use, %d max",
      mxstreams-n_free_streams, mxstreams);
   display_msg(0,msg);
   sprintf(msg,"  %d recovered", StreamsRecovered);
   display_msg(0,msg);
   if (stats_time != 0) {
      sprintf(msg,"  Creating   %.03f stream/s", (double)StreamsCreated/stats_time);
      display_msg(0,msg);
      sprintf(msg,"  Recovering %.03f stream/s", (double)StreamsRecovered/stats_time);
      display_msg(0,msg);
      }
   sprintf(msg, "Packet data (-a): %d in use, %d max",
      mxpktdatas-n_free_pktdata, mxpktdatas);
   display_msg(0,msg);
   sprintf(msg,"  %d recovered", PktsRecovered);
   display_msg(0,msg);
   if (stats_time != 0) {
      sprintf(msg,"  Creating   %.03f pktdata/s", (double)PktsCreated/stats_time);
      display_msg(0,msg);
      sprintf(msg,"  Recovering %.03f pktdata/s", (double)PktsRecovered/stats_time);
      display_msg(0,msg);
      }
   }
#endif

void show_chains(int reset)
{
   char msg[60];

   display_msg(1,"Ruleset Chain Info ..");
   sprintf(msg, "%d tests", rsc_n_searches);
   display_msg(0,msg);
   if (rsc_n_searches == 0) return;
   sprintf(msg, "Range: %d max, %d av", 
      rsc_mx_range, rsc_av_range/rsc_n_searches);
   display_msg(0,msg);
   sprintf(msg, "Tests: %d max, %d av", 
      rsc_mx_tests, rsc_av_tests/rsc_n_searches);
   display_msg(0,msg);

   if (reset) {
      rsc_n_searches = 
         rsc_mx_range = rsc_av_range =
         rsc_mx_tests = rsc_av_tests = 0;
      }
   }

Bit32 active_flows(void)
{
   Bit32 af = 0, j, n;
   struct rtinf_rec *rip;

   for (j = 0; j != sizeof(ri)/sizeof(struct rtinf_rec); ++j) {
      rip = &ri[j];
      if (rip->ri_Status != RS_ACTIVE &&
            rip->ri_Status != RS_NOTINSERVICE
         ) continue;
      n = rip->ri_FlowRecords < rip->ri_recoverable_flows ? 0 :
         rip->ri_FlowRecords - rip->ri_recoverable_flows;
      af += n;
      }
   return af;
   }

void show_flow_table(void)
{
   char msg[60];
   Bit32 f;
   display_msg(1,"Flow Table Info ..");
   sprintf(msg, "InactTime %d, Flood %d",
      InactivityTimeout,FloodMark);
   display_msg(0,msg);
   sprintf(msg, "Flows: %d active, %u used",
      active_flows(), nflows);
   display_msg(0,msg);
   sprintf(msg,"  %d recovered (GC: %u s, %u flow)",
      FlowsRecovered, gc_interval,gc_f);
   display_msg(0,msg);
   if (stats_time != 0) {
      f = FlowsCreated*1000L/stats_time;
      sprintf(msg,"  Creating   %d.%03lu flow/s", f/1000,f%1000);
      display_msg(0,msg);
      f = FlowsRecovered*1000L/stats_time;
      sprintf(msg,"  Recovering %d.%03lu flow/s", f/1000,f%1000);
      display_msg(0,msg);
      }
   }

void show_protocols(void)
{
   int j;
   char msg[60];
   display_msg(1,"Protocols being metered ..");
   sprintf(msg,"      %3.3s  %s",  adj_reqd ? "Yes" : " No", "Adjacent");
   display_msg(0,msg);
#if NF_OCX_BGP
   sprintf(msg,"      %3.3s  %s",  asn_lookup ? "Yes" : " No", "ASN lookup");
   display_msg(0,msg);
#endif
   for (j = 0; j != sizeof(proto_reqd)/sizeof(int); ++j) {
      if (proto_names[j][0] != '\0') {
         sprintf(msg,"  %2d  %3.3s  %s",
            j, proto_reqd[j] ? "Yes" : " No", proto_names[j]);
         display_msg(0,msg);
         }
      }
   }

void show_stream_stats(void)
{
   char msg[60];
   int af = active_flows(),
      as = mxstreams - n_free_streams;
   int s_count[SS_TERM+1];
   char *s_name[SS_TERM+1] = 
      {"free", "got", "unknown", "matched", "ignore", "term"};
   struct stream *sp;  int j, unk;
   struct pkt_match_info *pmi;
#define N_F 50
   int k, f_count[N_F+1], badf;
   Bit32 fx;

   memset(s_count, 0, sizeof(s_count));
   memset(f_count, 0, sizeof(f_count));
   badf = unk = 0;
//20   mutex_lock("sss", &stream_state_mutex);
   mutex_lock("sss", &stream_mutex);
   for (sp = sfa, j = 0; j != mxstreams; ++j, ++sp) {
      if (sp->state > SS_TERM) ++unk; 
      else ++s_count[sp->state];
      if (sp->state == SS_MATCHED) {
         pmi = &sp->f_pmi;
         for (k = 0; k != pmi->n_flows; ++k) {
	    fx = pmi->fx[k];
            if (fx > N_F ) ++badf; 
            else ++f_count[fx];
            }
         }
      }
//20   mutex_unlock("sss", &stream_state_mutex);
   mutex_unlock("sss", &stream_mutex);
   printf("Stream counts:\n");
   for (j = 0; j <= SS_TERM; ++j)
      printf("  %7s %d", s_name[j], s_count[j]);
   printf("  unknown %d\n", unk);
   for (j = 1; j <= N_F; ++j) {
      if (f_count[j] != 0)
         printf(" %d:%d", j,f_count[j]);
      }
   printf(" >%d|%d\n", N_F,badf);

   if (af == 0) return;  /* Don't divide by zero */
   display_msg(1,"Stream Statistics ..");
   snprintf(msg,sizeof(msg), " %u streams over %u flows (%d s/f)",
      as, af, as/af);
   display_msg(0, msg);
   snprintf(msg,sizeof(msg), " %u matches, %u were cached (%0.2f %%)",
      n_matches,n_cached_matches, 100.0*n_cached_matches/n_matches);
   display_msg(0, msg);
   if (stats_time != 0) {
      sprintf(msg,"  Creating   %02.2f stream/s",
         ((double)StreamsCreated)/stats_time);
      display_msg(0,msg);
      sprintf(msg,"  Recovering %02.2f stream/s",
         ((double)StreamsRecovered)/stats_time);
      display_msg(0,msg);
      }
   }

void zero_stats(int show_msg)
{
   FlowsCreated = FlowsRecovered =
#if NEW_ATR
      PktsCreated = PktsRecovered =  StreamsCreated = StreamsRecovered =
#endif
      n_cached_matches = n_matches = n_hash_compares = n_hash_searches = 0;
   clear_pkt_stats = 1;
   if (show_msg) display_msg(0,"Statistics Zeroed");
   }

void show_stats(void)
{
   int i,j;
   double aps;
   Bit32 apb, kdp;
   char msg[60];
   if (stats_time == 0 || npackets == 0)
      return;  /* Avoid divides by zero */

   display_msg(1,"Meter Statistics ..");
   aps = (double)npackets/(double)stats_time;
   sprintf(msg,"Av pkt/s %.02f, max pkt/s %u", aps,max_pkt_rate);
   display_msg(0,msg);

   sprintf(msg,"%d flows active (max %d)",active_flows(),mxflows);
   display_msg(0,msg);
   sprintf(msg,"%u masks in table (max %u)", n_masks,MXMASKS);
   display_msg(0,msg);
   i = (n_matches*100L+5L)/(npackets*10L);
   j = (n_hash_searches*100L+5L)/(npackets*10L);
   sprintf(msg,"%u.%u rules/pkt, %u.%u searches/pkt",
      i/10,i%10, j/10,j%10);
   display_msg(0,msg);
   if (n_hash_searches != 0) {
      i = (n_hash_compares*100L+5L)/(n_hash_searches*10L);
      sprintf(msg,"%u.%u compares/search", i/10,i%10);
      display_msg(0,msg);
      sprintf(msg,"%u hash slots, %u in use, ", fthashmod,n_hash_ents);
      }
   }

void show_rulesets(void)
{
   int c,j,t;
   struct rtinf_rec *rip;
   char buf[80];
   display_msg(1,"Ruleset Table ..");
   for (c = t = j = 0; j != sizeof(ri)/sizeof(struct rtinf_rec); ++j) {
      rip = &ri[j];
      if (rip->ri_Status != RS_ACTIVE &&
            rip->ri_Status != RS_NOTINSERVICE
         ) continue;
      ++c;
      if (t == 0) {
         display_msg(0,"   Rdy  Flows Rules  Name    Owner");
         t = 1;
         }
      sprintf(buf, "%2d  %c  %5lu  %5u  %s  %s",
         j+1,  /* 1-org */
         rip->ri_Status == RS_ACTIVE ? 'T' : 'F',
         rip->ri_FlowRecords,
         rip->ri_Size, rip->ri_Name,
         rip->ri_Owner);
      display_msg(0,buf);
      }
   if (c == 0) display_msg(0,"No ruleset rows active");
   }

void show_managers(void)
{
   int c,j,t;
   struct mgr_rec *mip;
   char buf[80];
   display_msg(1,"Manager Table ..");
   for (c = t = j = 0; j != sizeof(mi)/sizeof(struct mgr_rec); ++j) {
      mip = &mi[j];
      if (mip->mi_Status != RS_ACTIVE) continue;
      ++c;
      if (t == 0) {
         display_msg(0,"  Crnt Stby HWM RSby  Owner");
         t = 1;
         }
      sprintf(buf, "%2d  %2d  %2d  %3d   %c  %s",
         j+1,  /* 1-org */
         mip->mi_CurrentRuleSet, mip->mi_StandbyRuleSet,
         mip->mi_HighWaterMark,
         mip->mi_RunningStandby == TV_TRUE ? 'T' : 'F',
         mip->mi_Owner);
      display_msg(0,buf);
      }
   if (c == 0) display_msg(0,"No manager rows active");
   }

void show_readers(void)
{
   int c,j,t;
   struct rdr_rec *cip;
   Bit32 now;
   char buf[80];
   display_msg(1,"Meter Reader Table ..");
   now = uptime_cs();
   for (c = t = j = 0; j != sizeof(ci)/sizeof(struct rdr_rec); ++j) {
      cip = &ci[j];
      if (cip->ci_Status != RS_ACTIVE) continue;
      ++c;
      if (t == 0) {
         display_msg(0,"    Rsx  Last  Prev  Owner");
         t = 1;
         }
      sprintf(buf, "%2d  %2d  %5lu %5lu  %s",
         j+1,  /* 1-org */
         cip->ci_RuleSet,
         (now - cip->ci_LastTime)/100L, (now - cip->ci_PrevTime)/100L,
         cip->ci_Owner);
      display_msg(0,buf);
      }
   if (c == 0) display_msg(0,"No reader rows active");
   }

#if MULTI
# if THREAD_TRACE || MUTEX_TRACE
double ntm_start_s;
# endif
# if THREAD_TRACE
void mutex_lock(char *caller, pthread_mutex_t *m)
{
   struct timeval t; double ml_s, mp_s, w;

   gettimeofday(&t, (struct timezone *)0);
   ml_s = t.tv_sec + t.tv_usec/1000000.0;
   fprintf(multi_log, "c %s %x\n", caller, m);  fflush(multi_log);
   tw_pthread_mutex_lock(m);

   gettimeofday(&t, (struct timezone *)0);
   mp_s = t.tv_sec + t.tv_usec/1000000.0;

   if ((w = mp_s - ml_s) > 0.000150) {  /* 150 us */
      fprintf(multi_log, "t %12.6f %10.6f m %3s %x\n",
         ml_s - ntm_start_s, w, caller, m);
      fflush(multi_log);
      }
   }

void ntm_semaphore_down(char *caller, Semaphore *s)
{
   struct timeval t;  double sl_s, sp_s, w;
   int v;

   gettimeofday(&t, (struct timezone *)0);
   sl_s = t.tv_sec + t.tv_usec/1000000.0;

   v = semaphore_down(s);

   gettimeofday(&t, (struct timezone *)0);
   sp_s = t.tv_sec + t.tv_usec/1000000.0;

   if ((w = sp_s - sl_s) > 0.000300) {  /* 300 us */
      fprintf(multi_log, "t %12.6f %10.6f s %3s %x\n",
         sl_s - ntm_start_s, w, caller, s);
      fflush(multi_log);
      }
   }
# endif

void init_mutex(pthread_mutex_t *m, char *name)
{
# if THREAD_TRACE || MUTEX_TRACE
   fprintf(multi_log, "a m %x %s\n", m, name);
# endif
   if (pthread_mutex_init(m, pthread_mutexattr_default) == -1)
      log_msg(LOG_ERR, 21, "Couldn't init %s", name);

   }

void init_mutexes(void)
{
# if THREAD_TRACE || MUTEX_TRACE
   struct timeval t;
   gettimeofday(&t, (struct timezone *)0);
      /* Initiallise threads tod clock */
   ntm_start_s = t.tv_sec + t.tv_usec/1000000.0;
   if ((multi_log = fopen("multi_log.txt", "w")) == NULL)
      log_msg(LOG_ERR, 999, "Couldn't open multi_log.txt file");
# endif

   init_mutex(&pkt_fl_mutex, "pkt_fl_mutex");
   init_mutex(&dist_fl_mutex, "dist_fl_mutex");
   init_mutex(&event_fl_mutex, "event_fl_mutex");
   init_mutex(&stdata_fl_mutex, "stdata_fl_mutex");
   init_mutex(&stream_fl_mutex, "stream_fl_mutex");
   init_mutex(&pktdata_fl_mutex, "pktdata_fl_mutex");
//20   init_mutex(&stream_state_mutex, "stream_state_mutex");
   init_mutex(&stream_mutex, "stream_state_mutex");
# if OAM
   init_mutex(&oamrec_fl_mutex, "oamrec_fl_mutex");
# endif
# if STR_ANAL
   init_mutex(&hlist_fl_mutex, "hlist_fl_mutex");
# endif
   init_mutex(&flow_table_mutex, "flow_table_mutex");
//20   init_mutex(&stream_q_mutex, "stream_q_mutex");
//20   init_mutex(&strhash_q_mutex, "strhash_q_mutex");
   init_mutex(&pktdata_q_mutex, "pktdata_q_mutex");
   init_mutex(&dev_q_mutex, "dev_q_mutex");
   init_mutex(&event_q_mutex, "event_q_mutex");
# if OAM
   init_mutex(&oam_q_mutex, "oam_q_mutex");
# endif
   init_mutex(&if_pkt_q_mutex, "start_time_mutex");
   init_mutex(&if_pkt_q_mutex, "log_file_mutex");
   init_mutex(&if_pkt_q_mutex, "if_pkt_q_mutex");

# if MUTEX_TRACE && !THREAD_TRACE
   fclose(multi_log);
# endif
   }

void destroy_mutex(pthread_mutex_t *m, char *name)
{
   if (pthread_mutex_destroy(m) == -1)
     log_msg(LOG_ERR, 21, "Couldn't destroy %s", name);
   }

void destroy_mutexes(void)
{
   destroy_mutex(&pkt_fl_mutex, "pkt_fl_mutex");
   destroy_mutex(&dist_fl_mutex, "dist_fl_mutex");
   destroy_mutex(&event_fl_mutex, "event_fl_mutex");
   destroy_mutex(&stdata_fl_mutex, "stdata_fl_mutex");
   destroy_mutex(&stream_fl_mutex, "stream_fl_mutex");
   destroy_mutex(&pktdata_fl_mutex, "pktdata_fl_mutex");
//20   destroy_mutex(&stream_state_mutex, "stream_state_mutex");
   destroy_mutex(&stream_mutex, "stream_mutex");
# if OAM
   destroy_mutex(&oamrec_fl_mutex, "oamrec_fl_mutex");
# endif
# if STR_ANAL
   destroy_mutex(&hlist_fl_mutex, "hlist_fl_mutex");
# endif
   destroy_mutex(&if_pkt_q_mutex, "start_time_mutex");
   destroy_mutex(&if_pkt_q_mutex, "log_file_mutex");
   destroy_mutex(&if_pkt_q_mutex, "if_pkt_q_mutex");

   destroy_mutex(&flow_table_mutex, "flow_table_mutex");
//20   destroy_mutex(&stream_q_mutex, "stream_q_mutex");
//20   destroy_mutex(&strhash_q_mutex, "strhash_q_mutex");
   destroy_mutex(&pktdata_q_mutex, "pktdata_q_mutex");
   destroy_mutex(&dev_q_mutex, "dev_q_mutex");
   destroy_mutex(&event_q_mutex, "event_q_mutex");
# if OAM
   destroy_mutex(&oam_q_mutex, "oam_q_mutex");
# endif
   }

# if THREAD_TRACE
void ntm_semaphore_init(char *name, int id, Semaphore *s)
{
   if (id != 0) fprintf(multi_log, "a s %x %s.%d\n", s, name,id);
   else fprintf(multi_log, "a s %x %s\n", s, name);
   semaphore_init(s);
   }
# endif
#endif

unsigned init_monitor(int n_interfaces)
{
   int j;

   start_uptime_clock();
   if ((pkta = (struct pkt *)malloc(
        sizeof(struct pkt)*mxpkts)) == NULL)
      log_msg(LOG_ERR, 1, "Not enough memory for pkt data!");
   free_pkts = pkta;  n_free_pkts = mxpkts;
   for (j = 0; j != mxpkts-1; ++j) pkta[j].next_pp = &pkta[j+1];
   pkta[mxpkts-1].next_pp = NULL;
#if PKT_DEBUG >= 1
   for (j = 0; j != mxpkts; ++j) pkta[j].pkt_state = 0;
#endif

   if (mxflows < 30) {
      log_msg(LOG_ERR, 2, "-f specified < 30 flows!  Using -f 50");
      mxflows = 50;
      }
   if ((fa = (struct flow *)malloc(
        sizeof(struct flow)*mxflows)) == NULL)
      log_msg(LOG_ERR, 3, "Not enough memory for flow table!");
   for (j = 0; j != mxflows; ++j) fa[j].FlowRuleSet = 0;

   for (j = 0; j != mxflows-1; ++j) fa[j].rs_next = j+2;  /* 1-org */
   fa[mxflows-1].rs_next = 0;
   free_flow_head = 1;  free_flow_tail = mxflows;

   rsc_n_searches = 
      rsc_mx_range = rsc_av_range =
      rsc_mx_tests = rsc_av_tests = 0;

#if FLOW_INDEX
   if ((flow_ix = (struct flow **)malloc(
         sizeof(struct flow *)*mxflows)) == NULL)
      log_msg(LOG_ERR, 4, "Not enough memory for flow index!");
   for (j = 0; j != mxflows; ++j) flow_ix[j] = NULL;
#endif

   nflows = 0;  /* No accounting flows in use yet */
   fthashmod = mxflows;  /* Doesn't need to be mutually prime */

#if 0
   for (j = sizeof(p2primes)/sizeof(Bit32)-1; 
      j != 8 && p2primes[j] > mxflows; --j) ;
   fthashmod = p2primes[j];  /* Min fthashmod = p2primes[8] (2039) */
#endif
   flow_ht = (struct flow **)malloc(fthashmod*sizeof(struct flow *));
   if (flow_ht == NULL)
      log_msg(LOG_ERR, 5, "Not enough memory for flow hash table!");
   memset(flow_ht, 0, fthashmod*sizeof(struct flow *));

   if ((rule_space = (struct rule *)malloc(
         sizeof(struct rule)*mxrules)) == NULL)
      log_msg(LOG_ERR, 6, "Not enough memory for rulesets!");
   inuse_rules = 0;

   mxrulehash = 1 + mxrules*3/2;
   if ((rule_hash_space = (Bit16 *)malloc(
         sizeof(Bit16)*mxrulehash)) == NULL)
      log_msg(LOG_ERR, 7, "Not enough memory for rule hashes!");
   inuse_rule_hash = 0;

   if ((sfa = (struct stream *)malloc(
        sizeof(struct stream)*mxstreams)) == NULL)
      log_msg(LOG_ERR, 8, "Not enough memory for streams!");
   free_streams = sfa;  n_free_streams = mxstreams;
   for (j = 0; j != mxstreams-1; ++j) {
      sfa[j].next_sp = &sfa[j+1];  sfa[j].state = SS_FREE;
      }
   sfa[mxstreams-1].next_sp = NULL;  sfa[mxstreams-1].state = SS_FREE;
   free_stream_tail = &sfa[mxstreams-1];
   sfhashmod = mxstreams;  /* Doesn't need to be mutually prime */
   t_sq = t_sq_tail = NULL;  t_active_streams = 0;

#if 0
   for (j = sizeof(p2primes)/sizeof(Bit32)-1; 
      j != 8 && p2primes[j] > mxstreams; --j) ;
   sfhashmod = p2primes[j];
#endif
   sfht = (struct stream **)malloc(sfhashmod*sizeof(struct stream *));
   if (sfht == NULL)
      log_msg(LOG_ERR, 9, "Not enough memory for stream hash table!");
   memset(sfht, 0, sfhashmod*sizeof(struct stream *));

   if ((strb = (struct stream_data *)malloc(
        sizeof(struct stream_data)*mxstr_blocks)) == NULL)
      log_msg(LOG_ERR, 10, "Not enough memory for stream data blocks!");
   free_str_blocks = strb;  n_free_str_blocks = mxstr_blocks;
   for (j = 0; j != mxstr_blocks-1; ++j) strb[j].next = &strb[j+1];
   strb[mxstr_blocks-1].next = NULL;

#if NEW_ATR
   if ((distribs = (struct distribution *)malloc(
        sizeof(struct distribution)*mxdistribs)) == NULL)
      log_msg(LOG_ERR, 11, "Not enough memory for distributions!");
   free_distribs = distribs;  n_free_distribs = mxdistribs;
   for (j = 0; j != mxdistribs-1; ++j) distribs[j].next = &distribs[j+1];
   distribs[mxdistribs-1].next = NULL;

   if ((events = (struct event *)malloc(
        sizeof(struct event)*mxdistevents)) == NULL)
      log_msg(LOG_ERR, 12, "Not enough memory for distrib events!");
   free_events = events;  n_free_events = mxdistevents;
   for (j = 0; j != mxdistevents-1; ++j) {
      events[j].next = &events[j+1];  events[j].type = FREE_EVENT;
      }
   events[mxdistevents-1].next = NULL;
   events[mxdistevents-1].type = FREE_EVENT;

   if ((packet_data = (struct pktdata *)malloc(
        sizeof(struct pktdata)*mxpktdatas)) == NULL)
      log_msg(LOG_ERR, 13, "Not enough memory for packet data blocks!");
   free_pkt_dat = packet_data;  n_free_pktdata = mxpktdatas;
   for (j = 0; j != mxpktdatas-1; ++j) packet_data[j].next = &packet_data[j+1];
   packet_data[mxpktdatas-1].next = NULL;

# if OAM
   if ((oar = (struct OAMrec *)malloc(
        sizeof(struct OAMrec)*mxoamrecs)) == NULL)
      log_msg(LOG_ERR, 14, "Not enough memory for OAM data records!");
   free_oamrecs = oar;  n_free_oamrecs = mxoamrecs;
   for (j = 0; j != mxoamrecs-1; ++j) oar[j].next = &oar[j+1];
   oar[mxoamrecs-1].next = NULL;
   memset(oaf, 0, sizeof(oaf));
# endif

# if STR_ANAL
   if ((hla = (struct host_list *)malloc(
        sizeof(struct host_list)*mxhlists)) == NULL)
      log_msg(LOG_ERR, 15, "Not enough memory for host_lists!");
   free_hlists = hla;  n_free_hlists = mxhlists;
   for (j = 0; j != mxhlists-1; ++j) hla[j].next = &hla[j+1];
   hla[mxhlists-1].next = NULL;
# endif
#endif  /* NEW_ATR */

   memset(ri, 0, sizeof(dist_atr));  memset(ri, 0, sizeof(stream_atr));
   for (j = 0; ; ++j) {
      if (attribs[j].attr == 0) break;
      dist_atr[attribs[j].attr] = attribs[j].distrib;
      stream_atr[attribs[j].attr] = attribs[j].stream;
      }
   memset(ri, 0, sizeof(atr_ta));  memset(ri, 0, sizeof(atr_oa));
   for (j = 0; ; ++j) {
      if (addr_atrs[j].attr == 0) break;
      atr_ta[addr_atrs[j].attr] = addr_atrs[j].ta;
      atr_oa[addr_atrs[j].attr] = addr_atrs[j].oa;
      }

   memset(ri, 0, sizeof(ri));
   memset(rs_rdr, 0, sizeof(rs_rdr));  /* For nifty */
   /* Create a ruleset info entry for the default rule table */
   ri[0].ri_rule_table = (struct rule *)default_rule_table;
   ri[0].ri_Size = DRT_SIZE;
   optimise_rule_table(1);  /* Alloc rule hash */
   memcpy(ri[0].ri_Owner, "NeTraMet", 8);
   memcpy(ri[0].ri_Name, "1", 1);  /* For compatibility with v3 */
   ri[0].ri_TimeStamp = 1;
   ri[0].ri_Status = RS_ACTIVE;

   tbl_size[2] = n_interfaces;  /* Interface table */

   memset(mi, 0, sizeof(mi));  /* Manager info */
   for (j = 1; j != sizeof(mi)/sizeof(struct mgr_rec); ++j)
      mi[j].mi_RunningStandby = TV_FALSE;
   memcpy(mi[0].mi_Owner, "NeTraMet", 8);
   mi[0].mi_CurrentRuleSet = 1;
   mi[0].mi_RunningStandby = TV_FALSE;
   mi[0].mi_TimeStamp = 1;
   mi[0].mi_Status = RS_ACTIVE;

   for (j = 0; j != sizeof(proto_reqd)/sizeof(int); ++j) proto_reqd[j] = 0;
   adj_reqd = 0;
   open_rule_set(1, 1); /* Open set 1 */

   memset(ci, 0, sizeof(ci));  /* Reader (collector) info */

#if USER_INFO
   UserInfoInit();
#endif

#if APP_DATA
# if OLD_P2P_TEST
   for (j = 1; j != n_strings; ++j)
      strings[j].strlen = strlen(strings[j].str);
# endif
#endif

#if CHECK_UPTIME_DECR
   for (j = 0; j != N_PME_THREADS; ++j)
      memset(&utc_data[j], 0, sizeof(struct uptime_check));
#endif

   return show_memory(0);
   }

void show_help(void)
{
   display_msg(0,"Copyright (C) 1992-2003");
   display_msg(0,"  Nevil Brownlee");
   display_msg(0,"  ITSS Technology Development");
   display_msg(0,"  The University of Auckland");
   display_msg(0,"");
   display_msg(0,"Keyboard commands ..");
   display_msg(0,"");
   display_msg(0,"  e: show mEter reader table");
   display_msg(0,"  a: show mAnager table");
   display_msg(0,"  u: show rUleset table");
   display_msg(0,"  f: show Flow Table");
   display_msg(0,"  p: show Protocols");
   display_msg(0,"  s: show Statistics");
   display_msg(0,"  v: show meter Version");
   if (kb_enabled) {
      display_msg(0,"");
      display_msg(0,"ESC ESC: stop metering, exit NeTraMet");
      }
   }

void handle_kb(int ch)
{
   int c;
   switch (ch) {
   case 'a':
      show_managers();
      break;
#if NEW_ATR
   case 'd':
      show_new_attrib();
      break;
#endif
   case 'C':
   case 'c':
      show_chains(ch == 'C');
      break;
   case 'e':
      show_readers();
      break;
   case 'f':
      show_flow_table();
      break;
   case 'h':
      show_flow_hash();
      break;
   case 'H':
      show_stream_hash(1);
      break;
   case 'm':
      show_memory(1);
      break;
   case 'p':
      show_protocols();
      break;
   case 'u':
      show_rulesets();
      break;
   case 's':
      show_stats();
      break;
   case 'S':
      show_stream_stats();
      break;
   case 't':
      show_meter_time();
      break;
#if PM_DEBUG
   case '\\':
      printf("pm_debug: ");
      c = getchar();
      if (c == '\\') {
         pm_dbg = 0;
         printf("ri[] off");
         }
      else {
         c -= '0';
         pm_dbg = c == 0 ? 10 : c;
         printf("ri[%d]", pm_dbg);
         }
      break;
#endif
   case 'z':
      zero_stats(1);
      break;
   case '?':
      show_help();
      break;
   default:
      break;
      }
   display_msg(2, "");  /* fflush(stdout) */
   }
